###############################################################################
#                 BAYESIAN TREEVARIATE COPULA LEARNING                        #
###############################################################################

###### Generic MCMC function

##
## Calls copula MCMC method. Files will be stored accordingly.
##

copula.tree.mcmc <-
   function(simulation.size, # Number of MCMC iterations
            train.x,         # Training set
            test.x,          # Test set. Can be set to NULL.
            marg.type,       # Vector of strings specifying marginal type, or a
                             # single string
                             #     ("norm", "t", "gamma" or "exponential")
            marg.priors,     # List of priors
            marg.mh,         # Metropolis-Hastings parameters
            cop.type,        # String specifying copula type
                             #     ("norm", "frank", "gb", "amh", "clayton", "fgm")
            cop.priors,      # List of priors
            cop.mh,          # Metropolis-Hastings parameters
            mix.priors,      # Dirichlet process mixture hyperparameters
            opt              # General MCMC options
            )
{
  if (!is.null(test.x) && sum(sum(is.na(test.x))) > 0)               
      stop("\n INPUT PROBLEM: missing values in test set are not allowed!\n")

  c.train.x <- complete.set(train.x)
  stacked.data.train <- stack.data(c.train.x)
  stacked.data.test  <- stack.data(test.x)
  n.vars <- ncol(train.x)
  
  if (length(marg.type) != 1 && length(marg.type) != n.vars)
      stop("\n INPUT PROBLEM: number of marginal types should be one or equal to number of variables!\n")
   if (!is.null(test.x) && ncol(test.x) != n.vars)
      stop("\n INPUT PROBLEM: number of variables in test set should be equal to number in train set!\n")

  all.options <- validate.options(opt, train.x)
  
  if (is.null(mix.priors)) {
      mix.priors <- list()
      mix.priors$a <- 0.1
      mix.priors$b <- 1
  } else if (!is.numeric(mix.priors$a) || !is.numeric(mix.priors$b) ||
             mix.priors$a < 0 || mix.priors$b < 0) 
      stop("\n INPUT PROBLEM: two non-negative hyperparameters for the DP mixture are necessary!\n")
      
  if (all.options$init.cluster.t == "mle")
      initial.clustering <- cluster.data(c.train.x, all.options$initial.K, 
                                         marg.type, cop.type)
  else
      initial.clustering <- cluster.data(c.train.x, all.options$initial.K, 
                                         marg.type, "")
  initial.models <- mle.trees.array.C(c.train.x, initial.clustering,
                                      all.options$initial.K, cop.type)
  initial.trees  <- initial.models$trees
  initial.params <- initial.models$params

  r <- simple.mcmc.treevariate(TRUE, 
          marg.type, marg.priors, marg.mh, cop.type, cop.priors, cop.mh,
          simulation.size, all.options$burn.in, all.options$mcmc.step,
          ncol(train.x), nrow(train.x), stacked.data.train,
          all.options$missing.pattern,
          nrow(test.x), stacked.data.test, recode.missing.pattern(test.x), 
          mix.priors$a, mix.priors$b, 
          all.options$initial.K, initial.clustering, initial.trees, initial.params,
          all.options$mcmc.options, c.train.x)
  return(r)
}

## validate.options:
##
## Checks whether options are valid, and, most importantly set 
## default values for those options not present.
##
##   - burn.in: initial samples to ignore when doing evaluation
##   - mcmc.step: dump sample to file only at every "mcmc.step".
##                Set this to > 1 to save disk space while discarding
##                correlated samples.
##   - missing.pattern: specification of which entries should be considered 
##                      "missing" (if set to NULL, then it is interpreted as 
##                       no missing data)
##   - initial.K: initial number of clusters
##   - init.cluster.t: clustering initialization strategy 
##                     ("euclidean" or "mle")
##   - tree.sampler: proposal type for sampling trees (0, 1, 2 or 3)
##   - slice.sampler: use slice sampler? (TRUE or FALSE)
##   - fix.marginals: fix marginals after fitting them? (TRUE or FALSE)
##   - save.missing: save any imputed missing variables? (TRUE or FALSE)
##   - save.corr: save any implied correlation matrix?
##                (Gaussian copula only: TRUE or FALSE)

validate.options <- function(opt, train.x)
{
   # burn in and mcmc step
   if (is.null(opt$burn.in)) {
       opt$burn.in <- 1
   } else if (!is.numeric(opt$burn.in)) {
       stop("\n INPUT PROBLEM: invalid number of burn-in steps!\n")
   } else {
       opt$burn.in <- as.integer(opt$burn.in)
       if (opt$burn.in <= 0)
           stop("\n INPUT PROBLEM: invalid number of burn-in steps!\n")
   }
   if (is.null(opt$mcmc.step)) {
       opt$mcmc.step <- 1
   } else if (!is.numeric(opt$mcmc.step)) {
       stop("\n INPUT PROBLEM: invalid number of MCMC thinning steps!\n")
   } else {
       opt$mcmc.step <- as.integer(opt$mcmc.step)
       if (opt$mcmc.step <= 0)
           stop("\n INPUT PROBLEM: invalid number of MCMC thinning steps!\n")
   }
        
   # Cluster initialization fields
   if (is.null(opt$init.cluster.t))
       opt$init.cluster.t <- "euclidean"
   else if (opt$init.cluster.t != "euclidean" && opt$init.cluster.t != "mle")
       stop("\n INPUT PROBLEM: invalid cluster initialization!\n")
   if (is.null(opt$initial.K)) {
       opt$initial.K <- 2
   } else if (typeof(opt$initial.K) != "double" && typeof(opt$initial.K) != "integer") {
       stop("\n INPUT PROBLEM: invalid number of initial clusters!\n")
   } else {
       opt$initial.K <- as.integer(opt$initial.K)
       if (opt$initial.K <= 0)
           stop("\n INPUT PROBLEM: invalid number of initial clusters!\n")
   }
   
   # missing.pattern 
   if (is.null(opt$missing.pattern)) {
       opt$missing.pattern <- matrix(as.double(!is.na(train.x)), 
                                     ncol = ncol(train.x))
   } else if (!is.matrix(opt$missing.pattern)) {
       stop("\n INPUT PROBLEM: invalid missing pattern specification!\n")
   } else if (ncol(opt$missing.pattern) != ncol(train.x) || nrow(opt$missing.pattern) != nrow(train.x)) {
       stop("\n INPUT PROBLEM: missing pattern matrix should have same dimensions as training data!\n")   
   } else if (sum(sum(opt$missing.pattern == 1)) + sum(sum(opt$missing.pattern == 0)) != ncol(train.x) * nrow(train.x)) {
       stop("\n INPUT PROBLEM: missing pattern matrix should have zeros and ones only!\n")   
   }
   opt$missing.pattern <- stack.data(opt$missing.pattern)
   
   # Tree sampler
   if (is.null(opt$tree.sampler))
       opt$tree.sampler <- 0
   else if (opt$tree.sampler != 0 && opt$tree.sampler != 1 && 
            opt$tree.sampler != 2 && opt$tree.sampler != 3) 
       stop("\n INPUT PROBLEM: invalid tree sampler choice!\n")
   
   # binary indicators
   if (is.null(opt$slice.sampler))
       opt$slice.sampler <- FALSE
   else if (opt$slice.sampler != TRUE && opt$slice.sampler != FALSE)
       stop("\n INPUT PROBLEM: invalid slice sampler choice!\n")
   if (is.null(opt$fix.marginals))
       opt$fix.marginals <- FALSE
   else if (opt$fix.marginals != TRUE && opt$fix.marginals != FALSE)
       stop("\n INPUT PROBLEM: invalid fix marginals choice!\n")
   if (is.null(opt$save.missing))
       opt$save.missing <- FALSE
   else if (opt$save.missing != TRUE && opt$save.missing != FALSE)
       stop("\n INPUT PROBLEM: invalid missing data save choice!\n")
   if (is.null(opt$save.corr))
       opt$save.corr <- FALSE
   else if (opt$save.corr != TRUE && opt$save.corr != FALSE)
       stop("\n INPUT PROBLEM: invalid fix marginals choice!\n")
   if (is.null(opt$predict.from.file))
       opt$predict.from.file <- FALSE
   else if (opt$predict.from.file != TRUE && opt$predict.from.file != FALSE)
       stop("\n INPUT PROBLEM: invalid fix marginals choice!\n")
   
   # General MCMC options
   mcmc.options <- rep(1, 7)
   mcmc.options[1] <- as.double(opt$save.missing == TRUE)
   mcmc.options[2] <- as.double(opt$predict.from.file == TRUE)
   mcmc.options[3] <- as.double(opt$save.corr == TRUE)
   mcmc.options[4] <- as.double(opt$tree.sampler)   
   mcmc.options[5] <- as.double(opt$fix.marginals == TRUE)
   mcmc.options[6] <- as.double(opt$slice.sampler == TRUE)
   
   return(list(initial.K = opt$initial.K, init.cluster.t = opt$init.cluster.t,
               burn.in = opt$burn.in, mcmc.step = opt$mcmc.step,
               missing.pattern = opt$missing.pattern, mcmc.options = mcmc.options))
}
