## Core graph algorithms
## Current version: undirected graphs only
## Ricardo Silva, November 29th 2007

## class ugraph: created from adjacency matrix, using function as.ugraph
##
## utgraph is an undirected graph. To creat an utree, pass as a parameter the
## adjacency matrix of it. The fields in this class are:
##
## - adj.matrix: the adjacency matrix 
## - adj.list:   same information, as an adjacency list
## - edges:      same information, but now as a list with all edges
## - edge.index: given a pair of node indices (i, j), stores the position of
##               the respective edge within the list 'edges'
## - num.nodes,
##   num.edges:  self-explanatory

as.ugraph <- function(matrix) {
   adj.info <- get.adjlist.u(matrix)
   graph <- list(adj.matrix = matrix, adj.list = adj.info$adj.list,
                 edges = adj.info$edges, edge.index = adj.info$edge.index,
                 num.nodes = nrow(matrix), num.edges = adj.info$num.edges)
   return(graph)
}


## class utree: created from adjacency matrix, using function as.utree
##
## utree is an undirected tree. To creat an utree, pass as a parameter the
## adjacency matrix of it. The fields in this class, besides the ones in
## ugraph, are:
##
## - bf.order:   "breadth-first" order. An order that defines a directed tree.
##               Nodes later in the order are never parents of nodes earlier
##               in the order
## - parents:    the parent of each node according to the directed version of 
##               the tree. Useful for sampling                 

as.utree <- function(matrix) {
   adj.info <- get.adjlist.u(matrix)
   bf.order <- get.breadth.first.order(nrow(matrix), adj.info$adj.list)
   graph <- list(adj.matrix = matrix, adj.list = adj.info$adj.list,
                 edges = adj.info$edges, edge.index = adj.info$edge.index,
                 bf.order = bf.order[[1]], parents = bf.order[[2]],
                 num.nodes = nrow(matrix), num.edges = adj.info$num.edges)
   if (!is.tree(graph))
       error("Not a tree!")
   return(graph)
}

## Generate adjacency information from an adjacency matrix

get.adjlist.u <- function(matrix) {
   nr <- nrow(matrix); nc <- ncol(matrix)
   adj.list <- list(nr)   
   num.edges <- 0
   for (i in 1:nr) {
       n.adj <- sum(matrix[i, ] > 0)
       num.edges <- num.edges + n.adj
       adj.list[[i]] <- rep(0, n.adj)
       pos <- 1
       for (j in 1:nc) {
           if (matrix[i, j] > 0) {
               adj.list[[i]][pos] <- j
               pos <- pos + 1
           }
       }
   }
   num.edges <- num.edges / 2
   edges <- list(num.edges)
   edge.index <- matrix(nrow = nr, ncol = nr)
   pos <- 1
   for (i in 1:nr) {
       nn <- length(adj.list[[i]])
       for (j in 1:nn) {
           if (i < adj.list[[i]][j]) {
               edges[[pos]] <- c(i, adj.list[[i]][j])
               edge.index[i, adj.list[[i]][j]] <- pos
               edge.index[adj.list[[i]][j], i] <- pos
               pos <- pos + 1
           }
       }
   }
   return(list(adj.list = adj.list, num.edges = num.edges, edges = edges,
               edge.index = edge.index))
}

## Generate an order for the nodes of an undirected graph based on breadth-first 
## search. Useful for sampling from undirected trees. The graph is encoded as
## an adjacency list adj.list. THIS CURRENTLY ASSUMES THAT THE GRAPH IS CONNECTED.

get.breadth.first.order <- function(num.nodes, adj.list) {

    breadth.first.order.recursion <- function(current, adj.list) {
        nn <- length(adj.list[[current]])
        for (i in 1:nn) {
             if (!marked[adj.list[[current]][i]]) {
                 level <<- level + 1
                 b.order[level] <<- adj.list[[current]][i]        
                 parents[adj.list[[current]][i]] <<- current
             }
        }
        for (i in 1:nn) {
             if (!marked[adj.list[[current]][i]]) {
                 marked[adj.list[[current]][i]] <<- TRUE
                 breadth.first.order.recursion(adj.list[[current]][i], adj.list)
             }
        }    
    }

    b.order <- rep(0, num.nodes)
    parents <- rep(0, num.nodes)
    marked  <- rep(FALSE, num.nodes)
    
    parents[1] <- 0
    b.order[1] <- 1
    marked[1]  <- TRUE #default root is 1
    level <- 1
    
    breadth.first.order.recursion(1, adj.list) 

    return(list(b.order, parents))
}

## Check if undirected graph is a tree. This is done by depth-first search.
## input: object of type ugraph

is.tree <- function(ug) {
    
    is.tree.recursion <- function(node, parent) {
        marked[node] <<- TRUE
        for (i in ug$adj.list[[node]]) {        
            if (i == parent)
                next
            if (marked[i])
                return(FALSE)
            result <- is.tree.recursion(i, node)
            if (!result)
                return(FALSE)
        }
        return(TRUE)
    }
    
    marked <- rep(FALSE, ug$num.nodes)    
    is.tree.recursion(1, -1)
}

