###############################################################################
#                 BAYESIAN TREEVARIATE COPULA LEARNING                        #
###############################################################################

## encode.marginals:
##
## Checks the initialization for a specified marginal and encodes it as a 
## numerical vector.
##
## "marginal" should be a string, while "marginal.priors" is a listcontaining the
## prior hyperparameters, and "marginal.mh" contains the Metropolis-Hastings 
## parameters
##
## The initialization list "theta" specifies the parameter values.
##

encode.marginals <- function(marginal, theta, marg.priors, marg.mh)
{
    # check that theta, marg.priors and marg.mh should be lists
    if(!is.list(cop.priors) || !is.list(cop.mh))
        stop("build.marginals: input should be lists")

    ## for normal marginals
    if (marginal == "norm") {
        d <- 1 ## "norm" id

        ## check mu argument
        if (is.numeric(theta$mu) && length(theta$mu) == 1)
            d <- c(d, theta$mu)
        else 
            stop("norm marginal: mean should be a scalar")
        ## check s2 argument
        if (is.numeric(theta$s2) && length(theta$s2) == 1 && theta$s2 > 0)
            d <- c(d, theta$s2)
        else 
            stop("norm marginal: variance should be a positive scalar")
 
        ## check the mu.s2q RW proposal argument
        if (is.numeric(marg.mh$mu.s2q) && length(marg.mh$mu.s2q) == 1
            && marg.mh$mu.s2q > 0) {
            d <- c(d, marg.mh$mu.s2q)
        } else 
            stop("norm marginal: Metropolis-Hastings parameter for mean should be a positive scalar")
        ## check the s2.prop RW proposal argument
        if (is.numeric(marg.mh$s2.prop) && length(marg.mh$s2.prop) == 1
            && marg.mh$s2.prop > 0 && marg.mh$s2.prop > 0) {
            d <- c(d, marg.mh$s2.prop)
        } else 
            stop("norm marginal: Metropolis-Hastings parameter for variance should be a scale in (0, 1)")
    } else if (marginal == "t") {
        d <- 2 ## "t" id

        ## check parameters
        if (is.numeric(theta$mu) && length(theta$mu) == 1)
            d <- c(d, theta$mu)
        else 
            stop("t marginal: 'mean' should be a scalar")
        if (is.numeric(theta$s2) && length(theta$s2) == 1 && theta$s2 > 0)
            d <- c(d, theta$s2)
        else 
            stop("t marginal: 'variance' should be a positive scalar")
        if (is.numeric(theta$nu) && length(theta$nu) == 1 && theta$nu >= 1)
            d <- c(d, theta$nu)
        else 
            stop("t marginal: degrees of freedom should be >= 1")
 
        ## check the RW proposal arguments
        if (is.numeric(marg.mh$mu.s2q) && length(marg.mh$mu.s2q) == 1
            && marg.mh$mu.s2q > 0) {
            d <- c(d, marg.mh$mu.s2q)
        } else 
            stop("t marginal: Metropolis-Hastings parameter for 'mean' should be a positive scalar")
        if (is.numeric(marg.mh$s2.prop) && length(marg.mh$s2.prop) == 1
            && marg.mh$s2.prop > 0 && marg.mh$s2.prop <= 1) {
            d <- c(d, marg.mh$s2.prop)
        } else 
            stop("t marginal: Metropolis-Hastings parameter for 'variance' should be a scale in (0, 1)")
        if (is.numeric(marg.mh$nu.prop) && length(marg.mh$nu.prop) == 1
            && marg.mh$nu.prop > 0 && marg.mh$nu.prop <= 1) {
            d <- c(d, marg.mh$nu.prop)
        } else 
            stop("t marginal: Metropolis-Hastings parameter for degrees of freedom should be a scale in (0, 1)")
    } else if (marginal == "gamma") {
        d <- 3  ## "gamma" id
  
        ## check arguments
        if (is.numeric(theta$a) && length(theta$a) == 1 && theta$a > 0)
            d <- c(d, theta$a)
        else 
            stop("gamma marginal: first parameter should be a positive scalar")
        if (is.numeric(theta$b) && length(theta$b) == 1 && theta$b > 0)
            d <- c(d, theta$b)
        else 
            stop("gamma marginal: second parameter should be a positive scalar")
  
        ## check RW proposal arguments
        if (is.numeric(marg.mh$a.win) && length(marg.mh$a.win) == 1 && marg.mh$a.win > 0)
            d <- c(d, marg.mh$a.win)
        else 
            stop("gamma marginal: Metropolis-Hastings parameters should be positive scalars")
        if (is.numeric(marg.mh$b.win) && length(marg.mh$b.win) == 1 && marg.mh$b.win > 0)
            d <- c(d, marg.mh$b.win)
        else 
            stop("gamma marginal: Metropolis-Hastings parameters should be positive scalars")
          
        ## check the hyperpriors
        if (is.numeric(marg.priors$a.a) && length(marg.priors$a.a) == 1 && marg.priors$a.a > 0) 
            d <- c(d, marg.priors$a.a)
        else 
           stop("gamma marginal: hyperprior parameters should be positive scalars")
        if (is.numeric(marg.priors$a.b) && length(marg.priors$a.b) == 1 && marg.priors$a.b > 0) 
            d <- c(d, marg.priors$a.b)
        else 
           stop("gamma marginal: hyperprior parameters should be positive scalars")
        if (is.numeric(marg.priors$b.a) && length(marg.priors$b.a) == 1 && marg.priors$b.a > 0) 
            d <- c(d, marg.priors$b.a)
        else 
           stop("gamma marginal: hyperprior parameters should be positive scalars")
        if (is.numeric(marg.priors$b.b) && length(marg.priors$b.b) == 1 && marg.priors$b.b > 0) 
            d <- c(d, marg.priors$b.b)
        else 
           stop("gamma marginal: hyperprior parameters should be positive scalars")
    } else if (marginal == "exponential") {
        d <- 4 ## "exponential" id

        ## check argument
        if (is.numeric(theta$a) && length(theta$a) == 1 && theta$a > 0)
           d <- c(d, theta$a)
        else
            stop("exponential marginal: parameter should be a positive scalar")

        ## check the RW proposal argument
        if (is.numeric(marg.mh$a.win) && length(marg.mh$a.win) == 1 && marg.mh$a.win > 0)
            d <- c(d, marg.mh$a.win)
        else 
            stop("exponential marginal: Metropolis-Hastings parameters should be a positive scalar")

        ## check the hyperpriors
        if (is.numeric(marg.priors$a.a) && length(marg.priors$a.a) == 1 && marg.priors$a.a > 0)
            d <- c(d, marg.priors$a.a)
        else 
            stop("exponential marginal: hyperprior parameters should be positive scalars")
        if (is.numeric(marg.priors$a.b) && length(marg.priors$a.b) == 1 && marg.priors$a.b > 0)
            d <- c(d, marg.priors$a.b)
        else 
            stop("exponential marginal: hyperprior parameters should be positive scalars")
    } else if (marginal == "empirical") {
        d <- 5  ## "empirical" id
  
        ## check length of argument
        size <- length(theta$levels)      
        if (size < 3)
            stop("Size of empirical parameter space should be greater than 2! Check your data for discreteness.")
        if (length(theta$mass) != size)
            stop("Size of empirical parameter space not consistent")
        ## check ordering of argument
        for (i in 2:size) {
            if (theta$levels[i] < theta$levels[i - 1])
                stop("empirical marginal: Levels in empirical quantile space are not ordered") 
            if (theta$mass[i] < theta$mass[i - 1])
                stop("empirical marginal: Levels in empirical mass space are not ordered") 
        }
        if (theta$mass[1] != 0 || theta$mass[size] != 1)
            stop("empirical marginal: Mass vector should start at 0 and end at 1!")
        ## check [0, 1] constraints in the mass argument
        for (i in 1:size)
            if (theta$mass[i] < 0 || theta$mass[i] > 1)
                stop("empirical marginal: Levels in empirical mass space are in [0, 1]") 
        d <- c(d, size - 2, theta$levels, theta$mass)
    } else 
        ## otherwise a bad type was given
        stop(paste("copula type", copula, "unknown"))

    ## return numerical encoding
    return(d)
}

## encode.copula:
##
## Checks the initialization for a specified copula and encodes it as a 
## numerical vector.
##
## "copula" should be a string, while "cop.priors" is a list containing the
## prior hyperparameters, and "cop.mh" contains the Metropolis-Hastings 
## parameters.
##
## The initialization specifies the type of copula, and an initial
## value for the copula parameter ("theta")

encode.copula <- function(copula, theta, cop.priors, cop.mh)
{
   # check that cop.priors and cop.mh should be lists
   if(!is.numeric(theta) || !length(theta) == 1 ||
      !is.list(cop.priors) || !is.list(cop.mh)) 
       stop("build.copula: parameter should be a scalar and other inputs should be lists")

   # for gaussian copula
   if (copula == "norm") {
       d <- 1  ## for type

       ## check theta argument
       if (theta > -1 && theta < 1)
           d <- c(d, theta)
       else
           stop("norm copula: parameter should be a scalar in (-1,1)")

      ## notice: no prior hyperparameters (currently)
      
      ## check the RW proposal argument
      if (length(cop.mh$win) == 1 && is.numeric(cop.mh$win) && 
          cop.mh$win > 0 && cop.mh$win <= 1)
          d <- c(d, cop.mh$win)
      else
          stop("norm copula: Metropolis-Hastings parameter should be a scalar in (0,1]!")
   } else if (copula == "fgm") {
      #Farlie-Gumbel-Morgenstern copula
      d <- 2      
      
      ## check theta argument
      if (theta >= -1 && theta <= 1)
          d <- c(d, theta)
      else
          stop("fgm copula: parameter should be a scalar in [-1,1]");

      ## notice: no prior hyperparameters (currently)
      
      ## check the RW proposal argument
      if (length(cop.mh$win) == 1 && is.numeric(cop.mh$win) && 
          cop.mh$win > 0 && cop.mh$win <= 1)
         d <- c(d, cop.mh$win)
      else
         stop("fgm copula: Metropolis-Hastings parameter should be a scalar in (0,1]!")
   } else if (copula == "amh") {
      #Ali-Mikhail-Haq copula
      d <- 3

      ## check theta argument
      if (theta >= -1 && theta <= 1)
          d <- c(d, theta)
      else
          stop("amh copula: parameter should be a scalar in [-1,1]");

      ## notice: no prior hyperparameters (currently)
      
      ## check the RW proposal argument
      if (length(cop.mh$win) == 1 && is.numeric(cop.mh$win) && 
          cop.mh$win > 0 && cop.mh$win <= 1)
         d <- c(d, cop.mh$win)
      else
         stop("amh copula: Metropolis-Hastings parameter should be a scalar in (0,1]!")
   } else if (copula == "clayton") {
      # Temporary
      stop("clayton copula: not implemented yet!")
      
      #Clayton copula
      d <- 4

      ## check theta argument
      if (theta >= -1 && theta != 0)
          d <- c(d, theta)
      else
          stop("clayton copula: theta should be a scalar in [-1,Infinite) - {0}")

      ## check prior argument
      if (length(cop.priors) == 2 && is.numeric(cop.priors$mu) && 
          is.numeric(cop.priors$s2) && cop.priors$s2 > 0) {
          d <- c(d, cop.priors$mu)
          d <- c(d, cop.priors$s2)
      }
      else
         stop("clayton copula: prior hyperparameters are invalid!")
      
      ## check the RW proposal argument
      if (length(cop.mh$win) == 1 && is.numeric(cop.mh$win) && 
          cop.mh$win > 0 && cop.mh$win <= 1)
         d <- c(d, cop.mh$win)
      else
         stop("clayton copula: Metropolis-Hastings parameter should be a scalar in (0,1]!")
   } else if (copula == "gb") {
      #Gumbel-Barnett copula
      d <- 5

      if (theta >= 0 && theta <= 1)
         d <- c(d, theta)
      else
         stop("gb copula: parameter should be a scalar in [-1,1]");

      ## notice: no prior hyperparameters (currently)
      
      ## check the RW proposal argument
      if (length(cop.mh$win) == 1 && is.numeric(cop.mh$win) && 
          cop.mh$win > 0 && cop.mh$win <= 1)
         d <- c(d, cop.mh$win)
      else
         stop("gb copula: Metropolis-Hastings parameter should be a scalar in (0,1]!")
   }  else if (copula == "frank") {
      #Frank copula
      d <- 6

      if (theta != 0)
         d <- c(d, theta)
      else
         stop("frank copula: parameter should be a non-zero scalar");

      ## check prior argument
      if (length(cop.priors) == 2 && is.numeric(cop.priors$mu) && 
          is.numeric(cop.priors$s2) && cop.priors$s2 > 0) {
          d <- c(d, cop.priors$mu)
          d <- c(d, cop.priors$s2)
      }
      else
         stop("frank copula: prior hyperparameters are invalid!")
      
      ## check the RW proposal argument
      if (length(cop.mh$win) == 1 && is.numeric(cop.mh$win) && 
          cop.mh$win > 0 && cop.mh$win <= 1)
          d <- c(d, cop.mh$win)
      else
          stop("frank copula: Metropolis-Hastings parameter should be a scalar in (0,1]!")
   } else
      ## otherwise a bad type was given
      stop(paste("copula type", copula, "unknown"))

   ## return numerical encoding
   return(d)
}

## Transform data from matrix x into array form (row-wise).
## This is how data is passed to the C++ function (with NAs replaced
## by zeroes. Another matrix is used to indicate which values are missing.)

stack.data <- function(x)
{
   if (is.null(dim(x)) && length(x) > 0)
      return(x)
      
   stacked.data <- c()
   x.matrix <- as.matrix(x)
   for (i in 1:nrow(x))
        stacked.data <- c(stacked.data, x.matrix[i, ])
   stacked.data <- as.vector(stacked.data)
   for (i in 1:length(stacked.data))
        if (is.na(stacked.data[i]))
            stacked.data[i] <- 0
   return(stacked.data);
}

##Marginal expectation of tree edges
##
##Input matrix tree.data is a sample of edges in the following format: each row
##is a tree encoded by listing its edges. Each edge is takes two consecutive
##columns in tree.data. A column with entry -1 indicates the end of the edge
##list.
##
##This functions returns the average adjacency matrix. That is, each entry
##is the proportion of data points in which that particular edge appears in 
##the sample, starting from sample burn.in.

tree.marg.expectation <- function(nvar, tree.data, burn.in)
{
   output <- matrix(rep(0, nvar * nvar), ncol = n.vars, nrow = nvar)
   n <- nrow(tree.data)
   max.edges <- nvar - 1
   for (i in (burn.in + 1):n) {
       for (j in 1:max.edges) {
           if (tree.data[i, 2 * (j - 1) + 1] == -1)
               break
           pos.1 <- tree.data[i, 2 * (j - 1) + 1] + 1
           pos.2 <- tree.data[i, 2 * (j - 1) + 2] + 1
           output[pos.1, pos.2] <- output[pos.1, pos.2] + 1
           output[pos.2, pos.1] <- output[pos.2, pos.1] + 1
       }
   }
   for (i in 1:nvar)
       for (j in i:nvar) {
           output[i, j] <- output[i, j] / (n - burn.in)
           output[j, i] <- output[j, i] / (n - burn.in)
       }
   return(output)
}

## FUNCTION get.max.params
## Return the maximum 'number of parameters' among a set of marginals
## 'Number of parameters' is given by :
##
##           number of pdf params + proposal params + prior params + 
##           1 (= id of the marginal)

get.max.params <- function(marginals, train = NULL) 
{
   n.marg <- length(marginals)
   m.d <- c()
   for (i in 1:n.marg) {
        marginal <- marginals[i]
        if (marginal == "norm") {
            m.d <- c(m.d, 5)
        } else if (marginal == "t") {
            m.d <- c(m.d, 7)
        } else if (marginal == "gamma") {
            m.d <- c(m.d, 9)
        } else if (marginal == "exponential") {
            m.d <- c(m.d, 5)
        } else if (marginal == "empirical") {
            gp <- get.empirical(train[, i])   
            m.d <- c(m.d, 2 + length(gp[[1]]) + length(gp[[2]]))
        }
   }
   return(max(m.d))
}

###################################### MAIN FUNCTION

## FUNCTION simple.mcmc.treevariate
##
## Sample both trees and parameters within a Dirichlet process mixture. In the
## model, marginals are the same for every mixture distribution, while the
## copula is a infinite mixture of tree copulas. We call this a treevariate
## distribution.
##
## Because this is a memory-intensive procedure, the sampled trees and cluster
## indicators for the DP process are stored into files. The parameters are also
## stored.
##
## Missing data is allowed in the training set. However, if one wants to 
## evaluate predictive log-likelihood in a test set, then the test set must
## not contain any missing data. Notice that the data passed as input should be
## represented in a particular way. See the description of the parameters.
##
## This is a simplified procedure, where the same family of distribution is
## used in each marginal. The same is true for the copula. All marginals are
## initialized at their maximum likelihood estimate (you will need the 
## package MASS).  More flexible uses of the sampler, setting different families
## of marginals for each variable, can be constructed using the following code
## as a basis.
## 
## Arguments:
##
## - fit.marginals: if true, initialize the marginal parameters by their
##                  marginal likelihood estimators
## - marginal: a string ("norm", "t" or "gamma") indicating the family of
##             marginals used by each variable;
## - copula: a string ("norm", "fgm", "amh", "clayton", "gb" or "frank")
## - simulation.size: number of MCMC samples;
## - burn.in: number of initial MCMC samples to skip when calculating 
##            predictive log-likelihood;
## - adjacencies: adjacency matrix for a tree. For initialization purposes;
## - n.vars: number of variables in the domain;
## - n.train: number of training samples;
## - stacked.data.train: training data in an array format. Data should be stored
##                       rowise;
## - miss.pattern: binary array of the same size as stacked.data.train.
##                 miss.pattern[i] == 1 indicates that the respective variable
##                 is observed. miss.pattern[i] == 0 indicates missing data;
## - n.test: number of test samples. If you don't have any test data, just
##           set this to zero;
## - stacked.data.test: test data in an array format. Data should be stored
##                      rowise;
## - a, b: hyperparameters of the Dirichlet process parameter
##         alpha ~ Gamma(a, b);
## - initial.K: initial number of mixture components for the Dirichlet process;
## - train: original training data. The reason why we need both 
##          stacked.data.train and train is that computing stacked.data.train
##          might take a while, and one might want to call this function
##          multiple times (running different short chains to diagnose how well 
##          the proposal is doing).
## - mcmc.options: array of integers specifying some options. 

simple.mcmc.treevariate <- function(fit.marginals, 
                                    marginals, margs.priors, margs.mh,
                                    copula, cop.priors, cop.mh,
                                    simulation.size, burn.in, mcmc.step,
                                    n.vars,
                                    n.train, stacked.data.train,
                                    miss.pattern,
                                    n.test, stacked.data.test,
                                    miss.pattern.test,
                                    a, b, initial.K, 
                                    initial.clustering, initial.trees, initial.params,
                                    mcmc.options, train) 
{
   if (length(mcmc.options) != 7)
       stop("Array of options should be of length 7!")

   if (length(marginals) != n.vars && length(marginals) != 1)
       stop("You should specify marginal types for each variable or a single global marginal.")       
   if (length(marginals) != length(margs.priors))
       stop("You should specify marginal priors for each choice of marginal.")
   if (length(marginals) != length(margs.mh))
       stop("You should specify marginal Metropolis-Hastings parameters for each choice of marginal.")
   
   
   ##SET MARGINALS ############################################################

   m.d <- c()
   n.params <- 0
   d.complete <- complete.set(train)
   if (length(marginals) == 1)
       md.size <- get.max.params(rep(marginals, n.vars), train)
   else
       md.size <- get.max.params(marginals, train)

   for (i in 1:n.vars) {
      if (length(marginals) == 1) {
          marginal    <- marginals
          marg.priors <- margs.priors[[1]]
          marg.mh     <- margs.mh[[1]]
      } else {
          marginal    <- marginals[i]
          marg.priors <- margs.priors[[i]]
          marg.mh     <- margs.mh[[i]]
      }
      theta <- list()
      if (marginal == "norm") {
         #Gaussian marginals         
         if (fit.marginals) {
            theta$mu <- mean(d.complete[,i], na.rm = T)
            theta$s2 <- sqrt(var(d.complete[,i], na.rm = T))
         } else {
            theta$mu <- runif(1, -2, 2)
            theta$s2 <- runif(1, 0.5, 3)
         }
         cat(paste("Marginal #", i, ": normal [ mu = ", theta$mu,
                   ", s2 = ", theta$s2, "]\n"), sep = "")
         n.params <- n.params + 2
      } else if (marginal == "t") {
         #T marginals: use Gaussian fitted values to initialize it
         if (fit.marginals) {
             mle.out <- F.mle.t(d.complete[, i])
             theta$mu <- mle.out$params$mu
             theta$s2 <- mle.out$params$s2
             theta$nu <- mle.out$params$nu
         }
         else {
             theta$mu <- runif(1, -2, 2)
             theta$s2 <- runif(1, 0.5, 3)
             theta$nu <- 10
         }         
         cat(paste("Marginal #", i, ": t [ mu = ", theta$mu,
                   ", s2 = ", theta$s2, ", nu = ", theta$nu, "]\n"), sep = "")
         n.params <- n.params + 3
      } else if (marginal == "gamma") {
         #Gamma marginals
         cat(i, "gamma\n")
         if (sum(d.complete[,i] < 0) > 0)
             stop("Negative values found in the data. Gamma distribution can't be used.")
         if (fit.marginals) {
            mle.out <- F.mle.gamma(d.complete[, i])
            theta$a <- mle.out$params$shape
            theta$b <- mle.out$params$scale
         } else {
            theta$a <- runif(1, 1, 2)
            theta$b <- runif(1, 1, 2)
         }
         cat(paste("Marginal #", i, ": gamma [ shape = ", theta$a,
                   ", scale = ", theta$b, "]\n"), sep = "")
         n.params <- n.params + 2
       } else if (marginal == "exponential") {
         #Exponential marginals
         if (sum(d.complete[,i] < 0) > 0)
             stop("Negative values found in the data. Exponential distribution can't be used.")
         if (fit.marginals) {
             mle.out <- F.mle.exp(d.complete[, i])
             theta$a <- mle.out$lambda
         } else
             theta$a <- runif(1, 1, 2)    
         cat(paste("Marginal #", i, ": exponential [ a = ", theta$a,
                   "]\n"), sep = "")
         n.params <- n.params + 1
      }  else if (marginal == "empirical") {
         #Empirical marginals
         cat(i, "empirical\n")
         if (!fit.marginals || mcmc.options[5] != 1) {
            not.empirical <- marginals != "empirical"
            if (sum(not.empirical) > 0)
                stop("Error: you cannot combine empirical and parametric marginals if such marginals are not fitted and fixed!")
            else {
                warning("All marginals are empirical: forcing use of estimates")
                fit.marginals   <- TRUE
                mcmc.options[5] <- 1
            }
         }
         gp <- get.empirical(d.complete[, i])
         theta$levels <- gp[[1]]
         theta$mass   <- gp[[2]]
         n.params <- n.params + 1 + 2 * length(gp[[1]]);
      }
      new.d <- encode.marginals(marginal, theta, marg.priors, marg.mh)
      m.d <- c(m.d, pad(new.d, md.size))
   }

   ##SET COPULAS ##############################################################
   # All copulas are 1-parameter copulas.
      
   #Argument: type of copula. Use the following keys
   # "norm"    : Gaussian copula
   # "fgm"     : Farlie-Gumbel-Morgenstern copula
   # "amh"     : Ali-Mikhail-Haq copula
   # "clayton" : Clayton copula
   # "gb"      : Gumbel-Barnett copula
   # "frank"   : Frank copula

   cop.d <- c()
   rho.start <- 0
   if (copula == "frank" || copula == "clayton")
       rho.start <- 0.5
   for (i in 1:n.vars)
      for (j in 1:n.vars) {
          new.d <- encode.copula(copula, rho.start, cop.priors, cop.mh)
          cop.d <- c(cop.d, new.d)
          if (j > i)
              n.params <- n.params + 1
      }
   copd.size <- length(new.d) ##number of params + proposal params + 1 (= id)

   cat("Starting to sample [#parameters = ", n.params, "]\n", sep="");

   r <- .C("mcmc_structure_treevariate",
           n = as.integer(n.train),                ##data sample size
           p = as.integer(n.vars),                 ##number of variables
           a = as.double(a),                       ##hyperparameter for the Dirichlet process
           b = as.double(b),                       ##hyperparameter for the Dirichlet process
           data = as.double(stacked.data.train),   ##data
           miss.p = as.integer(miss.pattern),      ##missing data indicators
           ntest = as.integer(n.test),             ##size of test data
           test = as.double(stacked.data.test),    ##test data
           miss.ptst = as.integer(miss.pattern.test),  ##test data missing data indicators
           md = as.double(m.d), md.size = as.integer(md.size),
           copd = as.double(cop.d), copd.size = as.integer(copd.size),
           num.samples = as.integer(simulation.size),  ##MCMC sample size
           burn.in = as.integer(burn.in),           ##Burn-in (used in the predictive log-likelihood)
           mcmc.step = as.integer(mcmc.step),       ##Indicates which points to log in a file
           initial.K = as.integer(initial.K),       ##Initialize chain with this number of components
           initial.clustering = as.integer(initial.clustering),  ##Initialize chain with this number of components
           initial.trees = as.integer(initial.trees),  ##Initial tree for the chain
           initial.params = as.double(initial.params), ##Initial parameters for the chain
           mcmc.options = as.integer(mcmc.options), ##Set options for the procedure
           output = double(3))                      ##Stores predictive log-likelihood, among other things.

   cat("Done...\n")
   return(r$output)
}
