###############################################################################
#                 BAYESIAN TREEVARIATE COPULA LEARNING                        #
###############################################################################

##### General useful functions

## Generates a binary matrix of missing data indicators of the same
## size as x, but as an array. The probability of some entry being assigned
## zero is given by 'pr'. Good for testing the data augmentation sampler.
##
## The result is returned as an array of n by p entries.

generate.random.missing.pattern <- function(x, pr)
{
   n <- nrow(x)
   p <- ncol(x)
   output <- as.integer(runif(n * p, 0, 1) > pr)
   return(output)
}

## as before, but changes the data directly by setting some entries to NA
generate.random.missing.pattern.directly <- function(x, pr)
{
   n <- nrow(x)
   p <- ncol(x)
   output <- x
   for (i in 1:n)
       for (j in 1:p)
           if (runif(1, 0, 1) < pr)
               output[i, j] <- NA
   return(output)
}

##Generates an indicator matrix with '1' indicating recorded variables, and
##'0' indicating NA variables. This matrix is flattened into an array
##representation so it can be used by the C++ code.

recode.missing.pattern <- function(x)
{
   n <- nrow(x)
   p <- ncol(x)
   output <- data.matrix(matrix(as.integer(!is.na(x)), nrow=n, ncol=p, byrow=F))
   return(stack.data(output))
}

## Impute any missing values in a dataset with the corresponding average
## value in the training set. Convenient for comparison purposes, or for a
## quick-and-dirty sampler that does not require data augmentation.

complete.set <- function(dataset)
{
   m <- apply(dataset, 2, mean, na.rm = T)

   p <- ncol(dataset)        
   n <- nrow(dataset)
   for (j in 1:p) {
       nas <- which(is.na(dataset[,j]))
       if (length(nas) > 0)
           dataset[nas, j] <- m[j]
   }
   return(dataset)
}

## Take a complete dataset, make some values missing by using a given
## pattern, and then complete it with the default imputation method

complete.set.2 <- function(dataset, miss.pattern)
{
   n <- nrow(dataset)
   p <- ncol(dataset)
   if (length(miss.pattern) != n * p)
       stop("invalid missing pattern array!")

   for (i in 1:n)
      for (j in 1:p) {
         if (miss.pattern[(i - 1) * p + j] != 1)
            dataset[i, j] <- NA
      }
   cd <- complete.set(dataset)
   return(cd)
}

## Compare dataset against the respective values in imputed.data that were
## marked as missing by miss.pattern. Report the root mean square error.
## If "dataset" is a "n" x "p" data frame, "miss.pattern" should be a vector
## of size n * p, stored rowwise. In this vector, "0" indicates missing data.

compare.sets <- function(dataset, imputed.data, miss.pattern)
{
   n <- nrow(dataset)
   p <- ncol(dataset)
   if (n != nrow(imputed.data) || p != ncol(imputed.data))
      stop("datasets incomparable!")
   if (length(miss.pattern) != n * p)
      stop("invalid missing pattern array!")

   n.values <- 0
   error <- 0
   for (i in 1:n)
      for (j in 1:p) {
         if (miss.pattern[(i - 1) * p + j] != 1) {
            diff <- imputed.data[i, j] - dataset[i, j]
            error <- error + diff * diff
            n.values <- n.values + 1
         }
      }
   return(sqrt(error / n.values))
}

## Given a dataset 'dataset', generate k partitions of it
## and save the respective training and test sets.

generate.cross.validation <- function(dataset, k, file.name)
{
    n <- nrow(dataset)
    perm <- sample(n)
    fold.size <- n %/% k
    for (i in 1:k) {
        start.fold <- fold.size * (i - 1) + 1
        end.fold <- fold.size * i
        idx.train <- setdiff(perm[1:n], perm[start.fold:end.fold])
        idx.test <- perm[start.fold:end.fold]
        train <- dataset[idx.train,]
        test <- dataset[idx.test,]
        write.table(train, file = paste(file.name, ".train.", i, sep=""), row.names = FALSE)
        write.table(test, file = paste(file.name, ".test.", i, sep=""), row.names = FALSE)
    }
}

## Normalize to zero mean and unit variances

normalize.data <- function(train)
{
   m <- mean(train)
   p <- ncol(train)
   sds <- rep(0, p)
   for (i in 1:p) {       
       sd.i <- sd(train[, i])
       train[, i] <- (train[, i] - m[i]) / sd.i
   }
   return(train)
}

## Normalize both train and test set (using train's mean and covariance matrix)

normalize.datas <- function(train, test)
{
   p <- ncol(train)
   for (i in 1:p) {     
       not.na <- which(!is.na(train[, i]))  
       m.i <- mean(train[not.na, i])
       sd.i <- sd(train[not.na, i])
       train[, i] <- (train[, i] - m.i) / sd.i
       test[, i] <- (test[, i] - m.i) / sd.i
   }
   return(list(train = train, test = test))
}

## Return the indices of the columns in 'train' that have at least as 
## many different 'level' values.

discreteness.filter <- function(train, level)
{
   output <- c()
   for (i in 1:ncol(train))
       if (length(unique(train[, i])) >= level)
           output <- c(output, i)
   return(output)
}

## Returns a random tree structure of size nvars * nvars for initialization
## purposes. This structure is stored in a vector.

random.tree.structure <- function(nvars)
{
   m <- matrix(rep(0, nvars * nvars), ncol = nvars, nrow = nvars, byrow = T)
   s <- sample(1:nvars)
   for (i in 2:nvars) {
       if (i == 2)
           v1 <- s[1]
       else
           v1 <- sample(s[1:(i - 1)], 1)
       v2 <- s[i]
       m[v1, v2] <- 1
       m[v2, v1] <- 1
   }
   return(m)
}

## Append an array with zeros to get it to the desired size

pad <- function(x, size) {
   l <- length(x)
   if (l > size)
      stop("Can't pad an array larger than the desired size.")
   return (c(x, rep(0, size - l)))
}

## Returns a list of two items: the ordered unique values of x and the 
## probability mass associated with values less than or equal to x

get.empirical <- function(x)
{
   output <- list()
   sorted.x <- unique(sort(x))
   output[[1]] <- sorted.x
   output[[2]] <- rep(0, length(sorted.x))
   for (i in 1:length(sorted.x)) {
       count <- sum(x <= sorted.x[i])
       output[[2]][i] <- count / length(x)
   }
   output[[2]][1] <- 0;
   return(output)
}

## Pass the data through the empirical cdf.

empirical.cdf <- function(train)
{
   new.data <- train
   for (i in 1:ncol(train)) {
       emp.i <- get.empirical(train[, i])
       if (length(emp.i[[1]]) < 3)
           stop(paste("Not enough different values in column", i))
       min.v <- emp.i[[2]][2]
       max.v <- emp.i[[2]][length(emp.i[[1]]) - 1]
       for (j in 1:nrow(train)) {
           pos <- which(emp.i[[1]] == train[j, i])
           new.data[j, i] <- emp.i[[2]][pos]
           if (new.data[j, i] == 0)
               new.data[j, i] <- min.v / 2
           else if (new.data[j, i] == 1)
               new.data[j, i] <- (max.v + 1) / 2
       }
   }
   return(new.data)
}

## Log-likelihood of point under tree with adjacency matix adjs and 
## respective parameters params (assuming one-parameter copulas here),
## all of the same copula type.
##
## 'params' is an array, not a matrix
##
## IT IS ASSUMED THAT point COMES FROM A DISTRIBUTION WITH U(0, 1)
## MARGINALS!

pred.copula.loglikelihood <- function(adjs, params, point, copula.type)
{
   p <- ncol(adjs)
   if (p != nrow(adjs) || length(params) != p * p) {
       cat("\n")
       stop(paste("\n PREDICTIVE LOG-LIKELIHOOD: Parameter dimensions incompatible!"))
   }
   llik <- 0
   for (i in 1:(p - 1)) {
       for (j in (i + 1):p) {
            if (adjs[i, j] != 1)
                next
            u <- as.double(point[i])
            v <- as.double(point[j])
            theta <- params[(i - 1) * p + j]
            if (copula.type == "norm") {
                llik <- llik + cop.llik.norm(theta, u, v)
            } else if (copula.type == "fgm") {
                llik <- llik + cop.llik.fgm(theta, u, v)
            } else if (copula.type == "amh") {
                llik <- llik + cop.llik.amh(theta, u, v)
            } else if (copula.type == "clayton") {
                llik <- llik + cop.llik.clayton(theta, u, v)
            } else if (copula.type == "gb") {
                llik <- llik + cop.llik.gb(theta, u, v)
            } else if (copula.type == "frank") {
                llik <- llik + cop.llik.frank(theta, u, v)
            } else {
               cat("\n")
               stop(paste("\n PREDICTIVE LOG-LIKELIHOOD: copula of type", copula.type, 
                          "unknown!\n"))
            }
       }
   }
   return(llik)
}

## Given a table of indicators, calculate the proportion of the top v
## and return it.

z.statistics <- function(z.table, v)
{
   nr <- nrow(z.table)
   nc <- ncol(z.table)
   z <- matrix(as.double(as.matrix(z.table)), nrow = nr, byrow = F)
   output <- matrix(rep(0, nr * (v + 1)), nrow = nr)   
   for (i in 1:nr) {
       #cat("row =", i, "\n")
       max.r <- max(z[i, ])
       counts <- rep(0, max.r)
       for (j in 1:max.r)
           counts[j] <- sum(z[i, ] == j)
       s.c <- sort(counts, decreasing = T)
       size <- min(v, max.r)
       output[i, 1:size] <- s.c[1:size] / nc           
       output[i, v + 1] <- sum(s.c) / nc
   }
   return(output)
}
