extern "C"
{
#include "matrix.h"
#include "rhelp.h"
#include "rand.h"
}
#include "copula.h"
#include "copamh.h"
#include "marginal.h"
#include <math.h>
#include <Rmath.h>
#include <stdlib.h>
#include <stdio.h>
#include <assert.h>
#include <string.h>
#include <string>
#include <fstream>
#include "R.h"
using namespace std;

/*
 * CopAmh:
 *
 * basic constructor which sets the correlation parameter and the
 * other prior and RW proposal parameters com from an old(er) version
 */

CopAmh::CopAmh(double rho, const CopAmh* old) : Copula()
{
  /* assign rho */
  assert(rho >= -1.0 && rho <= 1.0);
  this->rho = rho;

  /* assign prior and RW proposal paramters from old */
  this->rho_win = old->rho_win;

  initialize();
}

void CopAmh::initialize()
{
  double aux[] = {-0.259783, -0.248312, -0.236647, -0.224781, -0.212706,
                  -0.200414, -0.187898, -0.175149, -0.162159, -0.148917,
                  -0.135412, -0.121636, -0.107574, -0.093216, -0.078547,
                  -0.063554, -0.048220, -0.032529, -0.016462, 0,
                   0.016879,  0.034198,  0.051984,  0.070264, 0.089070,
                   0.108438,  0.128406,  0.149018,  0.170325, 0.192383,
                   0.215256,  0.239023,  0.263771,  0.289608, 0.316664,
                   0.345102,  0.375132,  0.407037,  0.441229};

  double aux2[] = {-0.95, 0.95, 0.05};

  this->AMH_SPEARMAN = new double[39];
  memcpy(this->AMH_SPEARMAN, &aux, sizeof(double) * 39);
  this->AMH_SPEARMAN_B = new double[3];
  memcpy(this->AMH_SPEARMAN_B, &aux2, sizeof(double) * 3);
  this->AMH_SPEARMAN_SIZE = 39;
}

/*
 * CopAmh:
 *
 * basic constructor which sets the correlation parameter 
 * and the prior and hierarchical RW parameters
 */

CopAmh::CopAmh(double *d) : Copula()
{
  /* assign rho */  
  this->rho = d[0];
  assert(this->rho >= -1.0 && this->rho <= 1.0);

  /* assign win */
  this->rho_win = d[1];
  assert(this->rho_win > 0);

  initialize();
}


/* 
 * ~CopAmh:
 * 
 * destructor
 */

CopAmh::~CopAmh(void)
{
  delete[] this->AMH_SPEARMAN;
  delete[] this->AMH_SPEARMAN_B;
}


/*
 * Cuti:
 *
 * c_u^{-1}(t) for the AMH copula, useful for
 * obtaining random deviates -- u and v should be vectors
 * with uniform random deviates (of length n)
 */

void CopAmh::Cuti(unsigned int n, double *u, double *v) const
{
  for(unsigned int i = 0; i < n; i++) {
      double aux = this->rho * (1 - u[i]);
      double a = this->rho - aux * aux * v[i];
      double b = 1 - this->rho - 2 * aux * (1 - aux) * v[i];
      double c = - (1 - aux) * (1 - aux) * v[i];
      v[i] = -b / (2 * a) + sqrt(b * b - 4 * a * c) / (2 * a);
  }
}


/*
 * LLik:
 *
 * log likelihood for the AMH bivariate copula
 * calculated based on the marginal ranks (u and v)
 * and the parameter rho
 */

double CopAmh::LLik(unsigned int n, double *u, double *v) const
{
  if (this->rho >= 1 || this->rho <= -1)
      return ZERO_LOGLIKELIHOOD;

  double llik = 0.;
  for (unsigned int i = 0; i < n; i++) {
      double aux = this->rho * (1 - u[i]) * (1 - v[i]);
      llik += log(1 - this->rho + 2 * this->rho * u[i] * v[i] - aux + this->rho * aux) -
              3 * log(1 - aux);
  }
  return llik;
}

/**
 * For a single data point
 */

double CopAmh::LLik(double u, double v) const
{
  if (this->rho >= 1 || this->rho <= -1)
      return ZERO_LOGLIKELIHOOD;

  double aux = this->rho * (1 - u) * (1 - v);
  return log(1 - this->rho + 2 * this->rho * u * v - aux + this->rho * aux) -
         3 * log(1 - aux);
}

/*
 * As before, but going through a subset of the data instead of the
 * whole vectors u, v. The subset is indicated by 'idx'. 
 * Variable 'n' now refers to the size of the index vector
 */

double CopAmh::LLik(unsigned int n, double *u, double *v, unsigned int *idx) const
{
  if (this->rho >= 1 || this->rho <= -1)
      return ZERO_LOGLIKELIHOOD;

  double llik = 0.;
  for (unsigned int i = 0; i < n; i++) {
      double aux = this->rho * (1 - u[idx[i]]) * (1 - v[idx[i]]);
      llik += log(1 - this->rho + 2 * this->rho * u[idx[i]] * v[idx[i]] - aux + this->rho * aux) -
              3 * log(1 - aux);
  }
  return llik;
}

/*
 * As before, but going through a subset of the data instead of the
 * whole vectors u, v. The subset is indicated by integer set 'idx'
 * instead of an array.
 */

double CopAmh::LLik(double *u, double *v, set<int> *idx) const
{
  if (this->rho >= 1 || this->rho <= -1)
      return ZERO_LOGLIKELIHOOD;

  double llik = 0.;
  for (set<int>::iterator i = idx->begin(); i != idx->end(); i++) {
      double aux = this->rho * (1 - u[*i]) * (1 - v[*i]);
      llik += log(1 - this->rho + 2 * this->rho * u[*i] * v[*i] - aux + this->rho * aux) -
              3 * log(1 - aux);
  }
  return llik;
}

/*
 * NumParams:
 *
 * return the number of paramters managed by
 * this copula
 */

unsigned int CopAmh::NumParams(void) const
{
  return 1;
}


/*
 * Params:
 *
 * fill the out vector with the parameter values -- the out vector
 * must have the right amount of space pre-allocated
 */

void CopAmh::Params(double *out) const
{
  out[0] = rho;
}


/*
 * RW:
 *
 * Random Walk proposals for this copula -- the
 * which arguments specifies which proposal to augment
 * from 1:NumParams()
 */

Copula* CopAmh::RW(unsigned int which, double *lqfwd, double *lqbak) const
{
  /* choose a type of Randow Walk proposal */
  switch(which) {
    case 1: return RWrho(lqfwd, lqbak);
    default: error("CopAmh RW proposal %d unknown\n", which);
  }
  return NULL;
}


/*
 * RWrho:
 *
 * Randowm Walk proposal for the correlation parameter, rho
 */

Copula* CopAmh::RWrho(double *lqfwd, double *lqbak) const
{
    double rho_prime = unif_propose_bound(this->rho, -1., 1., this->rho_win, lqfwd, lqbak);
    return new CopAmh(rho_prime, this);
}


/*
 * RW (alternative):
 *
 * Random Walk proposals for this copula -- the
 * which arguments specifies which proposal to augment
 * from 1:NumParams(). These proposals are not based on the
 * current value of the parameter. Instead, an external value is
 * supplied. No probability of backward jump is given, since this
 * is usually used under a structural proposal context.
 */

Copula* CopAmh::RW(double start, unsigned int which, double *lqfwd) const
{
  /* choose a type of Randow Walk proposal */
  switch(which) {
    case 1: return RWrho(start, lqfwd);
    default: error("CopAmh RW proposal %d unknown\n", which);
  }
  return NULL;
}


/*
 * RWrho (alternative):
 *
 * Randowm Walk proposal for the correlation parameter, rho. The proposal
 * is with respect to a given point instead of its current value.
 */

Copula* CopAmh::RWrho(double start, double *lqfwd) const
{
  double lqbak;
  double rho_prime = unif_propose_bound(start, -1., 1., this->rho_win, lqfwd, &lqbak);
  return new CopAmh(rho_prime, this);
}

/*
 * LPrior:
 *
 * return the log (joint) prior of all of the parameters
 * -- this is a uniform prior
 */

double CopAmh::LPrior(void) const
{
  return log(0.5);
}

void CopAmh::randomize() 
{
    this->rho = runif(-1, 1);
}

void CopAmh::setParams(double *in)
{
    this->rho = in[0];
}

/**
 * spearmanRho:
 *
 * returns an (approximate) value of Spearman's Rho based on 
 * the copula parameters.
 */

double CopAmh::spearmanRho()
{
   Rprintf("HERE\n");
   int index = (int) ((this->rho + 1) / this->AMH_SPEARMAN_B[2]);
   if (index >= this->AMH_SPEARMAN_SIZE)
       index = this->AMH_SPEARMAN_SIZE - 1;
   return this->AMH_SPEARMAN[index];
   Rprintf("THERE\n");
}

/**
 * getImpliedSpearmanParams:
 *
 * from a given choice of Spearman's rho, return a choice of
 * parameters that (approximately) generates it.
 */

void CopAmh::getImpliedSpearmanParams(double srho, double out[])
{
  assert(srho < 1 && srho > -1);

  Rprintf("HERE\n");
  //Slow implementation: improve on it later
  for (int i = 0; i < this->AMH_SPEARMAN_SIZE; i++) {
      if (srho <= this->AMH_SPEARMAN[i]) {
          out[0] = this->AMH_SPEARMAN_B[0] + i * this->AMH_SPEARMAN_B[2];
          return;
      }
  }
  out[0] = this->AMH_SPEARMAN_B[1];
  Rprintf("THERE\n");
}

/*
 * evalRW:
 *
 * Evaluates the probability of obtaining the current parameter
 * value using the starting point 'value' (without actually
 * generating a new copula object). 
 */

double CopAmh::evalRW(unsigned int which, double value) const
{
  /* choose a type of Randow Walk proposal */
  switch(which) {
    case 1: return RWrhoEval(value);
    default: error("CopAmh RW proposal %d unknown\n", which);
  }
  return 0.;
}

double CopAmh::RWrhoEval(double start) const
{
  double lower = myfmax(-1., start - this->rho_win);
  double upper = myfmin( 1., start + this->rho_win);
  return -log(upper - lower);
}

/**
 * Returns the log-probability of obtaining the given parameter values from
 * the 'randomize()' distribution (used sometimes to generate copula parameters 
 * regardless of the state of other variables).
 *
 * This assumes 'params' are in the domain of 'randomize()'.
 */

double CopAmh::randomizeLL() const
{
  return 0.;
}

double CopAmh::getParam(unsigned int which) const
{
  switch(which) {
    case 1: return this->rho;
    default: error("CopAmh parameter %d unknown\n", which);
  }
  return 0.;
}

Copula* CopAmh::cloneCopula(unsigned int which, double new_param) const 
{
  if (which != 1) {
     error("CopAmh parameter %d unknown\n", which);
     return NULL;
  }
  return new CopAmh(new_param, this);
}

/**
 * Evaluate the log-posterior of the model (up to an additive constant)
 * using the slice sampler associated data (uSlice, vSlice, idxSlice) at 
 * the 'value' parameter.
 */

double CopAmh::evalSlice(unsigned int which, double value) 
{
  if (which != 1) {
     error("CopAmh slice sampler %d unknown\n", which);
     return 0;
  }
  double old_rho = this->rho;
  this->rho = value;
  double eval = LLik(this->uSlice, this->vSlice, this->idxSlice) +
                LPrior();
  this->rho = old_rho;
  return eval;
}

double CopAmh::priorSample() const
{
  return runif(-1, 1);
}
