extern "C"
{
#include "matrix.h"
#include "rhelp.h"
#include "rand.h"
}
#include "copula.h"
#include "copclayton.h"
#include "marginal.h"
#include <math.h>
#include <Rmath.h>
#include <stdlib.h>
#include <stdio.h>
#include <assert.h>
#include <string.h>
#include <string>
#include <fstream>
#include "R.h"
using namespace std;

/*
 * CopClayton:
 *
 * basic constructor which sets the correlation parameter and the
 * other prior and RW proposal parameters com from an old(er) version
 */

CopClayton::CopClayton(double rho, const CopClayton* old) : Copula()
{
  /* assign rho */
  assert(rho >= -1.0 && rho != 0.0);
  this->rho = rho;

  /* assign prior and RW proposal paramters from old */
  this->rho_mu  = old->rho_mu;
  this->rho_s2  = old->rho_s2;
  this->rho_win = old->rho_win;
}


/*
 * CopClayton:
 *
 * basic constructor which sets the correlation parameter 
 * and the prior and hierarchical RW parameters
 */

CopClayton::CopClayton(double *d) : Copula()
{
  /* assign rho */
  this->rho = d[0];
  assert(this->rho >= -1.0 && this->rho != 0.0);

  /* assign prior */
  this->rho_mu = d[1];
  this->rho_s2 = d[2];
  assert(this->rho_s2 > 0);

  /* assign win */
  this->rho_win = d[3];
  assert(this->rho_win > 0);
}


/* 
 * ~CopClayton:
 * 
 * destructor
 */

CopClayton::~CopClayton(void)
{
}


/*
 * Cuti:
 *
 * c_u^{-1}(t) for the copula, useful for
 * obtaining random deviates -- u and v should be vectors
 * with uniform random deviates (of length n)
 */

void CopClayton::Cuti(unsigned int n, double *u, double *v) const
{
  for(unsigned int i = 0; i < n; i++)
      v[i] = pow(1 + pow(u[i], -this->rho) * (pow(v[i], -this->rho / (this->rho + 1)) - 1), 
                 -1 / this->rho);
}


/*
 * LLik:
 *
 * log likelihood for the bivariate copula
 * calculated based on the marginal ranks (u and v)
 * and the parameter rho
 */

double CopClayton::LLik(unsigned int n, double *u, double *v) const
{
  double llik = 0.;
  for (unsigned int i = 0; i < n; i++)
      //llik += log(this->rho + 1) - (this->rho + 1) * log(u[i] * v[i]) -
      //        (2 * this->rho + 1) * log(pow(u[i], -this->rho) + pow(v[i], -this->rho) - 1) / this->rho;
      llik += log(this->rho + 1) + 
              log(pow(u[i] * v[i], -this->rho - 1) * 
                  pow(pow(u[i], -this->rho) + pow(v[i], -this->rho) - 1, -(2 * this->rho + 1) / this->rho));
  return llik;
}

/**
 * For a single data point
 */

double CopClayton::LLik(double u, double v) const
{
    //return log(this->rho + 1) - (this->rho + 1) * log(u * v) -
    //       (2 * this->rho + 1) * log(pow(u, -this->rho) + pow(v, -this->rho) - 1) / this->rho;
    return log(this->rho + 1) + 
           log(pow(u * v, -this->rho - 1) * pow(pow(u, -this->rho) + pow(v, -this->rho) - 1, 
                                                       -(2 * this->rho + 1) / this->rho));
}

/*
 * As before, but going through a subset of the data instead of the
 * whole vectors u, v. The subset is indicated by 'idx'. 
 * Variable 'n' now refers to the size of the index vector
 */

double CopClayton::LLik(unsigned int n, double *u, double *v, unsigned int *idx) const
{
  double llik = 0.;
  for (unsigned int i = 0; i < n; i++)
      //llik += log(this->rho + 1) - (this->rho + 1) * log(u[idx[i]] * v[idx[i]]) -
      //        (2 * this->rho + 1) * log(pow(u[idx[i]], -this->rho) + pow(v[idx[i]], -this->rho) - 1) / this->rho;
        llik += log(this->rho + 1) + 
                log(pow(u[idx[i]] * v[idx[i]], -this->rho - 1) * 
                    pow(pow(u[idx[i]], -this->rho) + pow(v[idx[i]], -this->rho) - 1, -(2 * this->rho + 1) / this->rho));

  return llik;
}

/*
 * As before, but going through a subset of the data instead of the
 * whole vectors u, v. The subset is indicated by integer set 'idx'
 * instead of an array.
 */

double CopClayton::LLik(double *u, double *v, set<int> *idx) const
{
  double llik = 0.;
  for (set<int>::iterator i = idx->begin(); i != idx->end(); i++)
      //llik += log(this->rho + 1) - (this->rho + 1) * log(u[*i] * v[*i]) -
      //        (2 * this->rho + 1) * log(pow(u[*i], -this->rho) + pow(v[*i], -this->rho) - 1) / this->rho;
      llik += log(this->rho + 1) + 
              log(pow(u[*i] * v[*i], -this->rho - 1) *
                  pow(pow(u[*i], -this->rho) + pow(v[*i], -this->rho) - 1, -(2 * this->rho + 1) / this->rho));
  return llik;
}

/*
 * NumParams:
 *
 * return the number of paramters managed by
 * this copula
 */

unsigned int CopClayton::NumParams(void) const
{
  return 1;
}


/*
 * Params:
 *
 * fill the out vector with the parameter values -- the out vector
 * must have the right amount of space pre-allocated
 */

void CopClayton::Params(double *out) const
{
  out[0] = rho;
}


/*
 * RW:
 *
 * Random Walk proposals for this copula -- the
 * which arguments specifies which proposal to augment
 * from 1:NumParams()
 */

Copula* CopClayton::RW(unsigned int which, double *lqfwd, double *lqbak) const
{
  /* choose a type of Randow Walk proposal */
  switch(which) {
  case 1: return RWrho(lqfwd, lqbak);
  default: error("CopClayton RW proposal %d unknown\n", which);
  }
  return NULL;
}


/*
 * RWrho:
 *
 * Randowm Walk proposal for the correlation parameter, rho
 */

Copula* CopClayton::RWrho(double *lqfwd, double *lqbak) const
{
    double rho_prime = unif_propose_bound(this->rho, -1., 10000., this->rho_win, lqfwd, lqbak); //TODO: find a better proposal
    return new CopClayton(rho_prime, this);
}


/*
 * RW (alternative):
 *
 * Random Walk proposals for this copula -- the
 * which arguments specifies which proposal to augment
 * from 1:NumParams(). These proposals are not based on the
 * current value of the parameter. Instead, an external value is
 * supplied. No probability of backward jump is given, since this
 * is usually used under a structural proposal context.
 */

Copula* CopClayton::RW(double start, unsigned int which, double *lqfwd) const
{
  /* choose a type of Randow Walk proposal */
  switch(which) {
    case 1: return RWrho(start, lqfwd);
    default: error("CopClayton RW proposal %d unknown\n", which);
  }
  return NULL;
}


/*
 * RWrho (alternative):
 *
 * Randowm Walk proposal for the correlation parameter, rho. The proposal
 * is with respect to a given point instead of its current value.
 */

Copula* CopClayton::RWrho(double start, double *lqfwd) const
{
  double lqbak;
  double rho_prime = unif_propose_bound(start, -1., 10000., this->rho_win, lqfwd, &lqbak);
  return new CopClayton(rho_prime, this);
}

/*
 * LPrior:
 *
 * return the log (joint) prior of all of the parameters
 */

double CopClayton::LPrior(void) const
{
    return -0.5 * (log(2 * 3.1415) + log(this->rho_s2) + 
                   (this->rho - this->rho_mu) * (this->rho - this->rho_mu) / this->rho_s2)
           - pnorm(-1, this->rho_mu, sqrt(this->rho_s2), 0, 1);
}

/*
 * Samples from the prior
 */

void CopClayton::randomize() 
{
    this->rho = sample_trunc_gaussian_l(this->rho_mu, this->rho_s2, -1);
    assert(this->rho > -1 && this->rho != 0);
}

void CopClayton::setParams(double *in)
{
    this->rho = in[0];
}

/**
 * spearmanRho:
 *
 * returns an (approximate) value of Spearman's Rho based on 
 * the copula parameters.
 */

double CopClayton::spearmanRho()
{
   return this->rho;
   //TODO: return the actual expression that has an arcsin
}

/**
 * getImpliedSpearmanParams:
 *
 * from a given choice of Spearman's rho, return a choice of
 * parameters that (approximately) generates it.
 */

void CopClayton::getImpliedSpearmanParams(double srho, double out[])
{
  //TODO !!!!!!!!!!!
  assert(srho < 1 && srho > -1);
  out[0] = 2 * sin(3.1415 * srho / 6);
}

/*
 * evalRW:
 *
 * Evaluates the probability of obtaining the current parameter
 * value using the starting point 'value' (without actually
 * generating a new copula object). 
 */

double CopClayton::evalRW(unsigned int which, double value) const
{
  /* choose a type of Randow Walk proposal */
  switch(which) {
    case 1: return RWrhoEval(value);
    default: error("CopClayton RW proposal %d unknown\n", which);
  }
  return 0.;
}

double CopClayton::RWrhoEval(double start) const
{
  double lower = myfmax(-1., start - this->rho_win);
  double upper = myfmin( 10000., start + this->rho_win);
  return -log(upper - lower);
}

/**
 * Returns the log-probability of obtaining the given parameter values from
 * the 'randomize()' distribution (used sometimes to generate copula parameters 
 * regardless of the state of other variables).
 *
 * This assumes 'params' are in the domain of 'randomize()'.
 */

double CopClayton::randomizeLL() const
{
  return 0.;
}

double CopClayton::getParam(unsigned int which) const
{
  switch(which) {
    case 1: return this->rho;
    default: error("CopNorm parameter %d unknown\n", which);
  }
  return 0.;
}

Copula* CopClayton::cloneCopula(unsigned int which, double new_param) const 
{
  if (which != 1) {
     error("CopClayton parameter %d unknown\n", which);
     return NULL;
  }
  return new CopClayton(new_param, this);
}

/**
 * Evaluate the log-posterior of the model (up to an additive constant)
 * using the slice sampler associated data (uSlice, vSlice, idxSlice) at 
 * the 'value' parameter.
 */

double CopClayton::evalSlice(unsigned int which, double value) 
{
  if (which != 1) {
     error("CopNorm slice sampler %d unknown\n", which);
     return 0;
  }
  double old_rho = this->rho;
  this->rho = value;
  double eval = LLik(this->uSlice, this->vSlice, this->idxSlice) +
                LPrior();
  this->rho = old_rho;
  return eval;
}

double CopClayton::priorSample() const
{
  return sample_trunc_gaussian_l(this->rho_mu, this->rho_s2, -1);
}
