extern "C"
{
#include "matrix.h"
#include "rhelp.h"
#include "rand.h"
}
#include "copula.h"
#include "copfgm.h"
#include "marginal.h"
#include <math.h>
#include <Rmath.h>
#include <stdlib.h>
#include <stdio.h>
#include <assert.h>
#include <string.h>
#include <string>
#include <fstream>
#include "R.h"
using namespace std;

/*
 * CopFgm:
 *
 * basic constructor which sets the correlation parameter and the
 * other prior and RW proposal parameters com from an old(er) version
 */

CopFgm::CopFgm(double rho, const CopFgm* old) : Copula()
{
  /* assign rho */
  assert(rho >= -1.0 && rho <= 1.0);
  this->rho = rho;

  /* assign prior and RW proposal paramters from old */
  this->rho_win = old->rho_win;
}


/*
 * CopFgm:
 *
 * basic constructor which sets the correlation parameter 
 * and the prior and hierarchical RW parameters
 */

CopFgm::CopFgm(double *d) : Copula()
{
  /* assign rho */
  this->rho = d[0];
  assert(this->rho >= -1.0 && this->rho <= 1.0);

  /* assign win */
  this->rho_win = d[1];
  assert(this->rho_win > 0);
}


/* 
 * ~CopFgm:
 * 
 * destructor
 */

CopFgm::~CopFgm(void)
{
}


/*
 * Cuti:
 *
 * c_u^{-1}(t) for the copula, useful for
 * obtaining random deviates -- u and v should be vectors
 * with uniform random deviates (of length n)
 */

void CopFgm::Cuti(unsigned int n, double *u, double *v) const
{
  for(unsigned int i = 0; i < n; i++) {
      double aux = this->rho * (2 * u[i] - 1);
      double aux2 = sqrt((0.5 - 0.5 / aux) * (0.5 - 0.5 / aux) + v[i] / aux);
      if (aux > 0)
          v[i] = 0.5 + 0.5 / aux + aux2;
      else
          v[i] = 0.5 - 0.5 / aux + aux2; 
  }
}


/*
 * LLik:
 *
 * log likelihood for the bivariate copula
 * calculated based on the marginal ranks (u and v)
 * and the parameter rho
 */

double CopFgm::LLik(unsigned int n, double *u, double *v) const
{
  double llik = 0.;
  for(unsigned int i = 0; i < n; i++)
    llik += log(1 + this->rho * (1 - 2 * u[i]) * (1 - 2 * v[i]));
  return llik;
}

/**
 * For a single data point
 */

double CopFgm::LLik(double u, double v) const
{
    return log(1 + this->rho * (1 - 2 * u) * (1 - 2 * v));
}

/*
 * As before, but going through a subset of the data instead of the
 * whole vectors u, v. The subset is indicated by 'idx'. 
 * Variable 'n' now refers to the size of the index vector
 */

double CopFgm::LLik(unsigned int n, double *u, double *v, unsigned int *idx) const
{
  double llik = 0.;
  for (unsigned int i = 0; i < n; i++)
      llik += log(1 + this->rho * (1 - 2 * u[idx[i]]) * (1 - 2 * v[idx[i]]));
  return llik;
}

/*
 * As before, but going through a subset of the data instead of the
 * whole vectors u, v. The subset is indicated by integer set 'idx'
 * instead of an array.
 */

double CopFgm::LLik(double *u, double *v, set<int> *idx) const
{
  double llik = 0.;
  for (set<int>::iterator i = idx->begin(); i != idx->end(); i++)
      llik += log(1 + this->rho * (1 - 2 * u[*i]) * (1 - 2 * v[*i]));
  return llik;
}

/*
 * NumParams:
 *
 * return the number of paramters managed by
 * this copula
 */

unsigned int CopFgm::NumParams(void) const
{
  return 1;
}


/*
 * Params:
 *
 * fill the out vector with the parameter values -- the out vector
 * must have the right amount of space pre-allocated
 */

void CopFgm::Params(double *out) const
{
  out[0] = rho;
}


/*
 * RW:
 *
 * Random Walk proposals for this copula -- the
 * which arguments specifies which proposal to augment
 * from 1:NumParams()
 */

Copula* CopFgm::RW(unsigned int which, double *lqfwd, double *lqbak) const
{
  /* choose a type of Randow Walk proposal */
  switch(which) {
  case 1: return RWrho(lqfwd, lqbak);
  default: error("CopFgm RW proposal %d unknown\n", which);
  }
  return NULL;
}


/*
 * RWrho:
 *
 * Randowm Walk proposal for the correlation parameter, rho
 */

Copula* CopFgm::RWrho(double *lqfwd, double *lqbak) const
{
    double rho_prime = unif_propose_bound(rho, -1., 1., this->rho_win, lqfwd, lqbak);
    return new CopFgm(rho_prime, this);
}


/*
 * RW (alternative):
 *
 * Random Walk proposals for this copula -- the
 * which arguments specifies which proposal to augment
 * from 1:NumParams(). These proposals are not based on the
 * current value of the parameter. Instead, an external value is
 * supplied. No probability of backward jump is given, since this
 * is usually used under a structural proposal context.
 */

Copula* CopFgm::RW(double start, unsigned int which, double *lqfwd) const
{
  /* choose a type of Randow Walk proposal */
  switch(which) {
    case 1: return RWrho(start, lqfwd);
    default: error("CopFgm RW proposal %d unknown\n", which);
  }
  return NULL;
}


/*
 * RWrho (alternative):
 *
 * Randowm Walk proposal for the correlation parameter, rho. The proposal
 * is with respect to a given point instead of its current value.
 */

Copula* CopFgm::RWrho(double start, double *lqfwd) const
{
  double lqbak;
  double rho_prime = unif_propose_bound(start, -1., 1., this->rho_win, lqfwd, &lqbak);
  return new CopFgm(rho_prime, this);
}

/*
 * LPrior:
 *
 * return the log (joint) prior of all of the parameters
 * -- this is a uniform prior
 */

double CopFgm::LPrior(void) const
{
  return log(0.5);
}

void CopFgm::randomize() 
{
    this->rho = runif(-1, 1);
}

void CopFgm::setParams(double *in)
{
    this->rho = in[0];
}

/**
 * spearmanRho:
 *
 * returns an (approximate) value of Spearman's Rho based on 
 * the copula parameters.
 */

double CopFgm::spearmanRho()
{
   return this->rho * 3;
}

/**
 * getImpliedSpearmanParams:
 *
 * from a given choice of Spearman's rho, return a choice of
 * parameters that (approximately) generates it.
 */

void CopFgm::getImpliedSpearmanParams(double srho, double out[])
{
  assert(srho < 1 && srho > -1);
  out[0] = srho / 3;
}

/*
 * evalRW:
 *
 * Evaluates the probability of obtaining the current parameter
 * value using the starting point 'value' (without actually
 * generating a new copula object). 
 */

double CopFgm::evalRW(unsigned int which, double value) const
{
  /* choose a type of Randow Walk proposal */
  switch(which) {
    case 1: return RWrhoEval(value);
    default: error("CopFgm RW proposal %d unknown\n", which);
  }
  return 0.;
}

double CopFgm::RWrhoEval(double start) const
{
  double lower = myfmax(-1., start - this->rho_win);
  double upper = myfmin( 1., start + this->rho_win);
  return -log(upper - lower);
}

/**
 * Returns the log-probability of obtaining the given parameter values from
 * the 'randomize()' distribution (used sometimes to generate copula parameters 
 * regardless of the state of other variables).
 *
 * This assumes 'params' are in the domain of 'randomize()'.
 */

double CopFgm::randomizeLL() const
{
  return 0.;
}

double CopFgm::getParam(unsigned int which) const
{
  switch(which) {
    case 1: return this->rho;
    default: error("CopFgm parameter %d unknown\n", which);
  }
  return 0.;
}

Copula* CopFgm::cloneCopula(unsigned int which, double new_param) const 
{
  if (which != 1) {
     error("CopFgm parameter %d unknown\n", which);
     return NULL;
  }
  return new CopFgm(new_param, this);
}

/**
 * Evaluate the log-posterior of the model (up to an additive constant)
 * using the slice sampler associated data (uSlice, vSlice, idxSlice) at 
 * the 'value' parameter.
 */

double CopFgm::evalSlice(unsigned int which, double value) 
{
  if (which != 1) {
     error("CopFgm slice sampler %d unknown\n", which);
     return 0;
  }
  double old_rho = this->rho;
  this->rho = value;
  double eval = LLik(this->uSlice, this->vSlice, this->idxSlice) +
                LPrior();
  this->rho = old_rho;
  return eval;
}

double CopFgm::priorSample() const
{
  return runif(-1, 1);
}
