extern "C"
{
#include "matrix.h"
#include "rhelp.h"
#include "rand.h"
}
#include "copula.h"
#include "copnorm.h"
#include "marginal.h"
#include <math.h>
#include <Rmath.h>
#include <stdlib.h>
#include <stdio.h>
#include <assert.h>
#include <string.h>
#include <string>
#include <fstream>
#include <limits>
#include "R.h"
using namespace std;


/*
 * CopNorm:
 *
 * basic constructor which sets the correlation parameter and the
 * other prior and RW proposal parameters com from an old(er) version
 */

CopNorm::CopNorm(double rho, const CopNorm* old) : Copula()
{
  /* assign rho */
  assert(rho >= -1.0 && rho <= 1.0);
  this->rho = rho;

  /* assign prior and RW proposal parameters from old */
  this->rho_win = old->rho_win;
}


/*
 * CopNorm:
 *
 * basic constructor which sets the correlation parameter 
 * and the prior and hierarchical RW parameters
 */

CopNorm::CopNorm(double *d) : Copula()
{
  /* assign rho */
  this->rho = d[0];
  assert(this->rho >= -1.0 && this->rho <= 1.0);

  /* assign win */
  this->rho_win = d[1];
  assert(this->rho_win > 0 && this->rho_win <= 1.0);
}


/* 
 * ~Normal:
 * 
 * destructor
 */

CopNorm::~CopNorm(void)
{
}


/*
 * Cuti:
 *
 * c_u^{-1}(t) for the Gaussian copula, useful for
 * obtaining random deviates -- u and v should be vectors
 * with uniform random deviates (of length n)
 */

void CopNorm::Cuti(unsigned int n, double *u, double *v) const
{
  for (unsigned int i = 0; i < n; i++) {
       double qv = qnorm(v[i], 0.0, 1.0, 1, 0);
       double qu = qnorm(u[i], 0.0, 1.0, 1, 0);
       double for_p = sqrt(1 - rho * rho) * qv + rho * qu;
       v[i] = pnorm(for_p, 0.0, 1.0, 1, 0);
  }
}


/*
 * LLik:
 *
 * log likelihood for the normal bivariate copula
 * calculated based on the marginal ranks (u and v)
 * and the correllation parameter rho
 */

double CopNorm::LLik(unsigned int n, double *u, double *v) const
{
  if (this->rho >= 1 || this->rho <= -1)
      return ZERO_LOGLIKELIHOOD;

  /* calculate sufficient statistics */
  double squ2, sqv2, squv;
  squ2 = sqv2 = squv = 0;
  for (unsigned int i = 0; i < n; i++) {

    /* A <- sum(qnorm(u)^a2) */
    double qu = qnorm(u[i], 0.0, 1.0, 1, 0);
    squ2 += qu*qu;

    /* B <- sum(qnorm(v)^2) */
    double qv = qnorm(v[i], 0.0, 1.0, 1, 0);
    sqv2 += qv*qv;

    /* C <- sum(qnorm(u)*qnorm(v)) */
    squv += qu*qv;

  }

  /*  evaluate likelihood */
  double rho2 = rho*rho;
  double llik = -0.5*n * log(1-rho2);  
  llik -= (rho2 * (squ2 + sqv2) - 2 * rho * squv) / (2 * (1 - rho2));

  /* return the likelihood */
  return llik;
}

/**
 * For a single data point
 */

double CopNorm::LLik(double u, double v) const
{
  if (this->rho >= 1 || this->rho <= -1)
      return ZERO_LOGLIKELIHOOD;

  /* calculate sufficient statistics */
  double squ2, sqv2, squv;
  squ2 = sqv2 = squv = 0;

  /* A <- sum(qnorm(u)^a2) */
  double qu = qnorm(u, 0.0, 1.0, 1, 0);
  squ2 += qu*qu;

  /* B <- sum(qnorm(v)^2) */
  double qv = qnorm(v, 0.0, 1.0, 1, 0);
  sqv2 += qv*qv;

  /* C <- sum(qnorm(u)*qnorm(v)) */
  squv += qu*qv;

  /*  evaluate likelihood */
  double rho2 = rho*rho;
  double llik = -0.5 * log(1-rho2);
  llik -= (rho2*(squ2+sqv2) - 2*rho*squv)/(2*(1-rho2));

  /* return the likelihood */
  return llik;
}

/*
 * As before, but going through a subset of the data instead of the
 * whole vectors u, v. The subset is indicated by 'idx'. 
 * Variable 'n' now refers to the size of the index vector
 */

double CopNorm::LLik(unsigned int n, double *u, double *v, unsigned int *idx) const
{
  if (this->rho >= 1 || this->rho <= -1)
      return ZERO_LOGLIKELIHOOD;

  /* calculate sufficient statistics */
  double squ2, sqv2, squv;
  squ2 = sqv2 = squv = 0;
  for(unsigned int i=0; i<n; i++) {

    /* A <- sum(qnorm(u)^a2) */
    double qu = qnorm(u[idx[i]], 0.0, 1.0, 1, 0);
    squ2 += qu*qu;

    /* B <- sum(qnorm(v)^2) */
    double qv = qnorm(v[idx[i]], 0.0, 1.0, 1, 0);
    sqv2 += qv*qv;

    /* C <- sum(qnorm(u)*qnorm(v)) */
    squv += qu*qv;

  }

  /*  evaluate likelihood */
  double rho2 = rho*rho;
  double llik = -0.5*n * log(1-rho2);
  llik -= (rho2*(squ2+sqv2) - 2*rho*squv)/(2*(1-rho2));

  /* return the likelihood */
  return llik;
}

/*
 * As before, but going through a subset of the data instead of the
 * whole vectors u, v. The subset is indicated by integer set 'idx'
 * instead of an array.
 */

double CopNorm::LLik(double *u, double *v, set<int> *idx) const
{  
  if (this->rho >= 1 || this->rho <= -1)
      return ZERO_LOGLIKELIHOOD;

  /* calculate sufficient statistics */
  double squ2, sqv2, squv;
  squ2 = sqv2 = squv = 0;
  
  for (set<int>::iterator i = idx->begin(); i != idx->end(); i++) {

    /* A <- sum(qnorm(u)^a2) */
    double qu = qnorm(u[*i], 0.0, 1.0, 1, 0);
    squ2 += qu*qu;

    /* B <- sum(qnorm(v)^2) */
    double qv = qnorm(v[*i], 0.0, 1.0, 1, 0);
    sqv2 += qv*qv;

    /* C <- sum(qnorm(u)*qnorm(v)) */
    squv += qu*qv;

  }

  /*  evaluate likelihood */
  double rho2 = rho*rho;
  double llik = -0.5*idx->size() * log(1-rho2);
  llik -= (rho2*(squ2+sqv2) - 2*rho*squv)/(2*(1-rho2));

  /* return the likelihood */
  return llik;
}

/*
 * NumParams:
 *
 * return the number of paramters managed by
 * this copula
 */

unsigned int CopNorm::NumParams(void) const
{
  return 1;
}


/*
 * Params:
 *
 * fill the out vector with the parameter values -- the out vector
 * must have the right amount of space pre-allocated
 */

void CopNorm::Params(double *out) const
{
  out[0] = rho;
}


/*
 * RW:
 *
 * Random Walk proposals for this copula -- the
 * which arguments specifies which proposal to augment
 * from 1:NumParams()
 */

Copula* CopNorm::RW(unsigned int which, double *lqfwd, double *lqbak) const
{
  /* choose a type of Randow Walk proposal */
  switch(which) {
    case 1: return RWrho(lqfwd, lqbak);
    default: error("CopNorm RW proposal %d unknown\n", which);
  }
  return NULL;
}


/*
 * RWrho:
 *
 * Randowm Walk proposal for the correlation parameter, rho
 */

Copula* CopNorm::RWrho(double *lqfwd, double *lqbak) const
{
  double rho_prime = unif_propose_bound(this->rho, -1., 1., this->rho_win, lqfwd, lqbak);
  return new CopNorm(rho_prime, this);
}

/*
 * RW (alternative):
 *
 * Random Walk proposals for this copula -- the
 * which arguments specifies which proposal to augment
 * from 1:NumParams(). These proposals are not based on the
 * current value of the parameter. Instead, an external value is
 * supplied. No probability of backward jump is given, since this
 * is usually used under a structural proposal context.
 */

Copula* CopNorm::RW(double start, unsigned int which, double *lqfwd) const
{
  /* choose a type of Randow Walk proposal */
  switch(which) {
    case 1: return RWrho(start, lqfwd);
    default: error("CopNorm RW proposal %d unknown\n", which);
  }
  return NULL;
}


/*
 * RWrho (alternative):
 *
 * Randowm Walk proposal for the correlation parameter, rho. The proposal
 * is with respect to a given point instead of its current value.
 */

Copula* CopNorm::RWrho(double start, double *lqfwd) const
{
  double lqbak;
  //double rho_prime = unif_propose_bound(start, -1., 1., this->rho_win, lqfwd, &lqbak); //UNIFORM
  double rho_prime = truncgaussian_propose_bound(start, -1., 1., this->rho_win, lqfwd, &lqbak); //TRUNCATED GAUSSIAN
  return new CopNorm(rho_prime, this);
}

/*Copula* CopNorm::RWrho(double start, double *lqfwd) const
{
  //
    2 * BETA(alpha, beta) - 1 (a beta rescaled to [-1, 1]). Here we have
      -- [2 * alpha / (alpha + beta) - 1] == mean(start) (that is, we set the mean as a function of 'start')
      -- 4 * alpha * beta  / [(alpha + beta)^2 * (alpha + beta + 1)] = this->rho_win (that is, the variance is given by this->rho_win)
  //
  double mean = 0.5 * (start + 1);
  double t = mean / (1 - mean);
  double t_1 = t + 1;
  double beta = (t / (this->rho_win * t_1 * t_1) - 1) / t_1;
  if (beta < 1.001)
      beta = 1.001;
  double alpha = (beta * mean - 2 * mean + 1) / (1 - mean);
  if (alpha < 1.001)
      alpha = 1.001;
  //double alpha = beta * t;
  double rho_prime_beta = rbeta(alpha, beta);
  *lqfwd = dbeta(rho_prime_beta, alpha, beta, 1);
  double rho_prime = 2 * rho_prime_beta - 1;
  //Rprintf("mean = %f, alpha = %f, beta = %f, rho_prime = %f\n", mean, alpha, beta, rho_prime);
  //double a_b = alpha + beta;
  //Rprintf("   alpha / (alpha + beta) = %f, impl_var = %f\n", alpha / a_b, alpha * beta / (a_b * a_b * (a_b + 1)));
  return new CopNorm(rho_prime, this);
}*/

/*
 * LPrior:
 *
 * return the log (joint) prior of all of the parameters
 * -- this is a uniform prior
 */

double CopNorm::LPrior(void) const
{
  return log(0.5);
}

void CopNorm::randomize() 
{
    this->rho = runif(-1, 1);
}

void CopNorm::setParams(double *in)
{
    this->rho = in[0];
}

/**
 * spearmanRho:
 *
 * returns an (approximate) value of Spearman's Rho based on 
 * the copula parameters.
 */

double CopNorm::spearmanRho()
{
   return (6. / 3.141593) * asin(0.5 * this->rho);
}

/**
 * getImpliedSpearmanParams:
 *
 * from a given choice of Spearman's rho, return a choice of
 * parameters that (approximately) generates it.
 */

void CopNorm::getImpliedSpearmanParams(double srho, double out[])
{
  if (!(srho < 1 && srho > -1))
      Rprintf("srho = %f\n", srho);
  assert(srho <= 1 && srho >= -1);
  out[0] = 2 * sin((3.141593 / 6) * srho);
}


/*
 * evalRW:
 *
 * Evaluates the probability of obtaining the current parameter
 * value using the starting point 'value' (without actually
 * generating a new copula object). 
 */

double CopNorm::evalRW(unsigned int which, double value) const
{
  /* choose a type of Randow Walk proposal */
  switch(which) {
    case 1: return RWrhoEval(value);
    default: error("CopNorm RW proposal %d unknown\n", which);
  }
  return 0.;
}

/* OLD ONE BASED ON UNIFORM PROPOSALS  
double CopNorm::RWrhoEval(double start) const
{
  double lower = myfmax(-1., start - this->rho_win);
  double upper = myfmin( 1., start + this->rho_win);
  if (this->rho >= lower && this->rho <= upper)
      return -log(upper - lower);
  else
      return -std::numeric_limits<double>::max(); // ``Log(0) == Negative infinity''
}*/

/* OLD ONE BASED ON TRUNCATED GAUSSIANS
 *
 * This evaluation is based on the [-1, 1] truncated Gaussian proposal and
 * DOES NOT include the full normalizing constant of the truncated Gaussian
 * (which is an additive term in thie log scale, as used here). Since the
 * main application of this is to compute ratios of proposals and since
 * the constant is also not used in 'truncgaussian_propose_bound' 
 * (see CopNorm::RWrho), it shouldn't be a problem.
 */

double CopNorm::RWrhoEval(double start) const
{
  double diff = this->rho - start;
  double sq_win = sqrt(this->rho_win);
  double z = log(pnorm(1, start, sq_win, 1, 0) - pnorm(-1, start, sq_win, 1, 0)); //log-normalizing constant (mean-dependent terms);
  return -0.5 * diff * diff / this->rho_win - z - 0.5 * log(this->rho_win);
}

/**
 * This evaluation is based on a 2 * X - 1 distribution, where X is 
 * taken as a beta(alpha, beta) random variable.
 */

/*double CopNorm::RWrhoEval(double start) const
{
  double mean = 0.5 * (this->rho + 1);
  double t = mean / (1 - mean);
  double t_1 = t + 1;
  double beta = (t / (this->rho_win * t_1 * t_1) - 1) / t_1;
  if (beta < 1.001)
      beta = 1.001;
  double alpha = (beta * mean - 2 * mean + 1) / (1 - mean);
  if (alpha < 1.001)
      alpha = 1.001;
  //double alpha = beta * t;
  double start_beta = 0.5 * (start + 1);
  return dbeta(start_beta, alpha, beta, 1);
}*/

/**
 * Returns the log-probability of obtaining the given parameter values from
 * the 'randomize()' distribution (used sometimes to generate copula parameters 
 * regardless of the state of other variables).
 *
 * This assumes 'params' are in the domain of 'randomize()'.
 */

double CopNorm::randomizeLL() const
{
  return 0.;
}

double CopNorm::getParam(unsigned int which) const
{
  switch(which) {
    case 1: return this->rho;
    default: error("CopNorm parameter %d unknown\n", which);
  }
  return 0.;
}

Copula* CopNorm::cloneCopula(unsigned int which, double new_param) const 
{
  if (which != 1) {
     error("CopNorm parameter %d unknown\n", which);
     return NULL;
  }
  return new CopNorm(new_param, this);
}

/**
 * Evaluate the log-posterior of the model (up to an additive constant)
 * using the slice sampler associated data (uSlice, vSlice, idxSlice) at 
 * the 'value' parameter.
 */

double CopNorm::evalSlice(unsigned int which, double value) 
{
  if (which != 1) {
     error("CopNorm slice sampler %d unknown\n", which);
     return 0;
  }
  double old_rho = this->rho;
  this->rho = value;
  double eval = LLik(this->uSlice, this->vSlice, this->idxSlice) +
                LPrior();
  this->rho = old_rho;
  return eval;
}

/**
 * Returns a sample from the prior
 */

double CopNorm::priorSample() const
{
  return runif(-1, 1);
}
