extern "C"
{
#include "matrix.h"
#include "rhelp.h"
}
#include "copula.h"
#include <R.h>
#include <Rmath.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <math.h>
#include <fstream>
using namespace std;


/*
 * Copula:
 *
 * a (for now) mostly empty constructor
 */

Copula::Copula(void)
{
  this->mSlice = 3;
}


/*
 * ~Copula:
 * 
 * the usual destructor function 
 */

Copula::~Copula(void)
{
}


/*
 * LLik:
 *
 * Joint likelihood of the copula based on the two passed
 * in marginals -- right now both marginals have to have 
 * the same length data (n), this will change in the future
 */

double Copula::LLik(Marginal *Mx, Marginal *My) const
{
  assert(Mx->N() == My->N());

  /* likelihood of the bivariate copula */
  double llik = LLik(Mx->N(), Mx->Ranks(), My->Ranks());

  /* return the joint likelihood */
  return llik;
}

double Copula::LLik(Marginal *Mx, Marginal *My, double x, double y) const
{
  assert(Mx->N() == My->N());

  /* likelihood of the bivariate copula */
  double llik = LLik(Mx->CalcRanks(x), My->CalcRanks(y));

  /* return the joint likelihood */
  return llik;
}

double Copula::LLik(Marginal *Mx, Marginal *My, unsigned int d) const
{
  assert(Mx->N() == My->N());

  /* likelihood of the bivariate copula */
  double llik = LLik(Mx->GetRank(d), My->GetRank(d));

  /* return the joint likelihood */
  return llik;
}

/*
 * As before, but using a subset of the data index by
 * array 'idx'.
 */

double Copula::LLik(Marginal *Mx, Marginal *My, unsigned int *idx, unsigned int idx_n) const
{
  assert(Mx->N() >= idx_n);
  assert(My->N() >= idx_n);

  /* likelihood of the bivariate copula */
  double llik = LLik(idx_n, Mx->Ranks(), My->Ranks(), idx);

  /* return the joint likelihood */
  return llik;
}

/*
 * As before, but using a subset of the data index by
 * integer set 'idx'.
 */

double Copula::LLik(Marginal *Mx, Marginal *My, set<int> *idx) const
{
  assert(Mx->N() >= idx->size());
  assert(My->N() >= idx->size());

  /* likelihood of the bivariate copula */
  double llik = LLik(Mx->Ranks(), My->Ranks(), idx);

  /* return the joint likelihood */
  return llik;
}

/*
 * Main slice sampling routines:
 *
 * - which: index of the parameter whose posterior 
 *          we are sampling from (keeping everybody else fixed);
 * - x0:    starting point;
 * - u, v:  array containing the data used in the evaluation of the posterior;
 * - idx:   index of which data points within (u, v) should be used;
 *
 * The procedure returns a new copula where the which-th parameter
 * has been substituted by the sample.
 */

Copula* Copula::sliceSample(unsigned int which, double x0,
                            double *u, double *v, set<int> *idx)
{
  this->uSlice = u; this->vSlice = v; this->idxSlice = idx;
  this->wSlice = getWSlice(which);
  //Rprintf("- Slice width = %d\n", this->wSlice);
  double l, r;
  double e = rexp(1);
  double z = (LLik(u, v, idx) + LPrior()) - e;
  //Rprintf("- Log-heigth z = %f\n", z);
  sliceStepOut(which, x0, z, &l, &r);
  //Rprintf("- Interval = [%f %f]\n", l, r);
  double new_param = sliceShrinkage(which, x0, z, &l, &r);
  //Rprintf("- New param = %f\n", new_param);
  //Rprintf("-----------------------\n");
  Copula* out_copula = cloneCopula(which, new_param);
  return out_copula;
}

Copula* Copula::sliceSample(unsigned int which, 
                            double *u, double *v, set<int> *idx)
{
  double x0 = getParam(which);
  return sliceSample(which, x0, u, v, idx);
}

/*
 * The 'stepping out' procedure of slice sampling (see Figure 3 from Neal, 2003):
 *
 * - which: index of the parameter to be sampled;
 * - x0: starting point;
 * -  z: the vertical level defining the (log) slice;
 * - *l: buffer where we store a copy the sampled left boundary;
 * - *r: buffer where we store a copy the sampled right boundary;
 */

void Copula::sliceStepOut(unsigned int which, double x0, double z, 
                          double *l, double *r) 
{
  /* Initialize */
  double u = runif(0, 1);
  *l = x0 - this->wSlice * u;
  *r = *l + this->wSlice;
  double v = runif(0, 1);
  int j = (int) trunc(this->mSlice * v);
  int k = (this->mSlice - 1) * j;

  /* Choose left boundary */
  double f_l = evalSlice(which, *l);
  while (j > 0 && z < f_l) {
      *l = *l - this->wSlice;
      f_l = evalSlice(which, *l);
      j--;
  }
  
  /* Choose right boundary */
  double f_r = evalSlice(which, *r);
  while (k > 0 && z < f_r) {
      *r = *r + this->wSlice;
      f_r = evalSlice(which, *r);
      k--;
  }
}

/*
 * The 'shrinkage' procedure of slice sampling (see Figure 5 from Neal, 2003):
 *
 * - which: index of the parameter to be sampled;
 * - x0: starting point;
 * -  z: the vertical level defining the (log) slice;
 * - *l: buffer where we store a copy the sampled left boundary;
 * - *r: buffer where we store a copy the sampled right boundary;
 *
 * This assumes the sliceStepOut procedure was used, and hence we do
 * not need to check if x1 is an acceptable point or not.
 *
 * At the end, we return the properly sampled 'which' parameter.
 */

double Copula::sliceShrinkage(unsigned int which, double x0, double z, 
                              double *l, double *r)
{
  do {
      double u = runif(0, 1);
      double x1 = *l + u * (*r - *l);
      if (z < evalSlice(which, x1))
          return x1;
      if (x1 < x0) 
          *l = x1;
      else
          *r = x1;
  } while (true);
}

double Copula::getWSlice(unsigned int which) const
{
    return 0.25; //TODO, MOVE THIS TO SUBCLASSES
}
