extern "C"
{
#include "matrix.h"
#include "rhelp.h"
#include "rand.h"
}

#include <Rmath.h>
#include "graphcore.h"
#include "R.h"
#include <assert.h>
#include <iostream>

/* 
 * This assumes that array adjacencies has num_nodex * num_nodes entries
 * from a matrix of respective size, ordered row-wise.
 */

Graph::Graph(int* adjacencies, unsigned int num_nodes) {
	assert(num_nodes > 0);
	this->num_nodes = num_nodes;
	this->adj_matrix = new int[num_nodes * num_nodes];
    memcpy(this->adj_matrix, adjacencies, sizeof(int) * num_nodes * num_nodes);
	generateAdjacencyInformation();
}

Graph::~Graph() {
	delete[] this->edge_index;
	delete[] this->edges;
    delete[] this->adj_matrix;
}

unsigned int Graph::getAdjacency(unsigned int pos1, unsigned int pos2) {
	assert(pos1 < this->num_nodes && pos2 < this->num_nodes);
	return this->adj_matrix[pos1 * this->num_nodes + pos2];
}

void Graph::generateAdjacencyInformation() {
	this->num_edges = 0;
	this->adj_list.resize(this->num_nodes);
	for (unsigned int i = 0; i < this->num_nodes; i++) {
		int num_neighbors = 0;
		for (unsigned int j = 0; j < this->num_nodes; j++)
			if (this->adj_matrix[i * this->num_nodes + j] != 0) {
				this->num_edges++;
                num_neighbors++;
			}
		this->adj_list[i].resize(num_neighbors);
		num_neighbors = 0;
		for (unsigned int j = 0; j < this->num_nodes; j++)
			if (this->adj_matrix[i * this->num_nodes + j] != 0)
				this->adj_list[i][num_neighbors++] = j;		
	}
	this->num_edges /= 2;
	this->edge_index = new unsigned int[this->num_nodes * this->num_nodes];
	this->edges = new unsigned int[this->num_edges * 2];
	int pos = 0;
	for (unsigned int i = 0; i < this->num_nodes; i++) {
		int nn = this->adj_list[i].size();
		for (int j = 0; j < nn; j++) {
			if (i >= this->adj_list[i][j])
				continue;
			this->edges[pos * 2    ] = i; 
			this->edges[pos * 2 + 1] = this->adj_list[i][j];
			this->edge_index[i * this->num_nodes + this->adj_list[i][j]] = 
			this->edge_index[this->adj_list[i][j] * this->num_nodes + i] = pos;
			pos++;
		}
	}
}



Tree::Tree(int* adjacencies, unsigned int num_nodes) : Graph(adjacencies, num_nodes) {
	generateTreeInformation();
}

/**
 * Verify whether adjacency matrix will generate a valid tree
 */
bool Tree::validAdjacencies(int* adjacencies, unsigned int num_nodes) 
{
    for (unsigned int i = 0; i < num_nodes; i++)
        if (adjacencies[i * num_nodes + i] == 1) {
            Rprintf("Invalid diagonal\n");
            return false;
        }

    for (unsigned int i = 0; i < num_nodes; i++)
        for (unsigned int j = i + 1; j < num_nodes; j++) {
            //Temporarily delete edge (i, j) if it exists
            if (adjacencies[i * num_nodes + j] != 1)
                continue;
            adjacencies[i * num_nodes + j] = adjacencies[j * num_nodes + i] = 0;
            bool con = isConnected(adjacencies, num_nodes, i, j);
            //Restore adjacencies
            adjacencies[i * num_nodes + j] = adjacencies[j * num_nodes + i] = 1;

            if (con) //Not supposed to be connected
                return false;
        }
    return true;
}

Tree::~Tree() {
	delete[] this->order;
	delete[] this->parents;
}

void Tree::generateTreeInformation() {
	this->order = new unsigned int[this->num_nodes];
	this->parents = new int[this->num_nodes];
	bool *marked = new bool[this->num_nodes];
	bool *pre_marked = new bool[this->num_nodes];
    int  level[1];
    unsigned int num_marked = 0;

	this->num_components = 0;
	for (unsigned int i  = 0; i < this->num_nodes; i++)
		marked[i] = false;

	for (unsigned int i = 0; i < this->num_nodes; i++) {
        memcpy(&pre_marked[0], &marked[0], sizeof(bool) * num_nodes);
		if (!marked[i]) {
			this->parents[i] = -1;
			this->order[num_marked] = i;
            marked[i] = true;
            level[0] = num_marked;          
            partialOrderRecursion(i, marked, level);
		}

		int component_size = 0;
		for (unsigned int j = 0; j < this->num_nodes; j++)
            if (pre_marked[j] != marked[j])
                component_size++;
        if (component_size > 0) {
            vector<unsigned int> new_component(component_size);
			this->components.insert(this->components.end(), new_component);
            component_size = 0;
			num_marked = 0;
            for (unsigned int j = 0; j < this->num_nodes; j++) {
                if (pre_marked[j] != marked[j]) {                    
                    components[this->num_components][component_size] = j;
					component_size++;
                }
				if (marked[j])
					num_marked++;
            }
			this->num_components++;
		}
        if (num_marked == this->num_nodes)
            break;
	}

	delete[] marked;
	delete[] pre_marked;
}

void Tree::partialOrderRecursion(int current_node, bool *marked, int *level) {
	int nn = this->adj_list[current_node].size();
    if (nn > 0) for (int i = 0; i < nn; i++) {
         if (!marked[this->adj_list[current_node][i]]) {
             level[0]++;
             this->order[level[0]] = this->adj_list[current_node][i];
             this->parents[this->adj_list[current_node][i]] = current_node;
         }
    }
    if (nn > 0) for (int i = 0; i < nn; i++) {
         if (!marked[this->adj_list[current_node][i]]) {
             marked[this->adj_list[current_node][i]] = true;
             partialOrderRecursion(this->adj_list[current_node][i], marked, level);
         }
    }        
}


void Tree::printTreeInformation() 
{
	cout << "num.nodes = " << this->num_nodes << "\n";
	cout << "num.edges = " << this->num_edges << "\n";
	cout << "num.components = " << this->num_components << "\n";

	cout << "\nCOMPONENTS:" << "\n\n";
	for (unsigned int i = 0; i < this->num_components; i++) {
		cout << "\t" << i << ":";
		for (unsigned int j = 0; j < this->components[i].size(); j++)
		    cout << " " << this->components[i][j];
		cout << "\n";
	}

	cout << "\nPARTIAL ORDER:" << "\n\n";
	for (unsigned int i = 0; i < this->num_nodes; i++)
		cout << "\t" << this->order[i] << "\n";

    cout << "\nPARENTS:" << "\n\n";
	for (unsigned int i = 0; i < this->num_nodes; i++)
		cout << "\t" << i << ": " << this->parents[i] << "\n" ;

	cout << "\nADJACENCY LIST:" << "\n\n";
	for (unsigned int i = 0; i < this->num_nodes; i++) {
		cout << "\t" << i << ":";
		for (unsigned int j = 0; j < this->adj_list[i].size(); j++)
		    cout << " " << this->adj_list[i][j];
		cout << "\n";
	}
}

void Tree::printTreeInformationR() 
{
	Rprintf("num.nodes = %d\n", this->num_nodes);
    Rprintf("num.edges = %d\n", this->num_edges);
    Rprintf("num.components = %d\n", this->num_components);

	Rprintf("\nADJACENCY LIST:\n\n");
	for (unsigned int i = 0; i < this->num_nodes; i++) {
        Rprintf("\t%d:", i);
		for (unsigned int j = 0; j < this->adj_list[i].size(); j++)
		    Rprintf(" %d", this->adj_list[i][j]);
		Rprintf("\n");
	}
}

Tree* Tree::proposeAddEdge(int *end_points, int *tabu) 
{
    if (this->num_edges == this->num_nodes - 1)
        return clone();
    
    int *valid_edges = new int[2 * this->num_nodes * (this->num_nodes - 1)];
    int count_edges = 0;
    for (unsigned int i = 0; i < this->num_components; i++)
        for (unsigned int n1 = 0; n1 < this->components[i].size(); n1++)
            for (unsigned int j = i + 1; j < this->num_components; j++)
                for (unsigned int n2 = 0; n2 < this->components[j].size(); n2++) {
                    if (tabu == NULL || 
                        (!(tabu[0] == components[i][n1] && tabu[1] == components[j][n2]) &&
                         !(tabu[0] == components[j][n2] && tabu[1] == components[i][n1]))) {
                        valid_edges[2 * count_edges]     = components[i][n1];
                        valid_edges[2 * count_edges + 1] = components[j][n2];
                        count_edges++;
                    }
                }
    
    int edge_choice = sample_uniform(count_edges);
    int pos_1 = valid_edges[2 * edge_choice], pos_2 = valid_edges[2 * edge_choice + 1];

    if (end_points != NULL) {
        end_points[0] = pos_1;
        end_points[1] = pos_2;
    }

    this->adj_matrix[pos_1 * this->num_nodes + pos_2] = 
                    this->adj_matrix[pos_2 * this->num_nodes + pos_1] = 1;
    Tree *new_tree = new Tree(this->adj_matrix, this->num_nodes);
    this->adj_matrix[pos_1 * this->num_nodes + pos_2] = 
                    this->adj_matrix[pos_2 * this->num_nodes + pos_1] = 0;

    delete[] valid_edges;
    return(new_tree);
}

Tree* Tree::proposeAddEdge() 
{
    return proposeAddEdge(NULL);
}

Tree* Tree::proposeAddEdge(int *end_points) {
    return proposeAddEdge(end_points, NULL);
}

Tree* Tree::proposeRemoveEdge(int *end_points) 
{
    int edge_choice = sample_uniform(this->num_edges);
    int pos_1 = this->edges[2 * edge_choice], pos_2 = this->edges[2 * edge_choice + 1];

    this->adj_matrix[pos_1 * this->num_nodes + pos_2] = 
                    this->adj_matrix[pos_2 * this->num_nodes + pos_1] = 0;
    Tree *new_tree = new Tree(this->adj_matrix, this->num_nodes);
    this->adj_matrix[pos_1 * this->num_nodes + pos_2] = 
                    this->adj_matrix[pos_2 * this->num_nodes + pos_1] = 1;
    if (end_points != NULL) {
        end_points[0] = pos_1;
        end_points[1] = pos_2;
    }
    return(new_tree);
}


Tree* Tree::proposeRemoveEdge() 
{
    return proposeRemoveEdge(NULL);
}

Tree* Tree::proposeMoveEdge(int *end_points_source, int *end_points_dest)
{
    Tree *thinner_tree = proposeRemoveEdge(end_points_source);
    Tree *new_tree = thinner_tree->proposeAddEdge(end_points_dest, end_points_source);
    delete thinner_tree;
    return(new_tree);
}

Tree* Tree::proposeMoveEdge() 
{
    Tree *thinner_tree = proposeRemoveEdge(NULL);
    Tree *new_tree = thinner_tree->proposeAddEdge(NULL);
    delete thinner_tree;
    return(new_tree);
}

/**
 * Proposes moving an edge from (u, v) to (u, t). Edge (u, v) is chosen uniformly
 * at random. Node t is chosen uniformly at random within the complementary
 * component that does not include u. Triplet (u, v, t) is then stored in buffer
 * array 'three_points'.
 *
 * THIS ASSUMES THE TREE IS CONNECTED!
 */

Tree* Tree::proposeMoveEdgeVertexOld(int *three_points) 
{
    int end_points_source[2];
    Tree *thinner_tree = proposeRemoveEdge(&end_points_source[0]);
    int pos_1, pos_old;
    if (this->adj_list[end_points_source[0]].size() == 1) {
        pos_1 = end_points_source[0];
        pos_old = end_points_source[1];
    } else if (this->adj_list[end_points_source[1]].size() == 1) {
        pos_1 = end_points_source[1];
        pos_old = end_points_source[0];
    } else if (runif(0, 1) < 0.5) {
        pos_1 = end_points_source[0];
        pos_old = end_points_source[1];
    } else {
        pos_1 = end_points_source[1];    
        pos_old = end_points_source[0];
    }
    
    int *valid_edges = new int[2 * thinner_tree->num_nodes * (thinner_tree->num_nodes - 1)];
    int count_edges = 0;

    for (unsigned int i = 0; i < thinner_tree->num_components; i++)
        for (unsigned int n1 = 0; n1 < thinner_tree->components[i].size(); n1++) {
            if (thinner_tree->components[i][n1] != pos_1)
                continue;
            for (unsigned int j = 0; j < thinner_tree->num_components; j++) {
                if (j == i)
                    continue;                
                for (unsigned int n2 = 0; n2 < thinner_tree->components[j].size(); n2++) {
                    if (thinner_tree->components[j][n2] == pos_old)
                        continue;
                    valid_edges[2 * count_edges]     = thinner_tree->components[i][n1];
                    valid_edges[2 * count_edges + 1] = thinner_tree->components[j][n2];
                    count_edges++;
                }
            }
            goto edges_collected;
            break;
        }
edges_collected:
    delete thinner_tree;    

    int edge_choice = sample_uniform(count_edges);
    int pos_2 = valid_edges[2 * edge_choice + 1];

    this->adj_matrix[pos_1 * this->num_nodes + pos_2] = 
                    this->adj_matrix[pos_2 * this->num_nodes + pos_1] = 1;
    this->adj_matrix[pos_1 * this->num_nodes + pos_old] = 
                    this->adj_matrix[pos_old * this->num_nodes + pos_1] = 0;
    Tree *new_tree = new Tree(this->adj_matrix, this->num_nodes);
    this->adj_matrix[pos_1 * this->num_nodes + pos_2] = 
                    this->adj_matrix[pos_2 * this->num_nodes + pos_1] = 0;
    this->adj_matrix[pos_1 * this->num_nodes + pos_old] = 
                    this->adj_matrix[pos_old * this->num_nodes + pos_1] = 1;
    three_points[0] = pos_1; three_points[1] = pos_old; three_points[2] = pos_2;
    delete[] valid_edges;
    return(new_tree);
}

Tree* Tree::proposeMoveEdgeVertex(int *three_points) 
{
    int end_points_source[2];
    Tree *thinner_tree = proposeRemoveEdge(&end_points_source[0]);
    int *allowed_adjacencies = new int[this->num_nodes];
    int count_adj = 0;

    for (unsigned int i = 0; i < this->adj_list[end_points_source[0]].size(); i++)
        if (this->adj_list[end_points_source[0]][i] != end_points_source[1])
            allowed_adjacencies[count_adj++] = this->adj_list[end_points_source[0]][i];
    int num_adj_first = count_adj;
    for (unsigned int i = 0; i < this->adj_list[end_points_source[1]].size(); i++)
        if (this->adj_list[end_points_source[1]][i] != end_points_source[0])
            allowed_adjacencies[count_adj++] = this->adj_list[end_points_source[1]][i];
    int endpoint_choice = sample_uniform(count_adj);

    /*Rprintf("num_adj_first = %d\n", num_adj_first);
    Rprintf("endpoint_choice = %d\n", endpoint_choice);
    Rprintf("count_adj = %d\n", count_adj);
    Rprintf("choice = %d (from [%d %d])\n", allowed_adjacencies[endpoint_choice], end_points_source[0], end_points_source[1]);*/

    if (endpoint_choice < num_adj_first) {
        //originally adjacent to the first node. arrange nodes accordingly.
        three_points[0] = end_points_source[1];
        three_points[1] = end_points_source[0];
    } else {
        //originally adjacent to the second node. arrange nodes accordingly.
        three_points[0] = end_points_source[0];
        three_points[1] = end_points_source[1];        
    }
    three_points[2] = allowed_adjacencies[endpoint_choice];
    int pos_1 = three_points[0], pos_old = three_points[1], pos_2 = three_points[2];
    this->adj_matrix[pos_1 * this->num_nodes + pos_2] = 
                    this->adj_matrix[pos_2 * this->num_nodes + pos_1] = 1;
    this->adj_matrix[pos_1 * this->num_nodes + pos_old] = 
                    this->adj_matrix[pos_old * this->num_nodes + pos_1] = 0;
    Tree *new_tree = new Tree(this->adj_matrix, this->num_nodes);
    this->adj_matrix[pos_1 * this->num_nodes + pos_2] = 
                    this->adj_matrix[pos_2 * this->num_nodes + pos_1] = 0;
    this->adj_matrix[pos_1 * this->num_nodes + pos_old] = 
                    this->adj_matrix[pos_old * this->num_nodes + pos_1] = 1;
    delete[] allowed_adjacencies;

    delete thinner_tree;
    return(new_tree);
}


/** 
 * Propose a choice of triangle (Y_0 - Y_1 - Y_2) choosen from
 * the *uniform* distribution (notice this is not the case for the 
 * proposeMoveEdgeVertex sampler, which is biased towards triangles with few
 * neighbors).
 *
 * The probability of the choice in the forward direction is stored in *prob_fwd,
 * and the reverse probability in *prob_back.
 */
Tree* Tree::proposeMoveUniformTreeangle(int *three_points, double *prob_fwd, double *prob_back) 
{
    /* Count weights for each edge */
    double *weights = new double[this->num_edges];
    double sum_w = 0.;
    for (unsigned int e = 0; e < this->num_edges; e++) {
        weights[e] = this->adj_list[this->edges[e * 2]].size()     - 1 +
                     this->adj_list[this->edges[e * 2 + 1]].size() - 1;
        sum_w += weights[e];
    }
    for (unsigned int e = 0; e < this->num_edges; e++)
        weights[e] /= sum_w;

    /* Sample edge proportional to its score */
    unsigned int e_choice = sample_discrete(weights, this->num_edges);
    int end_points_source[] = {this->edges[2 * e_choice], this->edges[2 * e_choice + 1]}; 
    delete[] weights;

    /* Sample the third point */
    int *allowed_adjacencies = new int[this->num_nodes];
    int count_adj = 0;

    for (unsigned int i = 0; i < this->adj_list[end_points_source[0]].size(); i++)
        if (this->adj_list[end_points_source[0]][i] != end_points_source[1])
            allowed_adjacencies[count_adj++] = this->adj_list[end_points_source[0]][i];
    int num_adj_first = count_adj;
    for (unsigned int i = 0; i < this->adj_list[end_points_source[1]].size(); i++)
        if (this->adj_list[end_points_source[1]][i] != end_points_source[0])
            allowed_adjacencies[count_adj++] = this->adj_list[end_points_source[1]][i];
    int endpoint_choice = sample_uniform(count_adj);

    if (endpoint_choice < num_adj_first) {
        //originally adjacent to the first node. arrange nodes accordingly.
        three_points[0] = end_points_source[1];
        three_points[1] = end_points_source[0];
    } else {
        //originally adjacent to the second node. arrange nodes accordingly.
        three_points[0] = end_points_source[0];
        three_points[1] = end_points_source[1];        
    }
    three_points[2] = allowed_adjacencies[endpoint_choice];
    int pos_1 = three_points[0], pos_old = three_points[1], pos_2 = three_points[2];
    this->adj_matrix[pos_1 * this->num_nodes + pos_2] = 
                    this->adj_matrix[pos_2 * this->num_nodes + pos_1] = 1;
    this->adj_matrix[pos_1 * this->num_nodes + pos_old] = 
                    this->adj_matrix[pos_old * this->num_nodes + pos_1] = 0;
    Tree *new_tree = new Tree(this->adj_matrix, this->num_nodes);
    this->adj_matrix[pos_1 * this->num_nodes + pos_2] = 
                    this->adj_matrix[pos_2 * this->num_nodes + pos_1] = 0;
    this->adj_matrix[pos_1 * this->num_nodes + pos_old] = 
                    this->adj_matrix[pos_old * this->num_nodes + pos_1] = 1;
    delete[] allowed_adjacencies;

    *prob_fwd = 1. / sum_w;
    sum_w += this->adj_list[three_points[2]].size() + 1 - this->adj_list[three_points[1]].size();
    *prob_back = 1. / sum_w;
    return(new_tree);
}


/**
 * Similar to the previous method, but relies only on the choice of the treeangle.
 */
void Tree::proposeMoveUniformTreeangle(int *three_points) 
{
    /* Count weights for each edge */
    double *weights = new double[this->num_edges];
    double sum_w = 0.;
    for (unsigned int e = 0; e < this->num_edges; e++) {
        weights[e] = this->adj_list[this->edges[e * 2]].size()     - 1 +
                     this->adj_list[this->edges[e * 2 + 1]].size() - 1;
        sum_w += weights[e];
    }
    for (unsigned int e = 0; e < this->num_edges; e++)
        weights[e] /= sum_w;

    /* Sample edge proportional to its score */
    unsigned int e_choice = sample_discrete(weights, this->num_edges);
    int end_points_source[] = {this->edges[2 * e_choice], this->edges[2 * e_choice + 1]}; 
    delete[] weights;

    /* Sample the third point */
    int *allowed_adjacencies = new int[this->num_nodes];
    int count_adj = 0;

    for (unsigned int i = 0; i < this->adj_list[end_points_source[0]].size(); i++)
        if (this->adj_list[end_points_source[0]][i] != end_points_source[1])
            allowed_adjacencies[count_adj++] = this->adj_list[end_points_source[0]][i];
    int num_adj_first = count_adj;
    for (unsigned int i = 0; i < this->adj_list[end_points_source[1]].size(); i++)
        if (this->adj_list[end_points_source[1]][i] != end_points_source[0])
            allowed_adjacencies[count_adj++] = this->adj_list[end_points_source[1]][i];
    int endpoint_choice = sample_uniform(count_adj);

    if (endpoint_choice < num_adj_first) {
        //originally adjacent to the first node. arrange nodes accordingly.
        three_points[0] = end_points_source[1];
        three_points[1] = end_points_source[0];
    } else {
        //originally adjacent to the second node. arrange nodes accordingly.
        three_points[0] = end_points_source[0];
        three_points[1] = end_points_source[1];        
    }
    three_points[2] = allowed_adjacencies[endpoint_choice];
}


/**
 * Finds the path from 'start' to 'end' and stores it in 'path_buffer'.
 * Please notice that 'start' is NOT included in the output array. The first element
 * is a node adjacent to 'start'.
 */

bool Tree::findPath(int start, int end, int path_buffer[]) 
{
    if (start == end) {
        path_buffer[0] = -1;
        return true;
    }
    bool *marked = new bool[this->num_nodes];
    for (unsigned int i = 0; i < this->num_nodes; i++)
        marked[i] = false;
    marked[start] = true;
    for (unsigned int i = 0; i < this->adj_list[start].size(); i++)
        if (!marked[this->adj_list[start][i]])
            if (findPathRecurse(this->adj_list[start][i], end, &path_buffer[1], &marked[0])) {
                path_buffer[0] = this->adj_list[start][i];
                delete[] marked;
                return true;
            }
    delete[] marked;
    return false;
}

bool Tree::findPathRecurse(int start, int end, int path_buffer[], bool marked[]) 
{
    marked[start] = true;
    if (start == end) {
        path_buffer[0] = -1;
        return true;
    }
    for (unsigned int i = 0; i < this->adj_list[start].size(); i++)
        if (!marked[this->adj_list[start][i]])
            if (findPathRecurse(this->adj_list[start][i], end, &path_buffer[1], &marked[0])) {
                path_buffer[0] = this->adj_list[start][i];
                return true;
            }
    return false;
}

/**
 * Returns true if nodes 'start' and 'end' are connected according to 'adjacencies'
 */
bool Tree::isConnected(int *adjacencies, unsigned int num_nodes, int start, int end) 
{
    if (start == end)
        return true;
    
    bool *marked = new bool[num_nodes];
    for (unsigned int i = 0; i < num_nodes; i++)
        marked[i] = false;
    marked[start] = true;
    
    for (unsigned int i = 0; i < num_nodes; i++) {
        if (adjacencies[start * num_nodes + i] == 0)
            continue;
        if (!marked[i])
            if (isConnectedRecurse(adjacencies, num_nodes, i, end, marked)) {
                delete marked;
                return true;
            }
    }
    
    delete[] marked;
    return false;
}

bool Tree::isConnectedRecurse(int *adjacencies, int num_nodes, int start, int end, bool marked[]) 
{
    marked[start] = true;
    if (start == end)
        return true;
    for (unsigned int i = 0; i < num_nodes; i++)
        if (adjacencies[start * num_nodes + i] == 1 && !marked[i])
            if (isConnectedRecurse(adjacencies, num_nodes, i, end, marked))
                return true;
    return false;
}


bool Tree::findPathDebug(int start, int end, int path_buffer[]) 
{
    Rprintf("***RECAP***\n");
    printTreeInformationR();
    Rprintf("START = %d, END = %d\n", start, end);
    if (start == end) {
        path_buffer[0] = -1;
        return true;
    }
    bool *marked = new bool[this->num_nodes];
    for (unsigned int i = 0; i < this->num_nodes; i++)
        marked[i] = false;
    marked[start] = true;
    for (unsigned int i = 0; i < this->adj_list[start].size(); i++) {
        Rprintf("Passing through edge (%d, %d) [neighbor number %d]\n", start, this->adj_list[start][i], i);        
        if (!marked[this->adj_list[start][i]]) {            
            if (findPathRecurseDebug(this->adj_list[start][i], end, &path_buffer[1], &marked[0], 1)) {
                Rprintf("Returning sucessfully from edge (%d, %d)\n", start, this->adj_list[start][i]);
                path_buffer[0] = this->adj_list[start][i];
                delete marked;
                return true;
            } else
                Rprintf("Returning unsuccessfully from edge (%d, %d)\n", start, this->adj_list[start][i]);
        }
    }
    delete marked;
    return false;
}

bool Tree::findPathRecurseDebug(int start, int end, int path_buffer[], bool marked[], int level) 
{
    marked[start] = true;
    if (start == end) {
        path_buffer[0] = -1;
        return true;
    }
    for (unsigned int i = 0; i < this->adj_list[start].size(); i++) {
        for (int x = 0; x < level; x++)
            Rprintf("   ");
        Rprintf("Passing through edge (%d, %d) [neighbor number %d]\n", start, this->adj_list[start][i], i);        
        if (!marked[this->adj_list[start][i]]) {
            if (findPathRecurseDebug(this->adj_list[start][i], end, &path_buffer[1], &marked[0], level + 1)) {
                for (int x = 0; x < level; x++)
                    Rprintf("   ");
                Rprintf("Returning unsuccessfully from edge (%d, %d)\n", start, this->adj_list[start][i]); 
                path_buffer[0] = this->adj_list[start][i];
                return true;
            } else {
                for (int x = 0; x < level; x++)
                    Rprintf("   ");
                Rprintf("Returning unsuccessfully from edge (%d, %d)\n", start, this->adj_list[start][i]); 
            }
        }
    }
    return false;
}

void Tree::copyAdjacencies(double *out) 
{
    out[0] = this->num_edges;
    int pos = 1;
    for (unsigned int i = 0; i < this->num_edges; i++) {
        out[pos++] = this->edges[2 * i];
        out[pos++] = this->edges[2 * i + 1];
    }
}

Tree* Tree::clone()
{
    return new Tree(this->adj_matrix, this->num_nodes);
}

/**
 * Samples a spanning tree of 'p' nodes from the uniform distribution
 */

Tree* sample_unif_tree(int p) 
{
    int *s = new int[p];    
    perm(s, p);
    int *adj = new int[p * p];
    for (int i = 0; i < p; i++) {
        for (int j = i + 1; j < p; j++)
            adj[i * p + j] = adj[j * p + i] = 0;
        adj[i * p + i] = 0;
    }
    for (int i = 1; i < p; i++) {
        int parent = sample_uniform(i);
        adj[s[parent] * p + s[i]] = adj[s[i] * p + s[parent]] = 1;
    }
    Tree *new_tree = new Tree(adj, p);
    delete[] adj; delete[] s;
    return new_tree;
}

/**
 * Generates a 1 - 2 - 3 - ... - p chain deterministically.
 */

Tree* sample_chain_tree(int p) 
{
    int *adj = new int[p * p];
    for (int i = 0; i < p; i++) {
        for (int j = i + 1; j < p; j++)
            adj[i * p + j] = adj[j * p + i] = 0;
        adj[i * p + i] = 0;
    }
    for (int i = 1; i < p; i++) {
        int parent = i - 1;
        adj[parent * p + i] = adj[i * p + parent] = 1;
    }
    Tree *new_tree = new Tree(adj, p);
    delete[] adj;
    return new_tree;
}

extern "C" {

  /*
   * Temporary debug code
   *
   * R interface function
   */

	void debug_tree(int *matrix, int *num_nodes, int *output) {
		Graph *graph = new Graph(matrix, (unsigned int) *num_nodes);
		output[0] = graph->num_nodes;
		output[1] = graph->num_edges;
		delete graph;
	}
}
