extern "C"
{
#include "matrix.h"
#include "rhelp.h"
#include "rand.h"
}
#include "marginal.h"
#include "margempirical.h"
#include <math.h>
#include <Rmath.h>
#include <stdlib.h>
#include <stdio.h>
#include <assert.h>
#include <string.h>
#include <string>
#include <fstream>
#include "R.h"
using namespace std;


/*
 * Empirical:
 * 
 * Parameterizes a marginal through an 'empirical cdf representation': essentially,
 * an histogram-like representation that is motivated by the empirical cdf estimator.
 *
 * We receive as parameters an array of size 2 * p + 5, where the first number indicates the
 * value of 'p'. The next 'p + 2' numbers indicate the cut-off points, and the
 * last 'p + 2' numbers, the corresponding cumulative mass.
 *
 * For instance, for c_i = params[i], m_i = params[(1 + (p + 2)) + i] equals to the statement
 * P(x <= c_i) = m_i.
 *
 * Two special cases: the quantity stored in params[1] and params[p + 2] 
 * (the first and last 'cut-offs'). The quantities in params[p + 3] and
 * params[2 * p + 4] (i.e., the corresponding cdf values) are ignored. The quantities
 * in params[1] and params[p + 2] are the 'bottom' and 'top' limits for the range
 * where we can impute missing data, if needed ('bottom' and 'top' are usually
 * set by looking at the training data).
 *
 * For instance, the number stored in 'this->mass[0]' is the probability of the
 * random variable being sampled at values less than 'this->levels[0]'.
 * WE ASSUME THAT THE SUPPORT OF THE DISTRIBUTION IS THE WHOLE REAL LINE.
 * However, we want to sample a number less than 'this->levels[0]', we will
 * sample uniformly in the interval ['this->bottom', 'this->levels[0]'].
 * An analogous imputation mechanism is used for the interval
 * ['this->levels[this->numLevels - 1]', 'this->top'].
 * 
 * NO PRIORS ARE USED WITH THIS CHOICE OF MARGINAL. IT IS ASSUMED TO BE FIXED,
 * AND THEREFORE WE DO NOT HAVE MCMC PROPOSALS FOR IT.
 */

/*
 * Empirical:
 * 
 * constructor function that takes new settings of the parameters,
 * but copies data and prior and RW proposal parameters from an old
 * module
 */


Empirical::Empirical(double *params, const Empirical *old)
{
  /* copy data from the old module */
  this->n = old->n;
  this->x = old->x;

  /* copy the new parameter settings */
  setInternalParams(params);

  /* calculate the ranks */
  CalcRanks();
}


/*
 * Empirical:
 * 
 * constructor function that takes all parameters in the model
 * including the priors and RW proposal parameters through the
 * double vector d
 */

Empirical::Empirical(unsigned int n, double *x, double *d) : Marginal(n, x)
{
  /* copy in the parameters */
  setInternalParams(d);

  /* calculate the ranks */
  CalcRanks();
}


/**
 * setInternalParams:
 *
 * Where all the dirty work of converting the raw representation in 'params' into
 * the parameters 'levels', 'mass', 'top' and 'bottom' is done.
 */

void Empirical::setInternalParams(double *params)
{
  this->numLevels = (unsigned int) trunc(params[0]);
  assert(this->numLevels > 0);
  /*if (this->numLevels <= 0) {
      Rprintf("???\n");
      for (int i = 0; i < 3; i++)
          Rprintf("levels[%d] = %f\n", i, params[i]);
      this->numLevels = 1;
      this->levels = new double[this->numLevels];
      this->levels[0] = 1;
      this->mass   = new double[this->numLevels];
      this->mass[0] = 0.5;
      return;
  }*/
  this->levels = new double[this->numLevels];
  this->mass   = new double[this->numLevels];
  memcpy(this->levels, &params[2], this->numLevels * sizeof(double));
  memcpy(this->mass, &params[this->numLevels + 4], this->numLevels * sizeof(double));
  this->bottom = params[1]; this->top = params[this->numLevels + 2];
  
  //Rprintf("Number of levels = %d, bottom = %f, top = %f\n", this->numLevels, this->bottom, this->top);
  //for (unsigned int i = 0; i < this->numLevels; i++)
  //    Rprintf("Level[%d] = %f, Mass[%d] = %f\n", i, this->levels[i], i, this->mass[i]);
  
  for (unsigned int i = 0; i < this->numLevels; i++) {
      if (this->mass[i] <= 0)
          error("NON-POSITIVE EMPIRICAL CDF MASS!\n");
      if (this->mass[i] >= 1)
          error("EMPIRICAL CDF MASS CANNOT BE GREATER THAN OR EQUAL TO 1!\n");
      if (i > 0 && this->mass[i] <= this->mass[i - 1])
          error("EMPIRICAL CDF MASS SHOULD BE STRICTLY INCREASING!\n");
  }
}

/* 
 * ~Empirical:
 * 
 * destructor
 */

Empirical::~Empirical(void)
{
    delete[] this->levels;
    delete[] this->mass;
}


/*
 * NewRanks:
 *
 * populate z with the ranks passed in and calculate
 * the corresponding sample x -- usually called by a
 * copula function
 */

double* Empirical::NewRanks(unsigned int n, double *z)
{
  this->n = n;

  assert(this->z == NULL);
  this->z = z;

  assert(x == NULL);
  x = new_vector(n);
  for (unsigned int i = 0; i < n; i++)
      x[i] = qempirical(z[i]);

  return x;
}

/*
 * qempirical: returns the variable associated with the quantile z
 *             according to the internally parameterized empirical distribution
 */

double Empirical::qempirical(double z) const
{
  assert(z <= 1 && z >= 0);
  if (z <= this->levels[0])
      return 0.5 * this->levels[0];
  if (z > this->levels[this->numLevels - 1])
      return 0.5 * (this->levels[numLevels - 1] + 1.);
  for (unsigned int i = 1; i < this->numLevels; i++)
      if (z <= this->mass[i])
         return 0.5 * (this->levels[i - 1] + this->levels[i]);
  error("Invalid marginal distribution object!\n");
  return 0;
}

/*
 * pempirical: CDF function for the given nonparametric CDF
 */

double Empirical::pempirical(double x) const
{
  for (unsigned int i = 0; i < this->numLevels; i++)
      if (x <= this->levels[i])
         return this->mass[i];
  return 1.;
}

/*
 * CalcRanks:
 *
 * calculate the ranks based on the sample, x,
 * and the parameters mu and s2
 */

void Empirical::CalcRanks(void)
{
  if (z) free(z);
  z = new_vector(n);
  for (unsigned int i = 0; i < n; i++) {
      z[i] = pempirical(x[i]);
      //Rprintf("x = %f, Rank = %f, d= %d\n", x[i], z[i], i);
      if (z[i] == 0.) z[i] = 1.e-16;
      else if (z[i] == 1.) z[i] = 1. - 1.e-16;
  }
}


/*
 * CalcRanks:
 *
 * calculate the ranks for a particular point, x
 * under this marginal
 */

double Empirical::CalcRanks(double x) const
{
  double z = pempirical(x);
    if (z == 0.) z = 1.e-16;
    else if (z == 1.) z = 1. - 1.e-16;
    if (z < 0) {
        for (unsigned int i = 0; i < this->numLevels; i++)
            Rprintf("Level[%d] = %f, Mass[%d] = %f\n", i, this->levels[i], i, this->mass[i]);
        error("NEGATIVE MASS\n");
    }
  return z;
}


/*
 * LLik:
 * 
 * returns the log likelihood based on the current
 * parameter setting.
 *
 * WARNING: THIS IS SOMEWHAT MEANINGLESS WHEN EVALUATION TEST DATA,
 * SINCE THE LOG-LIKELIHOOD IS JUST -INF IF THE POINT DOES NOT SHOW
 * IN WHATEVER DATASET WAS USED TO GENERATE THIS DISTRIBUTION.
 *
 * SINCE WE REQUIRE THIS FUNCTION TO IMPUTE MISSING DATA, WE NEED IT -
 * IN THIS CASE WE DEFINE IT AS THE EMPIRICAL PDF.
 */

double Empirical::LLik() const
{
  return -log((double) this->n);
}

/*
 * LLik(x):
 * 
 * calculate the log likelihood based on the current
 * parameter setting for a given data point x
 *
 * WE ASSUME THAT 'X' IS IN THE TRAINING DATA. SEE COMMENT ABOVE.
 */

double Empirical::LLik(double x) const
{
  return -log((double) this->n);
}

/*
 * NumParams:
 *
 * return the number of parameters managed by
 * this marginal
 */

unsigned int Empirical::NumParams(void) const
{
  return 1 + 2 * (this->numLevels + 2);
}


/*
 * Params:
 *
 * fill the out vector with the parameter values -- the out vector
 * must have the right amount of space pre-allocated
 */

void Empirical::Params(double *out) const
{
  out[0] = this->numLevels;
  out[1] = this->bottom;
  memcpy(&out[2], this->levels, this->numLevels * sizeof(double));
  out[this->numLevels + 2] = this->top;
  out[this->numLevels + 3] = 0.;
  memcpy(&out[this->numLevels + 4], this->mass, this->numLevels * sizeof(double));
  out[2 * this->numLevels + 4] = 1.;
}


/*
 * RW:
 *
 * Random Walk proposals for this marginal -- the
 * which arguments specifies which proposal to augment
 * from 1:NumParams()
 */

Marginal* Empirical::RW(unsigned int which, double *lqfwd, double *lqbak) const
{
  error("Empirical :: Random walk not implemented!\n");
  return NULL;
}


/*
 * RWvalue:
 *
 * Random Walk proposal for the variable at a current value 'value'.
 * Used for data imputation methods.
 *
 * This is very different from standard RWvalue proposals, since we sample
 * only in the space of cut-off values. THIS WILL BE PROBLEMATIC IF THERE ARE
 * MANY REPEATED VALUES.
 *
 * Consider that we have all training points sorted, and a particular choice of
 * value 'x'. We look at a window points to the left and right of
 * 'x' in the sorted training data, and pick one of these uniformly.
 *
 * The sorted data is implicitly encoded by 'levels', 'mass' and 'n'.
 */

double Empirical::RWvalue(double value, double *lqfwd, double *lqbak) const
{
  double window = 0.1;
  int    level = this->numLevels;
  double mass = 1;

  for (unsigned int i = 0; i < this->numLevels; i++)
      if (value == this->levels[i]) {
          level = i;
          mass = this->mass[i];
          break;
      }

  /* Find the window of mass probability */
  double l_prob = mass - window; if (l_prob < 0) l_prob = 0;
  if (l_prob > this->mass[this->numLevels - 1]) {
      double value_prime = this->top;
      *lqfwd = log(1.);
      *lqbak = log(1.);
      /* Will get stuck forever. Raise error */
      error("Invalid Markov chain for the empirical marginal: proposal does not provide ergodic chain\n");
      return value_prime;
  }
  double u_prob = mass + window; if (u_prob > 1) u_prob = 1;

  /* Choose probability cut-off */
  double prob_choice = runif(l_prob, u_prob);
  double value_prime = 0;

  /* Find respective levels */

  /* First possibility: 'top' value */
  if (prob_choice > this->mass[this->numLevels - 1]) {
      value_prime = this->top;

      *lqfwd = log((u_prob - this->mass[this->numLevels - 1]) / (u_prob - l_prob));

      double l_prob_prime = 1 - window;
      double u_prob_prime = 1;
      if (value > this->levels[this->numLevels - 1])
          *lqbak = log(1.);
      else {
          for (unsigned int i = 0; i < this->numLevels; i++)
              if (this->levels[i] > l_prob_prime) {
                  *lqbak = log((this->mass[i] - l_prob_prime) / (u_prob_prime - l_prob_prime));
                  break;
              }
      }

      return value_prime;
  }
 
  /* Alternative: non-'top' value */
  for (unsigned int i = 0; i < this->numLevels; i++)
      if (this->levels[i] > prob_choice) {
          value_prime = this->levels[i];
          
          if (i > 0 && l_prob < this->mass[i - 1])
             *lqfwd = log((this->mass[i] - this->mass[i - 1]) / (u_prob - l_prob));
          else
             *lqfwd = log((this->mass[i] - l_prob) / (u_prob - l_prob));
          
          double r_mass = this->mass[i];
          double l_prob_prime = r_mass - window; if (l_prob_prime < 0) l_prob_prime = 0;
          double u_prob_prime = r_mass + window; if (u_prob_prime > 1) u_prob_prime = 1;

          if (i > 0 && l_prob_prime < this->mass[i - 1])
             *lqbak = log((this->mass[i] - this->mass[i - 1]) / (u_prob_prime - l_prob_prime));
          else
             *lqbak = log((this->mass[i] - l_prob_prime) / (u_prob_prime - l_prob_prime));

          return value_prime;
      }

  error("Inconsistency empirical marginal sampling!\n");
  return 0.;
}

/*
 * LPrior:
 *
 * return the log (joint) prior of all of the parameters
 */

double Empirical::LPrior(void) const
{
  error("Empirical::LPrior not implemented!\n");
  return -1;
}

/*
 * setParams:
 *
 * Sets new parameter values and reset the ranked data
 */

void Empirical::setParams(double *in) 
{
  setInternalParams(in);
  CalcRanks();
}

/*
 * randomize:
 *
 * initialize parameters to random numbers.
 * Meaningless in the empirical marginal case.
 */

void Empirical::randomize()
{
  error("Empirical::randomize not implemented!\n");
}

