extern "C"
{
#include "matrix.h"
#include "rhelp.h"
#include "rand.h"
}
#include "marginal.h"
#include "marggamma.h"
#include <math.h>
#include <Rmath.h>
#include <stdlib.h>
#include <stdio.h>
#include <assert.h>
#include <string.h>
#include <string>
#include <fstream>
using namespace std;


/*
 * Gamma:
 * 
 * constructor function that sets x to be null
 */

Gamma::Gamma(double a, double b) : Marginal()
{
  this->a = a;
  assert(a > 0);
  this->a_win = 0.01;
  this->b = b;
  assert(b > 0);
  this->b_win = 0.01;

  this->a_a = 1;
  this->a_b = 1;
  this->b_a = 1;
  this->b_b = 1;
}


/*
 * Gamma:
 * 
 * constructor function that takes new settings of the parameters,
 * but copies data and prior and RW proposal parameters from an old
 * module
 */


Gamma::Gamma(double a, double b, double a_a, double a_b, double b_a, double b_b, const Gamma *old)
{
  /* copy data from the old module */
  this->n = old->n;
  this->x = old->x;

  /* copy priors and RW parameters form the old module */
  this->a_win = old->a_win;
  this->b_win = old->b_win;

  /* copy the new parameter settings */
  assert(a > 0); assert(b > 0);
  this->a = a;
  this->b = b;

  this->a_a = a_a; assert(this->a_a > 0);
  this->a_b = a_b; assert(this->a_b > 0);
  this->b_a = b_a; assert(this->b_a > 0);
  this->b_b = b_b; assert(this->b_b > 0);

  /* calculate the ranks */
  CalcRanks();
}


/*
 * Gamma:
 * 
 * constructor function that takes all parameters in the model
 * including the priors and RW proposal parameters through the
 * double vector d
 */

Gamma::Gamma(unsigned n, double *x, double *d) : Marginal(n, x)
{    
  /* copy in the parameters */
  this->a = d[0];
  assert(a > 0);
  this->b = d[1];
  assert(b > 0);
  this->a_win = d[2];  
  assert(a_win > 0);
  this->b_win = d[3];
  assert(b_win > 0);

  this->a_a = d[4]; assert(this->a_a > 0);
  this->a_b = d[5]; assert(this->a_b > 0);
  this->b_a = d[6]; assert(this->b_a > 0);
  this->b_b = d[7]; assert(this->b_b > 0);

  /* calculate the ranks */
  CalcRanks();
}


/* 
 * ~Gamma:
 * 
 * destructor
 */

Gamma::~Gamma(void)
{
}


/*
 * NewRanks:
 *
 * populate z with the ranks passed in and calculate
 * the corresponding sample x -- usually called by a
 * copula function
 */

double* Gamma::NewRanks(unsigned int n, double *z)
{
  this->n = n;

  assert(this->z == NULL);
  this->z = z;

  assert(x == NULL);
  x = new_vector(n);
  for(unsigned int i = 0; i < n; i++) 
    x[i] = qgamma(z[i], this->a, this->b, 1, 0);

  return x;
}


/*
 * CalcRanks:
 *
 * calculate the ranks based on the sample, x,
 * and the parameters mu and s2
 */

void Gamma::CalcRanks(void)
{
  if(z) free(z);
  z = new_vector(n);
  for(unsigned int i = 0; i < n; i++) {
    z[i] = pgamma(x[i], this->a, this->b, 1, 0);
    if (z[i] < 1.e-16) z[i] = 1.e-16;
    else if (z[i] > 1. - 1.e-16) z[i] = 1. - 1.e-16;
  }
}


/*
 * CalcRanks:
 * 
 * Calculate the rank of a particular data point x
 */

double Gamma::CalcRanks(double x) const
{
  double z = pgamma(x, this->a, this->b, 1, 0);
    if (z < 1.e-16) z = 1.e-16;
    else if (z > 1. - 1.e-16) z = 1. - 1.e-16;
  return z;
}


/*
 * LLik:
 * 
 * calculate the log likelihood based on the current
 * parameter setting
 */

double Gamma::LLik() const
{
  assert(x);
  double llik = 0;
  for(unsigned int i=0; i < n; i++)
    llik += dgamma(x[i], this->a, this->b, 1);
  return llik;
}

/*
 * LLik(x):
 * 
 * calculate the log likelihood based on the current
 * parameter setting for a given data point x
 */

double Gamma::LLik(double x) const
{
  return dgamma(x, this->a, this->b, 1);
}

/*
 * NumParams:
 *
 * return the number of paramters managed by
 * this marginal
 */

unsigned int Gamma::NumParams(void) const
{
  return 2;
}


/*
 * Params:
 *
 * fill the out vector with the parameter values -- the out vector
 * must have the right amount of space pre-allocated
 */

void Gamma::Params(double *out) const
{
  out[0] = this->a;
  out[1] = this->b;
}


/*
 * RW:
 *
 * Random Walk proposals for this marginal -- the
 * which arguments specifies which proposal to augment
 * from 1:NumParams()
 */

Marginal* Gamma::RW(unsigned int which, double *lqfwd, double *lqbak) const
{
  /* choose a type of Randow Walk proposal */
  switch (which) {
     case 1: return RWa(lqfwd, lqbak);
     case 2: return RWb(lqfwd, lqbak);
     default: error("Gamma RW proposal %d unknown\n", which);
  }
  return NULL;
}


/*
 * RWa:
 *
 * Randowm Walk proposal for a
 */

Marginal* Gamma::RWa(double *lqfwd, double *lqbak) const
{
  double a_prime = unif_propose_prop(this->a, this->a_win, lqfwd, lqbak);
  //double a_prime = truncgaussian_propose_bound(this->a, 0, 10000, this->a_win, lqfwd, lqbak);
  return new Gamma(a_prime, this->b, this->a_a, this->a_b, this->b_a, this->b_b, this);
}


/*
 * RWb:
 *
 * Randowm Walk proposal for b
 */

Marginal* Gamma::RWb(double *lqfwd, double *lqbak) const
{
  double b_prime = unif_propose_prop(this->b, this->b_win, lqfwd, lqbak);
  //double b_prime = truncgaussian_propose_bound(this->b, 0, 10000, this->b_win, lqfwd, lqbak);
  return new Gamma(this->a, b_prime, this->a_a, this->a_b, this->b_a, this->b_b, this);
}

/*
 * RWvalue:
 *
 * Randowm Walk proposal for the variable at a current value 'value'.
 */

double Gamma::RWvalue(double value, double *lqfwd, double *lqbak) const
{
  //double value_prime = unif_propose_bound(value, 0., 10000., 0.01, lqfwd, lqbak);
  double value_prime = unif_propose_prop(value, 0.01, lqfwd, lqbak);
  //double value_prime = truncgaussian_propose_bound(value, 0, 10000, 0.01, lqfwd, lqbak);
  return value_prime;
}


/*
 * LPrior:
 *
 * return the log (joint) prior of all of the parameters
 */

double Gamma::LPrior(void) const
{
  return dgamma(this->a, this->a_a, this->a_b, 1) +
         dgamma(this->b, this->b_a, this->b_b, 1);
  return 0.;
}

/*
 * setParams:
 *
 * Sets new parameter values and reset the ranked data
 */

void Gamma::setParams(double *in) 
{
  this->a = in[0];
  assert(this->a > 0);
  this->b = in[1];
  assert(this->b > 0);
  CalcRanks();
}

/*
 * randomize:
 *
 * initialize parameters to random numbers.
 * THIS IS NOT SAMPLING FROM THE PRIOR.
 */

void Gamma::randomize()
{    
    this->a = runif(1, 2);
    this->b = runif(1, 2);
    CalcRanks();
}
