extern "C"
{
#include "matrix.h"
#include "rhelp.h"
#include "rand.h"
}
#include "marginal.h"
#include "margstudent.h"
#include <math.h>
#include <Rmath.h>
#include <stdlib.h>
#include <stdio.h>
#include <assert.h>
#include <string.h>
#include <string>
#include <fstream>
using namespace std;


/*
 * Student:
 * 
 * constructor function that sets x to be NULL, and 
 * ignores the prior and RW proposal parameters
 */

Student::Student(double mu, double s2, double nu) : Marginal()
{
  assert(s2 > 0);
  this->mu = mu;
  this->s2 = s2;
  this->mu_s2q = 1;
  assert(nu >= 1);
  this->nu = nu;
}



/*
 * Student:
 * 
 * constructor function that takes new settings of the parameters,
 * but copies data and prior and RW proposal parameters from an old
 * module
 */


Student::Student(double mu, double s2, double nu, const Student *old)
{
  /* copy data from the old module */
  this->n = old->n;
  this->x = old->x;

  /* copy priors and RW parameters form the old module */
  this->mu_s2q = old->mu_s2q;
  this->s2_prop = old->s2_prop;
  this->nu_prop = old->nu_prop;

  /* copy the new parameter settings */  
  this->mu = mu;
  assert(s2 > 0);
  this->s2 = s2;
  assert(nu >= 1);
  this->nu = nu;
  
  /* calculate the ranks */
  CalcRanks();
}


/*
 * Student:
 * 
 * constructor function that takes all parameters in the model
 * including the priors and RW proposal parameters through the
 * double vector d
 */

Student::Student(unsigned n, double *x, double *d) : Marginal(n, x)
{
  /* Copy in model parameters */
  this->mu = d[0];
  this->s2 = d[1];
  assert(this->s2 > 0);
  this->nu = d[2];
  assert(this->nu >= 1);

  /* Copy in MH parameters */
  this->mu_s2q = d[3];
  assert(this->mu_s2q > 0);
  this->s2_prop = d[4];
  assert(s2_prop > 0 && s2_prop < 1);
  this->nu_prop = d[5];
  assert(nu_prop > 0 && nu_prop < 1);

  /* calculate the ranks */
  CalcRanks();
}


/* 
 * ~Student:
 * 
 * destructor
 */

Student::~Student(void)
{
}


/*
 * NewRanks:
 *
 * populate z with the ranks passed in and calculate
 * the corresponding sample x -- usually called by a
 * copula function -- returns the sample x
 */

double* Student::NewRanks(unsigned int n, double *z)
{
  this->n = n;

  assert(this->z == NULL);
  this->z = z;

  assert(x == NULL);
  x = new_vector(n);
  double ss2 = sqrt(s2);
  for (unsigned int i = 0; i < n; i++) {
       x[i] = qt(z[i], nu, 1, 0);
       x[i] = ss2 * x[i] + mu;
  }

  return x;
}


/*
 * CalcRanks:
 *
 * calculate the ranks based on the sample, x,
 * and the parameters mu and s2
 */

void Student::CalcRanks(void)
{
  if(z) free(z);
  z = new_vector(n);
  double ss2 = sqrt(s2);
  for (unsigned int i = 0; i < n; i++) {
       double y = (x[i] - mu) / ss2;
       z[i] = pt(y, nu, 1, 0);
       if (z[i] == 0.) 
           z[i] = 1.e-16;
       else if (z[i] == 1.) 
           z[i] = 1. - 1.e-16;
  }
}


/*
 * CalcRanks:
 * 
 * Calculate the rank of a particular data point x
 */

double Student::CalcRanks(double x) const
{
  double ss2 = sqrt(s2);
  double y = (x - mu) / ss2;
  double z = pt(y, nu, 1, 0);
      if (z == 0.) z = 1.e-16;
      else if (z == 1.) z = 1. - 1.e-16;
  return z;
}

/*
 * LLik:
 * 
 * calculate the log likelihood based on the current
 * parameter setting
 */

double Student::LLik() const
{
  assert(x);
  double llik = 0;
  double ss2 = sqrt(s2);
  for (unsigned int i = 0; i < n; i++) {
       double y = (x[i] - mu) / ss2;
       llik += dt(y, nu, 1);
  }
  return llik - n * log(ss2);
}

double Student::LLik(double x) const
{
  double ss2 = sqrt(s2);
  double y = (x - mu) / ss2;
  return dt(y, nu, 1) - log(ss2);
}

/*
 * NumParams:
 *
 * return the number of paramters managed by
 * this marginal
 */

unsigned int Student::NumParams(void) const
{
  return 3;
}


/*
 * Params:
 *
 * fill the out vector with the parameter values -- the out vector
 * must have the right amount of space pre-allocated
 */

void Student::Params(double *out) const
{
  out[0] = mu;
  out[1] = s2;
  out[2] = nu;
}


/*
 * RW:
 *
 * Random Walk proposals for this marginal -- the
 * which arguments specifies which proposal to augment
 * from 1:NumParams()
 */

Marginal* Student::RW(unsigned int which, double *lqfwd, double *lqbak) const
{
  /* choose a type of Randow Walk proposal */
  switch(which) {
  case 1: return RWmu(lqfwd, lqbak);
  case 2: return RWs2(lqfwd, lqbak);
  case 3: return RWnu(lqfwd, lqbak);
  default: error("Normal RW proposal %d unknown\n", which);
  }
  return NULL;
}


/*
 * RWmu:
 *
 * Randowm Walk proposal for the mean
 */

Marginal* Student::RWmu(double *lqfwd, double *lqbak) const
{
  double ss2q = sqrt(this->mu_s2q);
  double mu_prime = rnorm(this->mu, ss2q);
  *lqfwd = *lqbak = dnorm(this->mu, mu_prime, ss2q, 1);
  return new Student(mu_prime, this->s2, this->nu, this);
}


/*
 * RWs2:
 *
 * Randowm Walk proposal for the variance
 */

Marginal* Student::RWs2(double *lqfwd, double *lqbak) const
{
  double s2_prime = unif_propose_pos(this->s2, lqfwd, lqbak, this->s2_prop);
  return new Student(this->mu, s2_prime, this->nu, this);
}


/*
 * RWnu:
 *
 * Randowm Walk proposal for the degrees of freedom parameter
 */

Marginal* Student::RWnu(double *lqfwd, double *lqbak) const
{
  /* must make sure nu proposal is >= 1 */
  double nu_prime = unif_propose_pos(this->nu - 1, lqfwd, lqbak, this->nu_prop) + 1;
  return new Student(this->mu, this->s2, nu_prime, this);
}

/*
 * RWvalue:
 *
 * Randowm Walk proposal for the variable at a current value 'value'.
 */

double Student::RWvalue(double value, double *lqfwd, double *lqbak) const
{
  double ss2q = sqrt(this->mu_s2q);
  double value_prime = rnorm(value, ss2q);
  *lqfwd = *lqbak = dnorm(value, value_prime, ss2q, 1);
  return value_prime;
}

/*
 * LPrior:
 *
 * return the log (joint) prior of all of the parameters
 */

double Student::LPrior(void) const
{
  double lp = dgamma(this->nu, 1.0, 0.1, 1);
  return lp - log(this->s2);
}

/*
 * setParams:
 *
 * Sets new parameter values and reset the ranked data
 */

void Student::setParams(double *in) 
{
  this->mu = in[0];
  this->s2 = in[1];  
  this->nu = in[2];
  assert(this->s2 > 0);
  assert(this->nu >= 1);
  CalcRanks();
}

/*
 * randomize:
 *
 * initialize parameters to random numbers.
 * TODO SAMPLE FROM THE PRIORS
 */

void Student::randomize()
{
  this->mu = runif(-1, 1);
  this->s2 = runif(0.1, 2);
  this->nu = runif(1, 5);
  CalcRanks();
}
