#include "rand.h"
#include "rhelp.h"
#include "matrix.h"
#include <Rmath.h>
#include <assert.h>

/*
 * gaussian_propose_pos:
 * 
 * propose a new positive "ret" based on an old value "last"
 * by proposing from a Gaussian centered at "last" with variance "var".
 * Return the forward and backward log probabilities.
 */

double gaussian_propose_pos(double last, double *l_q_fwd, double *l_q_bak,
			                double var)
{
  double ret;

  /* check for a valid proportion argument */
  assert(var > 0);

  /* propose new d, and compute proposal probability */
  double move = rnorm(last, sqrt(var));
  ret = last + move;
  *l_q_fwd = *l_q_bak = -0.5 * (3.1415 + log(var) + (ret - move) * (ret - move) / var);
  return ret;
}

/*
 * unif_propose_pos:
 * 
 * propose a new positive "ret" based on an old value "last"
 * by proposing uniformly in [last * prop, last / prop], and return
 * the forward and backward probabilities; 
 */

double unif_propose_pos(double last, double *l_q_fwd, double *l_q_bak,
			            double prop)
{
  double left, right, ret;

  /* check for a valid proportion argument */
  assert(last > 0);
  assert(prop > 0 && prop < 1);

  /* propose new d, and compute proposal probability */
  left  = last * prop;
  right = last / prop;
  ret = runif(left, right);  
  *l_q_fwd = -log(right - left);

  /* compute backwards probability */
  *l_q_bak = -log(ret / prop - ret * prop);
  
  if (ret > 10e10) {
     warning("unif_propose_pos (%g) is bigger than max", ret);
     ret = 10;
  }
  return ret;
}


/*
 * unif_propose:
 *
 * random walk on (min, max) with a uniform sliding window
 * of width win.
 */

double unif_propose_bound(double x, double min, double max, double win, double *lqfwd, double *lqbak)
  {
    double xp, l, u;

    assert(x < max);
    assert(x > min);

    /* get lower and upper bounds of the window */
    l = myfmax(min, x-win);
    u = myfmin(max, x+win);
    //Rprintf("l = %f, u = %f, min = %f, max = %f, x = %f, win = %f\n", l, u, min, max, x, win);

    /* draw the proposal */
    xp = runif(l, u);

    /* get the forward (log) proposal probability */
    *lqfwd = dunif(xp, l, u, 1);

    /* get the backward (log) proposal probability */
    l = myfmax(min, xp-win);
    u = myfmin(max, xp+win);
    *lqbak = dunif(x, l, u, 1);

    /* return the proposed parameter */
    return xp;
  }

double unif_propose(double x, double win, double *lqfwd, double *lqbak)
  {
    return unif_propose_bound(x, 0., 1., win, lqfwd, lqbak);
  }

/*
 * unif_propose_prop:
 *
 * uniform random walk on (x * win, x / win)
 * of width win
 */

double unif_propose_prop(double x, double win, double *lqfwd, double *lqbak)
  {
    double xp, l, u;

    assert(win < 1 && win > 0);

    /* get lower and upper bounds of the window */
    l = x * win;
    u = x / win;

    /* draw the proposal */
    xp = runif(l, u);

    /* get the forward (log) proposal probability */
    *lqfwd = dunif(xp, l, u, 1);

    /* get the backward (log) proposal probability */
    l = xp * win;
    u = xp / win;
    *lqbak = dunif(x, l, u, 1);

    /* return the proposed parameter */
    return xp;
  }
/**
 * Sample an integer within [0, 1, 2, ..., max - 1].
 */

int sample_uniform(int max) 
{
    double u = runif(0, 1);
    if (u == 1.) u = 0.999;
    return (int) (u * max);
}

/**
 * Generates an uniformly distributed permutation of {0, 1, ..., n - 1}.
 */

void perm(int *out, int n)
{
    int i, choice, pos;
    int *marked = (int*) malloc(sizeof(int) * n);
    for (i = 0; i < n; i++)
        marked[i] = 0;

    for (i = 0; i < n; i++) {
        choice = sample_uniform(n - i);
        pos = 0;
        do {
           while (marked[pos++]);
           choice--;
        } while (choice >= 0);
        pos--;
        out[i] = pos;
        marked[pos] = 1;
    }
    free(marked);
}

/**
 * Samples from a truncated Gaussian with given upper (u) and lower (l) bounds.
 */

double sample_trunc_gaussian(double mean, double var, double l, double u) 
{
    assert(u > l);
    double std = sqrt(var);
    double fb = pnorm(u, mean, std, 1, 0);
    double fa = pnorm(l, mean, std, 1, 0);
    double v = runif(0, 1) * (fb - fa) + fa;
    return(qnorm(v, mean, std, 1, 0));
}

/**
 * Samples from a truncated Gaussian with a given lower (l) bound.
 */

double sample_trunc_gaussian_l(double mean, double var, double l) 
{
    double std = sqrt(var);
    double fa = pnorm(l, mean, std, 1, 0);
    double v = runif(0, 1) * (1 - fa) + fa;
    return(qnorm(v, mean, std, 1, 0));
}

/*
 * unif_propose:
 *
 * random walk on (min, max) with a centered truncated Gaussian
 * of width win
 */


double truncgaussian_propose_bound(double x, double min, double max, double win, double *lqfwd, double *lqbak)
  {
    double xp;

    assert(win > 0);
    assert(x < max);
    assert(x > min);

    /* draw the proposal */
    xp = sample_trunc_gaussian(x, win, min, max);

    /* get the forward (log) proposal probability (up to a multiplicative)*/
    double diff = xp - x;
    double sq_win = sqrt(win);
    double z = log(pnorm(max, x, sq_win, 1, 0) - pnorm(min, x, sq_win, 1, 0)); //log-normalizing constant
    *lqfwd = -0.5 * diff * diff / win - z - 0.5 * log(win);

    /* get the backward (log) proposal probability */
    double z2 = log(pnorm(max, xp, sq_win, 1, 0) - pnorm(min, xp, sq_win, 1, 0)); //log-normalizing constant
    *lqbak = *lqfwd + z - z2;

    /* return the proposed parameter */
    return xp;
  }

/**
 * Samples a discrete random variable according to table prob with n entries
 */

unsigned int sample_discrete(double prob[], unsigned int n)
{
  double s = runif(0, 1); 
  double acc = 0;
  for (unsigned int i = 0; i < n; i++) {
      acc += prob[i];
      if (acc >= s)
          return i;
  }
  return n - 1;
}
