extern "C"
{
#include "matrix.h"
#include "rhelp.h"
#include "rand.h"
#define dposv dposv_
extern void dposv(char*, int*, int*, double*, int*, double*, int*, int*);
}

#include "structure.h"
#include "marginal.h"
#include "copula.h"
#include "factory.h"

#include <R.h>
#include <Rmath.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <math.h>
#include <fstream>

/*
 * MCMCStructure: constructor
 *
 * Arguments:
 *
 * - n: number of training samples;
 * - p: number of variables;
 * - a, b: parameters for the hyperprior of the Dirichlet process (see Mike West's method)
 * - data: dataset. IT HAS TO BE ORDERED ROW-WISE. THAT IS, 1 SAMPLE OF
 *                  P VARIABLES IS FOLLOWED BY 1 SAMPLE OF THE SAME VARIABLES AND SO ON.
 * - md: array of marginal specifications;
 * - md_size: size of each marginal specification;
 * - copd: (redundant) p * p array of copula specifications;
 * - copd_size: size of each copula specification;
 * - initial_K: initial number of components to initialize the Markov chain;
 * - initial_clustering: initial cluster assignment (0, 1, ..., initial_K - 1) for the data points.
 *                       Hence, the array is assumed to be of size 'n' (or null - see below);
 * - initial_trees: tree-structures that initializes the Markov chain. SET THIS TO NULL
 *                  IF YOU DO NOT WANT TO SET AN INITIALIZATION POINT. If this is NOT null, then
 *                  all the 'initial_****' arrays have to be non- null;
 * - initial_params: initial parameter values to initialize the Markov chain. Currently, only used if
 *                   all copulas are one-parameter copulas. Hence, the array is assumed to be of
 *                   size p^2, even if this is redundant;
 *
 * While the treevariate class defines both a distribution and priors, and methods to sample
 * from the distribution and posteriors, this class is a further distribution over 
 * treevariate distributions by allowing a prior over trees.
 *
 * Notice this requires the specification of all copula pairwise distributions, and not
 * only those referring to a particular tree structure.
 *
 * It is assumed that the order of copulas in 'copd' is given by the order of variables 
 * (x_1, x_2, x_3, ..., x_p) in 'data': that is, copula are ordered as
 * (x_1, x_2), (x1_, x_3), ..., (x_1, x_p), (x_2, x_3), ..., (x_{p - 1}, x_p)
 *
 * SETTING this->M TO ZERO GIVES YOU THE FINITE MIXTURE MODEL WITH 'initial_K' COMPONENTS.
 */

MCMCStructure::MCMCStructure(unsigned int n, unsigned int p, double a, double b, 
                             double *data, int *mars_matrix,
                             double *md,   unsigned int md_size,
                             double *copd, unsigned int copd_size,
                             unsigned int initial_K, int *initial_clustering, Tree **initial_trees, double *initial_params)
{
  this->n = n; this->p = p;    
  this->dataSnapshot = data; this->mars_matrix = mars_matrix;
  this->data = data;
  this->data = new double[this->n * this->p];
  memcpy(this->data, this->dataSnapshot, sizeof(double) * this->n * this->p);
  initializeMissingDataInfo();
  this->md = md; this->copd = copd;
  this->md_size = md_size; this->copd_size = copd_size;
  this->initial_K = initial_K; this->initial_clustering = initial_clustering;
  this->a = a; this->b = b;
  
  this->M = 3;  //Notice: this defines the buffer of Radford Neal's Algorithm 8.
                //If set to zero, implies a finite mixture with initial_K being
                //the number of components in the mixture.

  this->n_test = 0;

  if (initial_trees != NULL) {
      this->random_tree = false;
      this->initial_trees = new Tree*[this->initial_K];
      for (unsigned int i = 0; i < this->initial_K; i++) {
          assert(initial_trees[i]->num_edges == p - 1); //Only connected trees are accepted
          this->initial_trees[i] = initial_trees[i]->clone();
      }           
  } else
      this->random_tree = true;  

  this->useCopulaSlice = false;

  this->all_M = new Marginal*[this->p];	
  this->all_cop = new Copula*[this->p * this->p];
  this->z = new int[this->n]; 

  /* 'marginal_data' will store 'data' in a column-wise way */
  this->marginal_data = new double[this->n * this->p];
  for (unsigned int i = 0; i < this->p; i++)
      for (unsigned int j = 0; j < this->n; j++)
           marginal_data[i * this->n + j] = this->data[j * p + i];

  /* Some flags describing characteristics of the copulas */
  this->gaussian_copula = true;
  this->one_param = true;
  this->quadrature_size = 50;

  /* The initial values for each parameter values are not given in the md/copd arrays:
     they are either generated randomly at the beginning of the chain or by 
     reading them from 'initial_params' */
  this->copula_quadrature = new double*[this->p * this->p];
  for (unsigned int i = 0 ; i < this->p; i++) {
      this->all_M[i] = new_marg(n, &marginal_data[i * this->n], &md[i * this->md_size]);
      for (unsigned int j = i + 1 ; j < this->p; j++) {
        this->all_cop[i * this->p + j] = new_copula(&copd[(i * this->p + j) * this->copd_size]); 
        this->all_cop[j * this->p + i] = this->all_cop[i * this->p + j]; 
        this->copula_quadrature[i * this->p + j] = startCopulaQuadrature(i, j);
        this->copula_quadrature[j * this->p + i] = this->copula_quadrature[i * this->p + j];
        if (copd[(i * this->p + j) * this->copd_size] != 1)
            this->gaussian_copula = false;
        if (this->all_cop[i * this->p + j]->NumParams() > 1)
            this->one_param = false;
        this->total_params += this->all_cop[i * this->p + j]->NumParams();
      }
  }   
  if (initial_trees != NULL && this->one_param) {
      this->initial_params = new double[this->initial_K * this->p * this->p];
      memcpy(this->initial_params, initial_params, sizeof(double) * this->initial_K * this->p * this->p);
  }
  this->verbose = true;  
  this->tree_prop = 1;
  this->do_marg_update = true;
  this->do_indcopula_update = false;

  /* Buffers for storing sufficient statistics */
  this->uv = new double[this->p * this->p];
  this->u_s = new double[this->p];
  this->u_sq = new double[this->p];
  this->u = new double*[this->p];
  this->edge_buffer = new unsigned int[2 * (this->p - 1)]; 
  this->del_score = new double[this->p - 1]; this->add_score = new double[this->p - 1];
}

/**  
 * initializeMissingDataInfo:
 * 
 * Index all rows and variables with missing information, then initialize
 * the values of the missing variables.
 */

void MCMCStructure::initializeMissingDataInfo()
{
  /* Index missing data */

  /* count the number of missing rows */
  this->num_missing_rows = 0;
  for (unsigned int i = 0; i < this->n; i++)
    for (unsigned int j = 0; j < this->p; j++)
      if (this->mars_matrix[i * this->p + j] == 0) {
          this->num_missing_rows++;
          break;
      }
  Rprintf("#Missing rows = %d\n", this->num_missing_rows);
   
  /* nothing to do if there is no missing data */
  if (this->num_missing_rows == 0)
    return;

  /* now go through each row record which rows and columns have missing
     data */

  /* allocate space for missing indices */
  this->missing_rows = new unsigned int[this->num_missing_rows];
  this->missing_columns = new unsigned int*[this->num_missing_rows];
  this->num_missing_cols = new unsigned int[this->num_missing_rows];
  this->num_missing_rows = 0;

  /* loop (twice) over each entry of the data matrix */
  for (unsigned int i = 0; i < this->n; i++) {
    for (unsigned int j = 0; j < this->p; j++) {
      if (this->mars_matrix[i * this->p + j] == 0) {
	
	/* count the number of missing columns within row i */
	this->missing_rows[this->num_missing_rows] = i;
	int missing_col = 0;
	for (unsigned int k = j; k < this->p; k++)
	  if (this->mars_matrix[i * this->p + k] == 0) 
	    missing_col++;
	this->missing_columns[this->num_missing_rows] = new unsigned int[missing_col];

	/* index the missing columns within that row */
	missing_col = 0;
	for (unsigned int k = j; k < this->p; k++)
	  if (this->mars_matrix[i * this->p + k] == 0) 
        this->missing_columns[this->num_missing_rows][missing_col++] = k;
        this->num_missing_cols[this->num_missing_rows] = missing_col;
        this->num_missing_rows++;
        break;
      }
    }
  }
  
  /* Fill in missing data with some perturbed average */
  double *avg = new_zero_vector(this->p);
  int *count = new_zero_ivector(this->p);

  /* replace missing values with the averages */
  /* first sum */
  for (unsigned int i = 0; i < this->n; i++) {
    for (unsigned int j = 0; j < this->p; j++)
      if (this->mars_matrix[i * this->p + j] == 1) {
        avg[j] += this->data[i * this->p + j];
        count[j]++;
      }   
  }
  
  /* then divide by the total count and print */
  for (unsigned int i = 0; i < this->p; i++) {
      avg[i] /= count[i];
      Rprintf("Average of variable %d = %f\n", i, avg[i]);
  }
    
  /* fill in missing values with the jittered averages */
  for (unsigned int i = 0; i < this->num_missing_rows; i++) {
      unsigned int row = this->missing_rows[i];
      for (unsigned int j = 0; j < this->num_missing_cols[i]; j++) {
        unsigned int col = this->missing_columns[i][j];
        this->data[row * this->p + col] = avg[col] + runif(0, 0.2);
      }
  }
  
  /* clean up */
  free(avg); free(count);
}


/**
 * startCopulaQuadrature:
 * 
 * Cache the quadrature points used in the numerical evaluation of the 
 * copulas marginal likelihoods (we always assume copula parameters are
 * independent in the prior).
 * 
 * TODO NOTES: points for the Gaussian copula assume the prior is uniform.
 *             Could be more generic. For the other copulas, a naive
 *             randomized algorithm is used. A smart quadrature method should
 *             be implemented instead.
 */

double* MCMCStructure::startCopulaQuadrature(unsigned int u, unsigned int v)
{
    double *output = new double[this->quadrature_size];
    if (this->gaussian_copula) {
        double step = 2 * 0.999 / (this->quadrature_size - 1);
        for (unsigned int i = 0; i < this->quadrature_size; i++)
            output[i] = -0.999 + i * step;
    } else
        for (unsigned int i = 0; i < this->quadrature_size; i++)
            output[i] = this->all_cop[u * this->p + v]->priorSample();
    return output;
}

/**
 * ~Treevariate:
 * 
 * the usual destructor function .
 */

MCMCStructure::~MCMCStructure()
{
  if (!this->random_tree) {
      for (unsigned int i = 0; i < this->initial_K; i++)
          delete this->initial_trees[i];
      delete[] this->initial_trees;
      delete[] this->initial_params;
  }
  for (unsigned int i = 0; i < this->p; i++) {
     delete this->all_M[i];
     for (unsigned int j = i + 1; j < this->p; j++) {
        delete this->all_cop[i * this->p + j];
        delete[] this->copula_quadrature[i * this->p + j];
     }
  }  
  delete[] this->all_M; delete[] this->all_cop;
  delete[] this->copula_quadrature;
  delete[] this->marginal_data;
  delete[] this->z;
  if (this->num_missing_rows > 0) {
     delete[] this->missing_rows;
     for (unsigned int i = 0; i < this->num_missing_rows; i++)
        delete[] this->missing_columns[i];
     delete[] this->missing_columns;
     delete[] this->num_missing_cols;
  }
  delete[] this->uv; delete[] this->u_s; delete[] this->u_sq; delete[] this->u; 
  delete[] this->edge_buffer; delete[] this->del_score; delete[] this->add_score;
  delete[] this->data;
}


/** 
 * inverse:
 *
 * Invert matrix M (of dimension m x m) and store output in itself.
 */

void MCMCStructure::inverse(double *M, int m)
{
   int n = 2 * m;
   double *A = new double[m * n];
   for (int i = 0; i < m; i++) {
       for (int j = 0; j < m; j++)
           A[i * n + j] = M[i * m + j];
       for (int j = m; j < n; j++)
           A[i * n + j] = 0;
       A[i * n + (m + i)] = 1;
   }

   int i = 0, j = 0;
   while (i < m && j < m) {
      //Rprintf("i = %d, j = %d\n", i, j);
      //Find pivot in column j, starting in row i:
      int maxi = i;
      for (int k = i + 1; k < m; k++)
          if (fabs(A[k * n + j]) > fabs(A[maxi * n + j]))
              maxi = k;
      if (A[maxi * n + j] != 0) { 
          //swap rows i and maxi, but do not change the value of i
          double *temp_row = new double[n];
          for (int v = 0; v < n; v++)
              temp_row[v] = A[i * n + v];
          for (int v = 0; v < n; v++) {
              A[i * n + v] = A[maxi * n + v];
              A[maxi * n + v] = temp_row[v];
          }
          delete[] temp_row;
          //Now A[i,j] will contain the old value of A[maxi,j].
        
          //divide each entry in row i by A[i,j]
          double den = A[i * n + j];
          for (int v = 0; v < n; v++)
              A[i * n + v] /= den;
          //Now A[i,j] will have the value 1.
          for (int u = 0; u < m; u++) {
              if (u == i) continue;
              //subtract A[u,j] * row i from row u
              double *sub = new double[n];
              for (int v = 0; v < n; v++)
                  sub[v] = A[u * n + j] * A[i * n + v];
              for (int v = 0; v < n; v++)
                  A[u * n + v] -= sub[v];
              delete[] sub;
              //Now A[u,j] will be 0, since A[u,j] - A[i,j] * A[u,j] = A[u,j] - 1 * A[u,j] = 0.
          }
          i++;
      }
      j++;
   }

   for (int i = 0; i < m; i++)   
       for (int j = 0; j < m; j++)
           M[i * m + j] = A[i * n + (m + j)];
   delete[] A;
}


/**
 * RoundsInit: 
 * 
 * The output is stored in four different text files.
 * Their content is formatted as follows.
 *
 * FILE ``Z.MCMC'':
 * 
 * The purpose of this file is to store the latent mixture indicators of
 * the Dirichlet process. Each line in the file contains the cluster
 * assignment for each data point in the training set. Hence, each line
 * contains 'p' integers.
 *
 * FILE ``TREES.MCMC'':
 *
 * The purpose of the file is to store each tree sampled in each of the T
 * iterations. All trees for iteration t are stored in the same line, as
 * follows. The very first item in the row is the number of trees 
 * (= number of clusters in the DP). For each tree, first we print its 
 * number of edges. Then, for each edge (i, j), we print 'i' and 'j'.
 * The order of the trees is the order of the cluster labels (0, 1, 2, ...).
 * Notice that some of such trees will not be associated with any point.
 * (However, once in a while we do a relabeling of the clusters and eliminate
 * all clusters with zero points. This doesn't affect previous trees.)
 *
 * FILE ``K.MCMC'':
 *
 * The file simply contains, for each row, the number of DP clusters with
 * at least one point assigned to it. Useful for plotting the statistics of
 * the distribution of the number of mixture components.
 *
 * FILE ``PARAM.MCMC'':
 *
 * The file that stores all sampled parameters, one line for each iteration
 * t = 1, 2, ..., T. All marginal parameters are dumped before any copula
 * parameter. Copula parameters are grouped by mixture component.
 * When an edge doesn't exist in the respective tree, the last value of the
 * respective copula is dumped: care should be taken when interpreting the
 * files, since this value is actually not an actual parameter value.
 *
 * FILE ``MARGINAL.MCMC'':
 *
 * A redundant file that stores only the parameters associated with the marginals.
 *
 * FILE ``ACCEPT_TREE.MCMC'':
 *
 * Tracks the proportion of trees that have been accepted so far at each sampling step.
 *
 * FILE ``COPULA.SEL.MCMC'':
 *
 * The file stores parameters of ``selected copulas'': those associated with the most
 * common cluster.
 *
 * FILE ``PARAM.IMPL.MCMC'':
 *
 * Used only when all copulas are set to be Gaussian. In this case, there is an implied
 * multivariate Gaussian copula with an associated ``implied'' covariance matrix.
 * This file stores some function of the implied covariances of each data point (e.g.,
 * the average of the sampled covariances matrices associated with each training point).
 *
 **
 * Other comments:
 *
 * Also, if the user asked for predictions to be made using a test set and predictions
 * stored in memory, flag 'predict' will be set to 1 (only effective if this->n_test > 0).
 *
 * Analogously, if the user asked for a evaluation of how well the algorithm is doing in
 * filling in missing values, the required data structures will be allocated if 'track_missing' == 1.
 */

void MCMCStructure::RoundsInit(unsigned int T, unsigned int predict_out, unsigned int track_missing)
{
  /* open the output files */
  z_file.open("z.mcmc");
  tree_file.open("trees.mcmc");
  k_file.open("k.mcmc");
  param_file.open("param.mcmc");
  marginal_file.open("marginal.mcmc");
  accepttree_file.open("accept_tree.mcmc");
  sel_copula_file.open("copula.sel.dat");
  impl_gauss_file.open("param.impl.dat");
  /* Print headers */
  for (unsigned int v1 = 0; v1 < this->p; v1++)
      for (unsigned int v2 = v1 + 1; v2 < this->p; v2++) {
          sel_copula_file << "COP" << (v1 + 1) << "." << (v2 + 1) << ".theta1\t";
          impl_gauss_file << "COR" << (v1 + 1) << "." << (v2 + 1) << ".rho\t";
      }
  /*for (unsigned int i = 0; i < 20; i++) //TODO CORRECT THIS
      for (unsigned int v1 = 0; v1 < this->p; v1++)
          for (unsigned int v2 = v1 + 1; v2 < this->p; v2++)
              impl_gauss_file << "COR" << (v1 + 1) << "." << (v2 + 1) << ".rho." << (i + 1) << "\t";*/
  sel_copula_file << "\n";
  impl_gauss_file << "\n";

  /* initialize pointers */
  this->K          = this->initial_K;
  this->tree_mix   = new Tree*[this->K + this->M];
  this->cop_mix    = new Copula*[(this->K + this->M) * (this->p * this->p)];
  this->z_count    = new int[this->K + this->M];
  this->idx        = new set<int>*[this->K];

  /* initialize counts and indices of the components */
  for (unsigned int k = 0; k < this->K; k++) {
      this->z_count[k] = 0;
      this->idx[k] = new set<int>();
  }

  /* randomly assign data points to mixture components */
  if (this->random_tree)
      for (unsigned int i = 0; i < this->n; i++) {
          this->z[i] = sample_uniform(this->K);
          this->z_count[this->z[i]]++;
          this->idx[this->z[i]]->insert(i);
      }
  else {
      for (unsigned int i = 0; i < this->n; i++) {
          this->z[i] = this->initial_clustering[i];
          this->z_count[this->z[i]]++;
          this->idx[this->z[i]]->insert(i);
      }
  }

  /* print the distribution of the initial clusters that result */
  printClusters();

  /* initialize trees in each component cluster */
  /* randomly */
  if (this->random_tree) {
      for (unsigned int k = 0; k < this->K + this->M; k++)
           this->tree_mix[k] = sample_unif_tree(this->p);
  /* read from R */
  } else {
      Rprintf("Using given starting trees...\n");
      for (unsigned int k = 0; k < this->K; k++) 
           this->tree_mix[k] = this->initial_trees[k]->clone();
      for (unsigned int k = this->K; k < this->K + this->M; k++)
           this->tree_mix[k] = sample_unif_tree(this->p);
  }

  /* initialize copulas for all pairs for variables */
  for (unsigned int k = 0; k < this->K + this->M; k++) {
       int offset = k * this->p * this->p;
       for (unsigned int i = 0 ; i < this->p; i++) {
            for (unsigned int j = i + 1 ; j < this->p; j++) {
                 this->cop_mix[offset + i * this->p + j] = new_copula(&copd[(i * this->p + j) * this->copd_size]); 
                 this->cop_mix[offset + j * this->p + i] = this->cop_mix[offset + i * this->p + j]; 
                 this->cop_mix[offset + i * this->p + j]->randomize();
            }
       }
  }
  if (!this->random_tree) {
      Rprintf("Using given starting parameters...\n");
      for (unsigned int k = 0; k < this->K; k++) {
          int offset = k * this->p * this->p;
          for (unsigned int i = 0 ; i < this->p; i++)
              for (unsigned int j = i + 1 ; j < this->p; j++) {
                   this->cop_mix[offset + i * this->p + j]->setParams(&this->initial_params[offset + i * this->p + j]);
                   //Rprintf("Setting parameter [%d, %d; %d] = %f\n", i, j, k, this->initial_params[offset + i * this->p + j]);
              }
      }
      Rprintf("Done with parameter initialization\n");
  }
  
  /* initialize counters for acceptance rates */
  this->accepted_tree = 0; this->accepted_parameter = 0;
 
  /* counters for the number of non-empty DP mixture components */
  this->nonzero_K = 0;
  for (unsigned int k = 0; k < this->K; k++)
      if (this->z_count[k] > 0)
         this->nonzero_K++;

  /* initialize the DP mixture concentration parammeter */
  this->alpha = 0.1;

  /* Initialize predictive log-likelihood data structures if there is a test set */
  if (this->n_test > 0 && predict_out != 1)
      this->predictions = new double[T * this->n_test];

  /* Initialize pseudo training data. We will store the average missing values
     generated by the MCMC algorithm. */
  if (track_missing) {
      this->imputedData = new double[this->n * this->p];
      for (unsigned int i = 0; i < this->n; i++)
           for (unsigned int j = 0; j < this->p; j++)
                this->imputedData[i * this->p + j] = 0.;      
  }
}


/**
 * Rounds:
 *
 * Perform T Metropolis-Hastings within Gibbs rounds. Parameters and 
 * tree-structures are sampled.
 *
 * The output is stored in four different text files that are opened by
 * the RoundsInit() function, and are described there.
 *
 */

void MCMCStructure::Rounds(unsigned int T, unsigned int predict_out, unsigned int track_missing, double output_buffer[])
{
  /* allocate space and open files */
  RoundsInit(T, predict_out, track_missing);

  unsigned int possible_trees = 0;

  /* used as temporary storage for writing to files at each iteration 
   * in WriteTreeFile */
  double *tree_buffer = new double[2 * (this->p - 1) + 1];

    /* now loop through MCMC rounds */
    for (unsigned int t = 0; t < T; t++) {

         /* all the work is done here */
         Rprintf("round %d [%d]\n", t + 1, this->tree_prop);
         step();        
         possible_trees += this->nonzero_K;
            
         /* now that the work is done we need to write stuff to files, skipping steps
            if it is necessary to save disk space */
         if (t % this->mcmc_step == 0) {
             WriteKFile();
             WriteTreeFile(tree_buffer);
             WriteParamsFile();
             WriteZFile();
             if (this->do_marg_update)
                 WriteMarginalFile();
             WriteSelectedCopulaFile(); WriteImpliedGaussianParameterSample();
         }

         WriteAcceptedTreeFile(possible_trees);
         /* store any statistics of sample t that might be returned to R */
         if (this->n_test > 0 && predict_out != 1)
             storePredictiveLogLikelihood(t);

         /* add missing data predictions if required */
         if (track_missing && t > this->burn_in) 
             addMissingDataPrediction();

         /* periodically clean up the mixture components that don't
            have any data points associate with it */
         if (this->M > 0 && t % 100 == 0) {
             Rprintf("Packing clusters...\n");
             packClusters();
         }

         /* print summary information about the iteration */
         printClusters();
         Rprintf("alpha = %f\n", this->alpha);

    }

    /* write closing info */
    Rprintf("#Accepted trees: %f\n", ((double) this->accepted_tree) / possible_trees);

    /* deal with R output */
    if (this->n_test > 0 && predict_out != 1) {
        output_buffer[0] = calculatePredictiveLoglikelihood(T);
        Rprintf("#Average predictive log-likelihood = %f\n", output_buffer[0]);
    }
    if (track_missing == 1) {
        for (unsigned int i = 0; i < this->n; i++)
             for (unsigned int j = 0; j < this->p; j++)
                  this->imputedData[i * this->p + j] /= (double) (T - this->burn_in);
        output_buffer[1] = evaluateMissingValues();
        Rprintf("#Missing data prediction error = %f\n", output_buffer[1]);
    }
    output_buffer[2] = ((double) this->accepted_tree) / possible_trees; /* For comparing samplers */

    /* clean up */
    delete[] tree_buffer;
    RoundsCleanup(predict_out, track_missing);
}


/* 
 * WriteZFile:
 *
 * write the cluster indicators out to the z file
 */

void MCMCStructure::WriteZFile(void)
{
  for (unsigned int i = 0; i < this->n; i++)
    this->z_file << this->z[i] << "\t";
  this->z_file << "\n";
}


/*
 * WriteParamsFile:
 *
 * write the marginal and copula parameters out to a line of the
 * params file
 */

void MCMCStructure::WriteParamsFile(void)
{

  /* first, write the marginal parameters to the file */
 if (this->do_marg_update)
    for (unsigned int v = 0; v < this->p; v++) {            
      double *buffer = new double[this->all_M[v]->NumParams()];
      this->all_M[v]->Params(buffer);
      for (unsigned int pm = 0; pm < this->all_M[v]->NumParams(); pm++)
        param_file << buffer[pm] << "\t";
      delete[] buffer;
    }

  /* and now the copula parameters */

  /* for each mixture component */
  for (unsigned int k = 0; k < this->K; k++) {

    /* for jumping though a 1-d array */
    int offset = k * this->p * this->p;

    /* for each pair of variables */
    for (unsigned int v1 = 0; v1 < this->p; v1++)
      for (unsigned int v2 = v1 + 1; v2 < this->p; v2++) {                

        /* copy the copula object into a allocated double vector buffer */
        double *buffer = new double[this->all_cop[v1 * this->p + v2]->NumParams()];         
        this->cop_mix[offset + v1 * this->p + v2]->Params(buffer);

        /* for each parameter in a copula */
        for (unsigned int pm = 0; pm < this->all_cop[v1 * this->p + v2]->NumParams(); pm++) {

          /* write the parameter to the file */
          param_file << buffer[pm] << "\t";
          /** 
              It would be more interesting to print 'NA' when the parameter is not used by
              the likelihood function (since such parameters are never used for anything). 
              Not doing this yet: the problem is that, when reading the params.mcmc file for
              computing predictive log-likelihoods, it is more convenient to assume that the
              file contains only real numbers. The alternative is illustrated below.
                 (But if this->do_indcopula_update is set to true, then we are all right, 
                  since samples will be generated)
          */
          //if (this->tree_mix[k]->getAdjacency(v1, v2) != 0)
          //    param_file << buffer[pm] << "\t";
          //else
          //    param_file << "NA\t"; 
        }

        /* clean up the buffer */
        delete[] buffer;
    }
  }

  /* finish the line */
  param_file << "\n";
}

/*
 * WriteSelectedCopulaFile:
 *
 * write the copula parameters of the most probable cluster.
 * For this to be meaningful, we have to be sampling parametes
 * for all copulas at all stages (and not only parameters associated
 * with edges in the current tree. Set this->do_indcopula_update = true)
 */
void MCMCStructure::WriteSelectedCopulaFile()
{
    /* Get the maximal cluster */
    int max_k = -1;
    for (int k = 0; k < (int) this->K; k++)
        if (this->z_count[k] > max_k)
            max_k = k;

    /* Dump only the copula parameters from that cluster */
    double par[5];
    int offset = max_k * this->p * this->p;
    for (unsigned int v1 = 0; v1 < this->p; v1++) {
        for (unsigned int v2 = v1 + 1; v2 < this->p; v2++) {	  
            this->cop_mix[offset + v1 * this->p + v2]->Params(par);                             
            double target_param = par[0];
            sel_copula_file << target_param << "\t";
        }
    }
    sel_copula_file << "\n";
}
/*
 * WriteImpliedGaussianParameterSample:
 *
 * Only for the case where all copulas are Gaussian. Computes
 * the implied multivariate Gaussian covariance matrix parameter
 * for each data point, then outputs a function of these covariance
 * matrices.
 */

void MCMCStructure::WriteImpliedGaussianParameterSample()
{
  if (!this->gaussian_copula)
    return;

  /* allocate space */
  double *avg_cov = new double[this->p * this->p];
  int    *path = new int[this->p];
  double *par = new double[1];

  for (unsigned int v1 = 0; v1 < this->p; v1++) {            
    for (unsigned int v2 = 0; v2 < this->p; v2++)
      avg_cov[v1 * this->p + v2] = 0.;
      avg_cov[v1 * this->p + v1] = 1.;
  }

  /* Reads trees and stores implied correlation */
  double *all_impl_corr = new double[this->K * this->p * this->p];

  /* calculate all implied correlations for all choices of k */
  for (int k = 0; k < (int) this->K; k++) {
    Tree *tree_t = tree_mix[k]; 
    int cor_pos = k * this->p * this->p;
    for (unsigned int v1 = 0; v1 < this->p; v1++) {
      for (unsigned int v2 = v1 + 1; v2 < this->p; v2++) {	  
        if (tree_t->adj_matrix[v1 * p + v2] == 0) {
          tree_t->findPath(v1, v2, path);
          this->cop_mix[cor_pos + v1 * this->p + path[0]]->Params(par);                             
          double impl_param = par[0];
          for (unsigned int i = 1; i < this->p; i++) {
            this->cop_mix[cor_pos + path[i - 1] * this->p + path[i]]->Params(par);
            impl_param *= par[0];
            if (path[i + 1] == -1)
              break;
          }
          all_impl_corr[cor_pos + v1 * this->p + v2] = impl_param;
        } else {
          this->cop_mix[cor_pos + v1 * this->p + v2]->Params(par); 
          all_impl_corr[cor_pos + v1 * this->p + v2] = par[0];
        }
      }        
    }
  }

  /* calculating the average covariance matrix  */
  for (int k = 0; k < (int) this->K; k++) {
    int cor_pos = k * this->p * this->p;
    for (unsigned int v1 = 0; v1 < this->p; v1++) 
      for (unsigned int v2 = v1 + 1; v2 < this->p; v2++)
        avg_cov[v1 * this->p + v2] += this->z_count[k] * all_impl_corr[cor_pos + v1 * this->p + v2];
  }
        
  /* Print averaged correlation matrix to the log file  */
  for (unsigned int v1 = 0; v1 < this->p; v1++) 
    for (unsigned int v2 = v1 + 1; v2 < this->p; v2++) {
      avg_cov[v1 * this->p + v2] /= this->n;
      impl_gauss_file << avg_cov[v1 * this->p + v2] << "\t";
    }                
  impl_gauss_file << "\n";

  /* Print implied correlation matrix of the most probable cluster to the log file */
  /*unsigned int max_k = -1;
  double max_count = -1;
  for (int k = 0; k < (int) this->K; k++)
      if (this->z_count[k] > max_count) {
          max_k = k;
          max_count = this->z_count[k];
      }
  int cor_pos = max_k * this->p * this->p;
  for (unsigned int v1 = 0; v1 < this->p; v1++) 
      for (unsigned int v2 = v1 + 1; v2 < this->p; v2++)
          impl_gauss_file << all_impl_corr[cor_pos + v1 * this->p + v2] << "\t";
  impl_gauss_file << "\n";*/

  /* Print implied correlation matrix for the first 10 data points to the log file */
  /*for (unsigned int i = 0; i < 20; i++) { //TODO CORRECT THIS
      int cor_pos = this->z[i] * this->p * this->p;
      for (unsigned int v1 = 0; v1 < this->p; v1++) 
          for (unsigned int v2 = v1 + 1; v2 < this->p; v2++)
              impl_gauss_file << all_impl_corr[cor_pos + v1 * this->p + v2] << "\t";
  }
  impl_gauss_file << "\n";*/

  /* Clean-up */
  delete[] all_impl_corr;
  delete[] avg_cov;
  delete[] path;
  delete[] par;
}

/*
 * WriteMarginalFile:
 *
 * write the marginal parameters out to a line of the
 * marginals file
 */

void MCMCStructure::WriteMarginalFile(void)
{
  /* write the marginal parameters to the file */
  for (unsigned int v = 0; v < this->p; v++) {            
    double *buffer = new double[this->all_M[v]->NumParams()];
    this->all_M[v]->Params(buffer);
    for (unsigned int pm = 0; pm < this->all_M[v]->NumParams(); pm++)
      marginal_file << buffer[pm] << "\t";
    delete[] buffer;
  }
  /* finish the line */
  marginal_file << "\n";
}

/*
 * WriteKFile:
 *
 * write the current number of (non-empty) clusters to the file 
 */

void MCMCStructure::WriteKFile(void)
{
  /* write to the k file */
  k_file << this->nonzero_K << "\t" << this->alpha << "\n";
}

/*
 * WriteTreeFile:
 *
 * copies the adjacency matrix to the file tree_file 
 */

void MCMCStructure::WriteTreeFile(double *tree_buffer)
{
  tree_file << this->K << "\t";
  for (unsigned int k = 0; k < this->K; k++)  {
    this->tree_mix[k]->copyAdjacencies(tree_buffer);
    int num_edges = (int) tree_buffer[0];
    tree_file << num_edges << "\t";
    for (int e = 0; e < num_edges; e++)
      tree_file << ((int) tree_buffer[1 + e * 2]) << "\t" << ((int) tree_buffer[1 + e * 2 + 1]) << "\t";
  }
  tree_file << "\n";
}


/*
 * WriteAcceptedTreeFile:
 *
 * copies the current proportion of accepted trees to the file accepttree_file
 */

void MCMCStructure::WriteAcceptedTreeFile(int possible_trees)
{
   this->accepttree_file << ((double) this->accepted_tree) / possible_trees << "\n"; 
}
/*
 * RoundsCleanup:
 *
 * close files and free storage initialized by the RoundsInit() function
 * and used by the Rounds function
 */

void MCMCStructure::RoundsCleanup(unsigned int predict_out, unsigned int track_missing)
{
    z_file.close();
    tree_file.close(); 
    k_file.close();
    param_file.close();
    marginal_file.close();
    accepttree_file.close();
    sel_copula_file.close();
    impl_gauss_file.close();

    /* Clean variables */
    for (unsigned int k = 0; k < this->K; k++)
        delete this->idx[k];
    delete[] this->idx;
    delete[] this->z_count;
    for (unsigned int k = 0; k < this->K + this->M; k++) {
        delete this->tree_mix[k];
        int offset = k * this->p * this->p;
        for (unsigned int i = 0 ; i < this->p; i++)
            for (unsigned int j = i + 1 ; j < this->p; j++)
                delete this->cop_mix[offset + i * this->p + j];
    }
    delete[] this->tree_mix; delete[] this->cop_mix;
    if (this->n_test > 0 && predict_out != 1)
        delete[] this->predictions;
    if (track_missing)
       delete[] this->imputedData;
}


/**
 * Step:
 *
 * MCMC round for sampling from the Bayesian posterior
 * for the treevariate copula, including tree structure.
 */

void MCMCStructure::step()
{
    //Update Dirichlet process parameter
    //Rprintf("Update alpha...\n");
    updateAlpha();

    //Update component indicators
    //Rprintf("Update Z...\n");
    updateZ();

    //Update tree structures by Metropolis--Hastings
    //Rprintf("Update structure...\n");
    for (unsigned int i = 0; i < this->K; i++)
        updateStructure(i); 

    //Update marginal parameters
    //Rprintf("Update marginals...\n");
    if (this->do_marg_update)
      for (unsigned int v = 0; v < this->p; v++)
           updateMarginals(v);

    //Update each of the copula parameters
    //Rprintf("Update copulas...\n");
    for (unsigned int k = 0; k < this->K; k++)
        for (unsigned int v1 = 0; v1 < this->p; v1++)
            for (unsigned int v2 = v1 + 1; v2 < this->p; v2++)
                updateCopulas(k, v1, v2);

    //Finally, sample missing variables if any (Data Augmentation)
    //Rprintf("Update missing values...\n");
    if (this->num_missing_rows > 0)
        updateMissingData();
}


/**
 * updateMarginals:
 *
 * Sample parameters for marginal 'v'.
 */

void MCMCStructure::updateMarginals(unsigned int v) 
{
  /* total number of parameters to update for marginal v */
  unsigned int nump = this->all_M[v]->NumParams();

  /* iterate through the parameters via which */
  for (unsigned int which = 1; which <= nump; which++) {

    /* get the v-th marginal */
    Marginal *M = this->all_M[v];

    /* make a random walk proposal for the which-th parameter */
    double lqfwd, lqbak;
    Marginal* Mprime = M->RW(which, &lqfwd, &lqbak);

    /* (log) likelihood ratio of the proposed marginal */
    double lalpha = Mprime->LLik() - M->LLik();

    //double temp_params1[2], temp_params2[2];
    //Mprime->Params(temp_params1); M->Params(temp_params2);
    //Rprintf(" - (m1 = %f, s1 = %f), (m2 = %f, s2 = %f), diff = %f, %f" , temp_params1[0], temp_params1[1], temp_params2[0], temp_params2[1], Mprime->LLik(), M->LLik()); 

    /* add in (log) likelihood ratio of copulas involving v */
    for (unsigned int v2 = 0; v2 < this->p; v2++) {
      
      /* besides itself */
      if (v == v2) continue;
      
      /* for each (tree) component in the DP mixture */
      for (unsigned int k = 0; k < this->K; k++) {
        /* for all neighbors of v in that tree component */
        if (this->tree_mix[k]->getAdjacency(v, v2) == 0) continue;
        	
        /* get the marginal of the adjacent vertex */
        Marginal *Mother = this->all_M[v2];

        /* get the copula between v and v2 in the k-th tree */
        int offset = k * this->p * this->p;
        Copula *c = this->cop_mix[offset + v * this->p + v2];

        /* log likelihood numerator of the proposed Marginal */
        lalpha += c->LLik(Mprime, Mother, this->idx[k]);

        /* log likelihood denominator of the old Marginal */
        lalpha -= c->LLik(M, Mother, this->idx[k]); 

        //double diff = c->LLik(Mprime, Mother, this->idx[k]) - c->LLik(M, Mother, this->idx[k]);
        /*if (fabs(diff) > 1) {
            double rho;
            c->Params(&rho);
            Rprintf(" - diff = %f, k = %d, rho = %f, size = %d" , diff, k, rho, this->idx[k]->size()); 
        }*/
      }

    }

    /* (log) prior ratio */
    lalpha += Mprime->LPrior() - M->LPrior();

    /* proposal ratio */
    lalpha += lqbak - lqfwd;

    /* accept or reject the proposed Mprime */
    if (runif(0, 1) < exp(lalpha)) {
        delete this->all_M[v];
        this->all_M[v] = Mprime;
        //Rprintf(" - ACCEPTED (alpha = %f)\n", exp(lalpha));
    } else {
        delete Mprime;
        //Rprintf(" - REJECTED (alpha = %f)\n", exp(lalpha));
    }
  }
}


/**
 * updateCopulas (random walk Metropolis-Hastings or slice sampling):
 *
 * Sample parameters for copula ('v1', 'v2') within mixture component 'k'.
 */

void MCMCStructure::updateCopulas(unsigned int k, unsigned int v1, unsigned int v2)
{
  if (this->useCopulaSlice)
     updateCopulasSlice(k, v1, v2);
  else
     updateCopulasMH(k, v1, v2);
}

void MCMCStructure::updateCopulasMH(unsigned int k, unsigned int v1, unsigned int v2)
{
  /* check that the edge between v1 and v2 is in the k-th tree; if not
     then nothing to do here, unless explicitly indicated by do_indcopula_update */
    
  if (this->tree_mix[k]->getAdjacency(v1, v2) == 0) {
     if (this->do_indcopula_update) {
        int offset = k * this->p * this->p;
        this->cop_mix[offset + v1 * this->p + v2]->randomize();
     }
     return;
  }

  /* get a pointer to the copula between v1 and v2 in the k-th tree */
  int offset = k * this->p * this->p;
  Copula *c = this->cop_mix[offset + v1 * this->p + v2];

  /* get the number of paramters in this copula */
  unsigned int nump = c->NumParams();

  /* foreach parameter to this copula */
  for (unsigned int which = 1; which <= nump; which++) {

    /* make a random walk proposal for the which-th parameter */
    double lqfwd, lqbak;
    Copula* cprime = c->RW(which, &lqfwd, &lqbak);
  
    /* get the marginals of v1 and v2 */
    Marginal *Mx = this->all_M[v1];
    Marginal *My = this->all_M[v2];

    /* log likelihood ratio of the proposed and old copula densities */
    double lalpha = cprime->LLik(Mx, My, this->idx[k]) - c->LLik(Mx, My, this->idx[k]);

    /* add in the (log) prior ratio */
    lalpha += cprime->LPrior() - c->LPrior();

    /* add in the (log) proposal ratio */
    lalpha += lqbak - lqfwd;

    /* accept or reject the proposed Cprime */        
    if (runif(0, 1) < exp(lalpha)) {
      delete this->cop_mix[offset + v1 * this->p + v2];
      this->cop_mix[offset + v1 * this->p + v2] = cprime;
      this->cop_mix[offset + v2 * this->p + v1] = cprime;
    } else
      delete cprime;

  }
}

void MCMCStructure::updateCopulasSlice(unsigned int k, unsigned int v1, unsigned int v2)
{
  /* check that the edge between v1 and v2 is in the k-th tree; if not
     then nothing to do here, unless explicitly indicated by do_indcopula_update */    
  if (this->tree_mix[k]->getAdjacency(v1, v2) == 0) {
     if (this->do_indcopula_update) {
        int offset = k * this->p * this->p;
        this->cop_mix[offset + v1 * this->p + v2]->priorSample();
     }
     return;
  }

  /* get a pointer to the copula between v1 and v2 in the k-th tree */
  int offset = k * this->p * this->p;
  Copula *c = this->cop_mix[offset + v1 * this->p + v2];

  /* get the number of paramters in this copula */
  unsigned int nump = c->NumParams();

  /* foreach parameter to this copula */
  for (unsigned int which = 1; which <= nump; which++) {

    /* get the marginals of v1 and v2 */
    Marginal *Mx = this->all_M[v1];
    Marginal *My = this->all_M[v2];

    /* slice-sample the which-th parameter */
    Copula* cprime = c->sliceSample(which, Mx->Ranks(), My->Ranks(), this->idx[k]);
    delete this->cop_mix[offset + v1 * this->p + v2];
    this->cop_mix[offset + v1 * this->p + v2] = cprime;
    this->cop_mix[offset + v2 * this->p + v1] = cprime;

  }
}

/**
 * updateMissingData:
 *
 * Sample all necessary missing values. Notice that the samplewd value is based 
 * on the type of marginal associated with the variable.
 */

void MCMCStructure::updateMissingData()
{
  /* iterate over all rows that have missing variables */
  for (unsigned int i = 0; i < this->num_missing_rows; i++) {
    unsigned int row = this->missing_rows[i];

    /* iteratate over all variabes that are missing in that particular row */
    for (unsigned int j = 0; j < this->num_missing_cols[i]; j++) {
      unsigned int col = this->missing_columns[i][j];

      /* get the mixture component (tree) associated with that data point */
      Tree *tree = this->tree_mix[this->z[row]];

      /* get the marginal and copula information for that data point */

      /* the old imputed value and rank of this missing data point */
      double value = this->data[row * this->p + col];       
      double value_rank = this->all_M[col]->GetRank(row);

      /* make a random walk proposal for a new imputed value and comput its rank */
      double lqfwd, lqbak;
      double value_prime = this->all_M[col]->RWvalue(value, &lqfwd, &lqbak);
      double value_prime_rank = this->all_M[col]->CalcRanks(value_prime);

      /* calculate the log likelihood ratio of the marginals */
      double lalpha = this->all_M[col]->LLik(value_prime) - this->all_M[col]->LLik(value);

      /* for each copula that includes the variable corresponing to the 
	     column that the missing data point is in */
      int offset = this->z[i] * this->p * this->p;
      for (unsigned int v = 0; v < tree->adj_list[col].size(); v++) {
        int other = tree->adj_list[col][v];
        double other_rank = this->all_M[other]->GetRank(row);
        Copula *c = this->cop_mix[offset + col * this->p + other];
    	
        /* calculate the log likelihood ratio */
        lalpha += c->LLik(value_prime_rank, other_rank);
        lalpha -= c->LLik(value_rank, other_rank);
      }

      /* add in the log proposal ratio */
      lalpha += lqbak - lqfwd;

      /* accept or reject */
      if (runif(0, 1) < exp(lalpha)) { /* accept */
        this->data[row * this->p + col] = value_prime;
        this->marginal_data[col * this->n + row] = value_prime;
        this->all_M[col]->SetRank(row, value_prime_rank); 
      }
    }
  }
}

/**
 * addMissingDataPrediction:
 *
 * After obtaining predicted missing values, add them to a buffer of predictions. 
 * This statistic will be used later to evaluate the quality of the 
 * missing value predictions if we happen to know the true values (for evaluation
 * purposes).
 */

void MCMCStructure::addMissingDataPrediction()
{
  /* iterate over all rows that have missing variables */
  for (unsigned int i = 0; i < this->num_missing_rows; i++) {
       unsigned int row = this->missing_rows[i];

       /* iterate over all variabes that are missing in that particular row */
       for (unsigned int j = 0; j < this->num_missing_cols[i]; j++) {
            unsigned int col = this->missing_columns[i][j];
      
            /* add the imputed value to imputedData */
            double value = this->data[row * this->p + col];       
            this->imputedData[row * this->p + col] += value;
       }
  }
}

/**
 * updateAlpha:
 *
 * Update the Dirichlet process parameter by using the method of
 * Mike West.
 */

void MCMCStructure::updateAlpha()
{
    if (this->M == 0) return; //Finite, non-DP, version
    double eta  = rbeta(this->alpha + 1., (double) this->n);
    double odds = (this->a + this->nonzero_K - 1) / (this->n * (this->b - log(eta)));
    double pi   = odds / (1. + odds);
    if (runif(0, 1) < pi)
        this->alpha = rgamma(this->a + this->nonzero_K, 1. / (this->b - log(eta)));
    else
        this->alpha = rgamma(this->a + this->nonzero_K - 1, 1. / (this->b - log(eta)));
}


/**
 * resampleParameters:
 *
 * Sample parameters for mixture component 'k' (useful when
 * a new component is created).
 */

void MCMCStructure::resampleParameters(unsigned int k)
{
    int offset = k * this->p * this->p;
    for (unsigned int i = 0; i < this->p; i++) 
        for (unsigned int j = i + 1; j < this->p; j++)
            this->cop_mix[offset + i * this->p + j]->randomize();
}




/**
 * Neal's ``Algorithm 8''.
 *
 * IT IS ASSUMED THAT THE TREES ARE GIVEN AN UNIFORM PRIOR.
 * (since it's easy to sample from).
 */

void MCMCStructure::updateZ()
{   
  /* Initialize variables that keep track of whether a new cluster had to be created */
  bool K_changed = false;

  /* Initialize array containing probability of each mixture (tree) component */
  double *p = new double[this->K + this->M];

  /* Initialize parameters of the auxiliary variables of Neal's algorithm */

  for (unsigned int m = 0; m < this->M; m++) {
    delete this->tree_mix[this->K + m];
    this->tree_mix[this->K + m] = sample_unif_tree(this->p);
    resampleParameters(K + m);
    this->z_count[this->K + m] = 0;
  }
    
  /* Resample mixture indicator for each of the n data points */

  for (unsigned int i = 0; i < this->n; i++) {

    /* Data point i was the only data point in its mixture component.
       Component will lose its only element, and data point i will
       be assigned to a new component. */

    if (this->z_count[this->z[i]] == 1 && this->M > 0) {
      delete this->tree_mix[this->K];
      this->tree_mix[this->K] = this->tree_mix[this->z[i]]->clone();
      this->z_count[this->z[i]] = 0;
      this->idx[z[i]]->erase(i);
      this->z[i] = this->K;
      this->z_count[this->K] = 1;            
    }

    computeComponentProbability(i, p);

    /* Physically remove data point i from its current mixture component. */

    this->z_count[z[i]]--;
    if (this->z[i] < (int) this->K)
      this->idx[z[i]]->erase(i);

    /* Variables used to decide in the multinomial sampling procedure that
       decides which mixture component to assign data point i to. */

    double u = runif(0, 1), acc = 0;
        
    /* Iterate through all mixture components (including auxiliary variables),
       stopping when finding the component k that corresponds to the sampled
       cumulative distribution u. */

    unsigned int max_k = this->K + this->M;
    bool updated = false;
    for (unsigned int k = 0; k < max_k; k++) {
      if (u < acc + p[k]) { /* Finding point k that corresponds to sampled probability u. */
                
        /* If chosen mixture component is none of the existing ones,
           then without loss of generality we can set it to be the first 
           auxiliary variable. This will make our implementation simpler
           when increasing the physical array of mixture components. */

        unsigned int cluster_choice;
        if (k >= this->K) {
          cluster_choice = this->K;
          if (k != this->K) {
            /* Make sure to copy the right tree to position K if
               the chosen mixture was not the first one. */
            delete this->tree_mix[this->K];
            this->tree_mix[this->K] = this->tree_mix[k]->clone();
          }
        }
        else
          cluster_choice = k;

        /* Update mixture component indicators. */
 
        this->z[i] = cluster_choice; updated = true;
        this->z_count[cluster_choice]++;

        if (cluster_choice >= this->K) {
          /* Adding a new component with data point i. */
          addClusters(this->K + 1, &i);
          K_changed = true;
          delete[] p;
          p = new double[this->K + this->M];                    
        } else {
          /* Making sure we index data point i correctly. */
          for (unsigned int m = 0; m < this->M; m++) {
            /* Sample new parameters and tree for auxiliary variables. */
            delete this->tree_mix[this->K + m];
            this->tree_mix[this->K + m] = sample_unif_tree(this->p);
            resampleParameters(K + m);
            /* Update counter for auxiliary variables. */
            this->z_count[this->K + m] = 0;
          }
          this->idx[this->z[i]]->insert(i);
        }
        break;
      }
      acc += p[k];
    }

    /* This might happen if the model is really bad (or badly initialized):
       the respective p[] array is full of NaNs because of zero-likelihood */

    if (!updated) {
      this->z[i] = (int) trunc(runif(0, 1) * (this->K - 1));
      this->z_count[this->z[i]]++;
      for (unsigned int m = 0; m < this->M; m++) {
        delete this->tree_mix[this->K + m]; this->tree_mix[this->K + m] = sample_unif_tree(this->p);
        resampleParameters(K + m); this->z_count[this->K + m] = 0;
      }
      this->idx[this->z[i]]->insert(i);
      Rprintf("WARNING: cluster value of %d was not updated! Setting it to %d randomly!\n", i, this->z[i]);
    }
  }

  /* Clean-up. */
  delete[] p; 

  /* Update count of non-zero mixture components. */

  this->nonzero_K = 0;
  for (unsigned int k = 0; k < this->K; k++)
    if (this->z_count[k] > 0)
      this->nonzero_K++;
}

/* 
 * computeComponentProbability:
 *
 * For data point i, compute the probability of each mixture
 * component (including possible new components) according to the
 * given Dirichlet process mixture prior (or finite mixture prior).
 */

void MCMCStructure::computeComponentProbability(int i, double *p) 
{
  double sum = 0., max = 0.;
  bool set_max = true;

  /* Calculate the (un-normalized) probability of each existing mixture component. */

  for (int k = 0; k < (int) this->K; k++) {

    /* Ignore mixture component if it is not associated with any data point. */

    if (this->z_count[k] == 0 && this->M > 0) {
      p[k] = 0.;
      continue;
    }

    /* Add log-likelihood of the data point according to the parameters 
       (and tree) associated with component k. */

    p[k] = pointLogLikelihood(i, k);

    /* Add log-prior (DP prior or finite mixture prior, depending on the
       number M of auxiliary variables). */

    if (this->M > 0) {
      /* Dirichlet process mixture prior */
      if (k != this->z[i])
        p[k] += log(this->z_count[k] / (this->n - 1 + this->alpha));
      else
        p[k] += log((this->z_count[k] - 1.) / (this->n - 1 + this->alpha));             
    } else {
      /* Finite mixture prior */
      if (k != this->z[i])
        p[k] += log((this->alpha / this->K + this->z_count[k]) / (this->n - 1 + this->alpha));
      else
        p[k] += log((this->alpha / this->K + this->z_count[k] - 1.) / (this->n - 1 + this->alpha));
    }

    /* Find maximum (un-normalized) log-probability (to re-scale probabilities without
       numerical problems later on). */

    if (set_max || p[k] > max) {
      max = p[k];
      set_max = false;
    }
  }

  /* Calculate the (un-normalized) probability of the auxiliary mixture 
     components of Neal's algorithm. */

  for (unsigned int m = 0; m < this->M; m++) {
    p[this->K + m] = pointLogLikelihood(i, this->K + m) +
                     log((this->alpha / this->M) / (this->n - 1 + this->alpha));
    if (p[this->K + m] > max)
      max = p[this->K + m];
  }

  /* Compute normalized probabilities (subtracting maximum term for the sake
     of numerical stability). */

  for (unsigned int k = 0; k < this->K + this->M; k++)
    if (this->z_count[k] > 0 || k >= this->K || this->M == 0) {
      p[k] = exp(p[k] - max);
      sum += p[k];
    }
  for (unsigned int k = 0; k < this->K + this->M; k++) 
    if (this->z_count[k] > 0 || k >= this->K || this->M == 0)
      p[k] /= sum;
}

/**
 * printClusters:
 *
 * A method to print the distribution of points over clusters.
 */

void MCMCStructure::printClusters()
{
    int zero_count = 0;
    Rprintf("#CLUSTERS = %d [ ", this->K);
    for (unsigned int k = 0; k < this->K; k++) {
        if (this->z_count[k] != 0)
            Rprintf("%d ", this->z_count[k]);
        else 
            zero_count++;
    }
    Rprintf("] (%d of them are empty)\n", zero_count);
}


/**
 * packClusters:
 *
 * Rearrange clusters such that all non-zero clusters are 
 * physically eliminated with the respective memory freed.
 */

void MCMCStructure::packClusters()
{
  /* Create a mapping to map each data point to a re-indexed cluster,
     where we skip mixture components with zero elements. Surviving
     mixture components are reindex by array reindexed, which maps their
     indices to the new index where we skipt the empty mixture components. */

  int *reindex = new int[this->K];
  int *map = new int[this->K];
  int pos = 0;

  for (unsigned int k = 0; k < this->K; k++)
    if (this->z_count[k] > 0) {
      map[k] = pos;
      reindex[pos] = k;
      pos++;
    }

  /* First verify if there is any empty mixture component. If
     there is none, then there is nothing to be done. */
  this->nonzero_K = pos;
  if (this->nonzero_K == this->K) {
      delete[] reindex; delete[] map;
      return;
  }

  /* At this point, there are components with zero elements, which
     will be eliminated. What follows is a rearrangement of the arrays
     z_count, idx, tree_mix and cop_mix such that information about
     non-zero components will be copied to new array while bypassing
     empty components. */

  /* Copy current points to backup arrays old_*. Allocate memory to
     the new (smaller) arrays. */

  int *old_z_count = this->z_count;
  this->z_count = new int[this->nonzero_K + this->M];
  set<int> **old_idx = this->idx;
  this->idx = new set<int>*[this->nonzero_K];
  Tree **old_tree_mix = this->tree_mix;
  this->tree_mix = new Tree*[this->nonzero_K + this->M];
  Copula **old_cop_mix = this->cop_mix;
  this->cop_mix = new Copula*[(this->nonzero_K + this->M) * this->p * this->p];

  /* Copy information from the old arrays to the new ones, skipping
     empty mixture components (those for which old_z_count[k] == 0). */

  pos = 0;
  for (unsigned int k = 0; k < this->K; k++) {
    if (old_z_count[k] > 0) {
      /* Surviving component: copy information. */
      copy_from_old(k, pos, old_z_count, old_tree_mix, old_cop_mix);
      this->idx[pos] = old_idx[k];
      pos++;
    } else {
      /* Empty component. Free the respective memory. */
      delete old_tree_mix[k];
      delete old_idx[k];
      int offset = k * this->p * this->p;
      for (unsigned int i = 0; i < this->p; i++)
          for (unsigned int j = i + 1; j < this->p; j++)
	          delete old_cop_mix[offset + i * this->p + j];
    }
  }

  /* Move allocated auxiliary variables to the end of the newly packed
     arrays of (non-empty) mixture components. */
  for (unsigned int m = 0; m < this->M; m++)
       copy_from_old(this->K + m, this->nonzero_K + m, old_z_count, old_tree_mix, old_cop_mix);

  /* Finally, update the indicator variables according to the
     new index system. */
  for (unsigned i = 0; i < this->n; i++)
    this->z[i] = map[this->z[i]];

  /* Delete old arrays of pointers, now that we freed the objects
     pointed by them. */
  delete[] old_z_count;
  delete[] old_idx;
  delete[] old_tree_mix; delete[] old_cop_mix;

  /* Clean up temporary arrays. */
  delete[] reindex; delete[] map;    

  /* Update number of current components. */
  this->K = this->nonzero_K;
}


/*
 * copy_from_old:
 *
 * copy components of the un-packed data data structures to the
 * packed data structure from cluster k to cluster pos
 */

void MCMCStructure::copy_from_old(unsigned int k, unsigned int pos,
				  int *old_z_count, 
				  Tree **old_tree_mix, Copula **old_cop_mix)
{
  this->z_count[pos] = old_z_count[k];
  this->tree_mix[pos] = old_tree_mix[k];
  int offset = pos * this->p * this->p, offset_k = k * this->p * this->p;
  for (unsigned int i = 0; i < this->p; i++)
    for (unsigned int j = i + 1; j < this->p; j++) {
      this->cop_mix[offset + i * this->p + j] = old_cop_mix[offset_k + i * this->p + j];
      this->cop_mix[offset + j * this->p + i] = old_cop_mix[offset_k + i * this->p + j];
    }
}


/**
 * AddClusters:
 *
 * Number of clusters (with zero or positive counts) increased.
 * We have to resize the data structures accordingly. The new
 * number of clusters is passed, as well as.
 *
 * We need to also update the data structure containing the trees
 * This assumes that all new trees (usually just one) to be kept
 * are located in positions tree_mix[this->K], tree_mix[this->K + 1],
 * ..., tree_mix[new_K - 1].
 *
 */

void MCMCStructure::addClusters(unsigned int new_K, unsigned int *data_points)
{
    assert(new_K > this->K && new_K <= this->K + this->M);

    /* All of these updates consist of adding an extra element to the
       end of the array. First we copy the current pointer to an old
       pointer (e.g., old_z_count). Second, we allocate a new array
       with one extra position. Then we copy information from the old
       array into the new array. Finally, we free the memory associated
       with the old pointers. */

    /* Update z_count */

    int *old_z_count = this->z_count;
    this->z_count = new int[new_K + this->M];
    memcpy(this->z_count, old_z_count, sizeof(int) * new_K);
    for (unsigned int k = new_K; k < new_K + this->M; k++)
        this->z_count[k] = 0;
    delete[] old_z_count;

    /* Update idx */

    set<int> **old_idx = this->idx;
    this->idx = new set<int>*[new_K];
    for (unsigned int k = 0; k < this->K; k++)
        this->idx[k] = old_idx[k];
    for (unsigned int k = this->K; k < new_K; k++) {
        this->idx[k] = new set<int>();
        this->idx[k]->insert(data_points[k - this->K]);
    }
    delete[] old_idx;

    /* Update tree */

    Tree **old_tree_mix = this->tree_mix;
    this->tree_mix = new Tree*[new_K + this->M];
    for (unsigned int k = 0; k < new_K; k++)
        this->tree_mix[k] = old_tree_mix[k];
    for (unsigned int k = new_K; k < new_K + this->M; k++)
        this->tree_mix[k] = sample_unif_tree(this->p);
    for (unsigned int k = new_K; k < this->K + this->M; k++)
        delete old_tree_mix[k];
    delete[] old_tree_mix;

    /* Update copulas. */

    Copula **old_cop_mix = this->cop_mix;
    this->cop_mix = new Copula*[(new_K + this->M) * this->p * this->p];
    
    /* copying the old copulas */
    
    for (unsigned int k = 0; k < new_K; k++) {
        int offset = k * this->p * this->p;
        for (unsigned int i = 0; i < this->p; i++)
            for (unsigned int j = i + 1; j < this->p; j++) {
                this->cop_mix[offset + i * this->p + j] = this->cop_mix[offset + j * this->p + i] = 
                                                          old_cop_mix[offset + i * this->p + j];
            }
    }

    /* initializing new copulas */
    for (unsigned int k = new_K; k < new_K + this->M; k++) {
        int offset = k * this->p * this->p;
        for (unsigned int i = 0; i < this->p; i++)
            for (unsigned int j = i + 1; j < this->p; j++) {
                this->cop_mix[offset + i * this->p + j] = new_copula(&copd[(i * this->p + j) * this->copd_size]); 
                this->cop_mix[offset + j * this->p + i] = this->cop_mix[offset + i * this->p + j]; 
                this->cop_mix[offset + i * this->p + j]->randomize();
            }
    }

    /* deleting the old copulas */
    for (unsigned int k = new_K; k < this->K + this->M; k++) {
        int offset = k * this->p * this->p;
        for (unsigned int i = 0; i < this->p; i++)
            for (unsigned int j = i + 1; j < this->p; j++) 
                delete old_cop_mix[offset + i * this->p + j];
    }
    delete[] old_cop_mix;

    /* Finally, make sure we also update the number of mixture components. */
    this->K = new_K;
}


/**
 * pointLogLikelihood:
 *
 * Calculates the log-likelihood for the d-th datapoint under
 * the k-th mixture component
 */

double MCMCStructure::pointLogLikelihood(unsigned int d, unsigned int k)
{
  double output = 0.; 
  int offset = k * this->p * this->p;

  /* add in the likelihood of the marginals */
  for (unsigned int v = 0; v < this->p; v++) {
       output += this->all_M[v]->LLik(this->data[d * this->p + v]);
       //if (isinf(this->all_M[v]->LLik(this->data[d * this->p + v])))
       //   Rprintf("point log[%d, %f] = %f\n", v, this->data[d * this->p + v], output);
  }
    
  /* add in the likelihood of each of the copulas */
  Tree *tree = this->tree_mix[k];    
  for (unsigned int e = 0; e < tree->num_edges; e++) {
       int v1 = tree->edges[e * 2], v2 = tree->edges[e * 2 + 1];        
       Copula *c = this->cop_mix[offset + v1 * this->p + v2];
       output += c->LLik(this->all_M[v1], this->all_M[v2], d);

       /* print if there are problems */
       if (isnan(output)) {
           int dpi = d * this->p;
           Rprintf("NaN: edge = (%d, %d), v1 = %f, u = %f, v2 = %f, v = %f, rho = %f, point = %d\n", v1, v2, 
               this->data[dpi + v1], this->all_M[v1]->GetRank(d), this->data[dpi + v2],this->all_M[v2]->GetRank(d), c->getParam(1), d);
       }
  }    
  return output;
}

/**
 * updateStructure:
 * 
 * Updates tree structure in cluster k.
 */

void MCMCStructure::updateStructure(unsigned int k) {  
  if (this->p <= 2)
      return;
  if (!this->gaussian_copula) {
      updateStructureStandard(k); //For now 
      return;
  }
  if (this->one_param && this->tree_prop == 0)
      updateStructureOneParam(k);
  else if (this->one_param && this->tree_prop == 2)
      updateStructureOneParamHybrid(k);
  else if (this->tree_prop == 1)
      updateStructureStandard(k);
  else if (this->tree_prop == 3)
      updateStructureExact(k);
  else
      updateStructureRandomized(k);
}

void MCMCStructure::updateStructureRandomWalk(unsigned int k) {
  if (this->one_param && this->tree_prop == 0)
      updateStructureOneParam(k);
  else if (this->one_param && this->tree_prop == 2)
      updateStructureOneParamHybrid(k);
  else
      updateStructureStandard(k);
}

/**
 * updateStructureStandard:
 *
 * Vanilla proposal for structures -- i.e., draw from the default copula proposal
 * function residing in the appropriate module.
 * 
 * Uses an uniformly distributed initialization parameter for new edges. See
 * method updateStructureOneParam for a more sophisticated method.
 *
 * This version proposes fully uniformly chosen moves.
 */

void MCMCStructure::updateStructureStandard(unsigned int k)
{
  /* Skip this update if tree is not associated with any data point */
  if (this->M > 0 && this->z_count[k] == 0)
    return;

  /* Find corresponding tree */
  Tree *tree = this->tree_mix[k];
  int offset = k * this->p * this->p;

  /* Propose and store indices of the endpoints of the removed and added edges. */
  int deleted_edge[2], added_edge[2];
  Tree *new_tree = tree->proposeMoveEdge(deleted_edge, added_edge);

  /* Propose a new copula 'cprime' and retrieve the old copula 'c' */
  Copula* cprime = new_copula(&this->copd[(added_edge[0] * this->p + added_edge[1]) * this->copd_size]); 
  cprime->randomize();
  Copula* c = this->cop_mix[offset + deleted_edge[0] * this->p + deleted_edge[1]];
  Copula* cOld = this->cop_mix[offset + added_edge[0] * this->p + added_edge[1]];
  cOld->randomize(); /* Postponed sampling of cOld from its prior (as if it happened in the previous step) */

  /* Evaluate proposal */
  double llik_prime   = cprime->LLik(this->all_M[added_edge[0]], this->all_M[added_edge[1]], this->idx[k]);
  double lprior_prime = cprime->LPrior();
  double lq_prime     = c->randomizeLL(); //Probability of generating c's params from the randomize() distribution
  double llik         = c->LLik(this->all_M[deleted_edge[0]], this->all_M[deleted_edge[1]], this->idx[k]);
  double lprior       = cOld->LPrior();
  double lq           = cprime->randomizeLL(); //Probability of generating cprime's params from the randomize() distribution

  /* Decide on acceptance */
  double ratio = exp((llik_prime + lprior_prime + lq_prime) - (llik + lprior + lq));
  if (runif(0, 1) < ratio) {
      delete this->cop_mix[offset + added_edge[0] * this->p + added_edge[1]];
      this->cop_mix[offset + added_edge[0] * this->p + added_edge[1]] = cprime;
      this->cop_mix[offset + added_edge[1] * this->p + added_edge[0]] = cprime;
      delete this->tree_mix[k];
      this->tree_mix[k] = new_tree;
      this->accepted_tree++;
  } else {
      delete new_tree;
      delete cprime;
  }        
}   

/**
 * updateStructureOneParam:
 *
 * Update tree structure. Assumes all copulas have one parameter. In this case,
 * proposes a new parameter based on the implied Spearman's rho coefficient.
 * The idea is proposing a move that moves from a chain X - Y - Z  to a chain
 * Y - Z - X. In order to do that, we find the implied Spearman's rho for (X, Z)
 * based on the original chain, and use it as a basis to propose the parameter of
 * the new copula Z - X.
 *
 * Parameter 'k' indicates which tree to update.
 */

void MCMCStructure::updateStructureOneParam(unsigned int k)
{
    assert(this->one_param);

    /* Skip this update if tree is not associated with any data point */
    if (this->M > 0 && this->z_count[k] == 0)
        return;

    //if (this->z_count[k] < 20) {
    //    updateStructureStandard(k);
    //    return;
    //}

    /* Get tree references and index of data points associated with it */
    Tree *tree = this->tree_mix[k];
    int offset = k * this->p * this->p;
    set<int> *idx = this->idx[k];

    /* Choose which edge to add (which implies which edge to remove). Fills 'move' accordingly. */
    int move[3];
    //Tree *new_tree = tree->proposeMoveEdgeVertex(move);
    double lq_tree_prime, lq_tree;
    Tree *new_tree = tree->proposeMoveUniformTreeangle(move, &lq_tree, &lq_tree_prime);

    /* Provides the endpoints of the deleted and added edge */
    int deleted_edge[] = {move[0], move[1]};
    int added_edge[]   = {move[0], move[2]};

    /* Find the two copulas corresponding to the existing path between the 
       endpoints of the new edge. Get also the current (unused) copula between
       the endpoints of the proposed edge. */
    Copula* c      = this->cop_mix[offset + move[0] * this->p + move[1]];
    Copula* cOther = this->cop_mix[offset + move[1] * this->p + move[2]];
    Copula* cOld   = this->cop_mix[offset + added_edge[0] * this->p + added_edge[1]];
    if (!this->do_indcopula_update) cOld->randomize(); /* Postponed sampling of cOld from its prior (as if it happened in the previous step) */
    
    /* Get the implied Spearman's rho for 'move[0]' and 'move[2]' according to current tree */
    double rho_01 = c->spearmanRho();
    double rho_12 = cOther->spearmanRho();

    /* Propose new copula 'cprime'. Samples parameter from a Gaussian random walk
       centralized on a 'implied parameter' (impl_rho). The implied parameter is an 
       approximation (exact in the case of the Gaussian copula) of the parameter in the
       copula between (move[0], move[1]) that is implied by the current tree. 
       Currently, tree-mixtures are restricted to use the same family of copulas for all edges. */
    //double impl_rho = rho_01 * rho_12;
    double impl_rho = 2 * rho_01 * rho_12 / (rho_12 * rho_12 + 1);
    
    //double a = rho_01, b = rho_12, c = 0;
    //minimizeCorrDist(a, &b, &c);
    //Rprintf("FYI: a = %f, old_b = %f, b = %f, c = %f\n", a, b, c);

    double *params_buffer = new double[1];    
    cOld->getImpliedSpearmanParams(impl_rho, params_buffer);    
    double proposal_param = params_buffer[0];
    double lq_u1;
    Copula* cprime = cOld->RW(proposal_param, 1, &lq_u1); 

    /* Get log-likelihood of reverse jump */
    //double reverse_impl_rho = cprime->spearmanRho() * rho_12;
    double reverse_impl_rho = 2 * cprime->spearmanRho() * rho_12 / (rho_12 * rho_12 + 1);
    c->getImpliedSpearmanParams(reverse_impl_rho, params_buffer);
    double reverse_proposal_param = params_buffer[0];
    double lq_u2 = c->evalRW(1, reverse_proposal_param); 

    delete[] params_buffer;

    /* Evaluate proposal */
    double llik_prime    = cprime->LLik(this->all_M[added_edge[0]], this->all_M[added_edge[1]], idx); //likelihood of the added copula
    double lprior_prime  = cprime->LPrior();                     //prior of the added copula parameter after the move
    //double lq_tree_prime = -log((double) new_tree->num_edges)    //reverse probability of jumping from tree_prime to tree
    //                       -log(new_tree->adj_list[move[0]].size() - 1. + new_tree->adj_list[move[2]].size() - 1.);    
    double lq_prime      = lq_u2;

    double llik          = c->LLik(this->all_M[deleted_edge[0]], this->all_M[deleted_edge[1]], idx); //likelihood of the *removed* copula
    double lprior        = cOld->LPrior();                       //prior of the added copula parameter *before* the move
    //double lq_tree       = -log((double) tree->num_edges)        //probability of jumping from tree to tree_prime
    //                       -log(tree->adj_list[move[0]].size() - 1. + tree->adj_list[move[1]].size() - 1.);
    double lq            = lq_u1;

    double ratio = exp((llik_prime + lprior_prime + lq_tree_prime + lq_prime) - 
                       (llik + lprior + lq_tree + lq));

    /* Decide on acceptance */
    if (runif(0, 1) < ratio) {
        delete this->cop_mix[offset + added_edge[0] * this->p + added_edge[1]];
        this->cop_mix[offset + added_edge[0] * this->p + added_edge[1]] = cprime;
        this->cop_mix[offset + added_edge[1] * this->p + added_edge[0]] = cprime;
        delete this->tree_mix[k];
        this->tree_mix[k] = new_tree;
        this->accepted_tree++;
    } else {
        delete new_tree;
        delete cprime;
    }
}   

/**
 * updateStructureOneParamHybrid:
 *
 * A hybrid Monte Carlo method. Numerically integrate the parameters (within a
 * triangular proposal), then use the numerical approximation as a Metropolis
 * proposal, along with parameter proposals.
 */

/**
 * Samples the next state of a tree random walk using its normalized weights
 */

void MCMCStructure::updateStructureOneParamHybrid(unsigned int k)
{
    assert(this->one_param);

    /* Skip this update if tree is not associated with any data point */
    if (this->M > 0 && this->z_count[k] == 0)
        return;

    /* Get tree references and index of data points associated with it */
    Tree *tree = this->tree_mix[k];
    int offset = k * this->p * this->p;
    set<int> *idx = this->idx[k];

    /* Choose which treenagle to perturb */
    int    treeangle[3];
    double dummy;
    Tree *temp_tree = tree->proposeMoveUniformTreeangle(treeangle, &dummy, &dummy); delete temp_tree;

    /* Calculates the numerical weight of the three edges */
    double w01 = calcPostEdgeWeight(k, treeangle[0], treeangle[1]);
    double w12 = calcPostEdgeWeight(k, treeangle[1], treeangle[2]);
    double w02 = calcPostEdgeWeight(k, treeangle[0], treeangle[2]);

    /* Tree proposal distribution, with sample */
    /*double z_t = w01 * w12 + w01 * w02 + w02 * w12;
    double t012 = w01 * w12 / z_t; 
    double t021 = w02 * w12 / z_t;
    double t102 = w01 * w02 / z_t;
    double t_dist[] = {t012, t021, t102}; Rprintf("tree_dist = %f, %f, %f\n", t012, t021, t102);
    int    tree_choice = sample_discrete(t_dist, 3);*/
    
    double z_t_prime = w01 * w02 + w02 * w12;
    double t021_prime = w02 * w12 / z_t_prime;
    double t102_prime = w01 * w02 / z_t_prime;
    double t_dist_prime[] = {t021_prime, t102_prime};
    int    tree_choice = sample_discrete(t_dist_prime, 2);
    int    move[] = {-1, -1, -1};

    bool   tree_changed = false;
    double z_t_reverse = w01 * w12;
    switch(tree_choice) {
        //case 0:
            //No new tree: sample parameters based on the current values using slice sampler
        //    move[0] = treeangle[0]; move[1] = treeangle[1]; move[2] = treeangle[2];             
        //    break;
        case 0:
            move[0] = treeangle[0]; move[1] = treeangle[1]; move[2] = treeangle[2];
            tree_changed = true;
            z_t_reverse += w01 * w02;
            break;
        case 1:
            move[0] = treeangle[2]; move[1] = treeangle[1]; move[2] = treeangle[0];
            tree_changed = true;
            z_t_reverse += w02 * w12;
            break;
    }

    /* Variables used in the evaluation of the proposal */
    double llik_prime;   //likelihood of the copulas after the move
    double lprior_prime; //prior of the copulas after the move
    double lq_prime;      //reverse parameter proposal new_copulas -> current_copulas
    //double lq_tree_prime = t012; //reverse probability of jumping from tree_prime to tree (i.e., probability of sampling current tree)
    double lq_tree_prime = w01 * w12 / z_t_reverse; //reverse probability of jumping from tree_prime to tree (i.e., probability of sampling current tree)

    double llik;         //likelihood of the current copulas
    double lprior;       //prior of current copulas
    double lq;           //forward parameter proposasl current_copulas -> new_copulas
    double lq_tree = t_dist_prime[tree_choice]; //probability of jumping from tree to tree_prime (i.e., probability of sampling tree_prime)

    if (tree_changed) { /* Tree changed: sample new parameters accordingly (i.e., via implied rank correlation matrix */        

        /* Provides the endpoints of the deleted and added edge */
        int deleted_edge[] = {move[0], move[1]};
        int added_edge[]   = {move[0], move[2]};        

        /* Find the two copulas corresponding to the existing path between the 
           endpoints of the new edge. Get also the current (unused) copula between
           the endpoints of the proposed edge. */
        Copula* c      = this->cop_mix[offset + move[0] * this->p + move[1]];
        Copula* cOther = this->cop_mix[offset + move[1] * this->p + move[2]];
        Copula* cOld   = this->cop_mix[offset + added_edge[0] * this->p + added_edge[1]];
        if (!this->do_indcopula_update) cOld->randomize(); /* Postponed sampling of cOld from its prior (as if it happened in the previous step) */
        
        /* Get the implied Spearman's rho for 'move[0]' and 'move[2]' according to current tree */
        double rho_01 = c->spearmanRho();
        double rho_12 = cOther->spearmanRho();

        /* Propose new copula 'cprime' as in the previous tree proposals */
        double impl_rho = 2 * rho_01 * rho_12 / (rho_12 * rho_12 + 1);
        
        double *params_buffer = new double[1];    
        cOld->getImpliedSpearmanParams(impl_rho, params_buffer);    
        double proposal_param = params_buffer[0];
        Copula* cprime = cOld->RW(proposal_param, 1, &lq); 

        /* Get log-likelihood of reverse jump */
        double reverse_impl_rho = 2 * cprime->spearmanRho() * rho_12 / (rho_12 * rho_12 + 1);
        c->getImpliedSpearmanParams(reverse_impl_rho, params_buffer);
        double reverse_proposal_param = params_buffer[0];
        lq_prime = c->evalRW(1, reverse_proposal_param); 

        /* Evaluate proposal factors */
        llik_prime    = cprime->LLik(this->all_M[added_edge[0]], this->all_M[added_edge[1]], idx);
        lprior_prime  = cprime->LPrior();
        llik          = c->LLik(this->all_M[deleted_edge[0]], this->all_M[deleted_edge[1]], idx);
        lprior        = cOld->LPrior();

        /* Evaluate proposal */
        double ratio = exp((llik_prime + lprior_prime + lq_tree_prime + lq_prime) - 
                           (llik + lprior + lq_tree + lq));

        /* Decide on acceptance */
        if (runif(0, 1) < ratio) {
            delete this->cop_mix[offset + added_edge[0] * this->p + added_edge[1]];
            this->cop_mix[offset + added_edge[0] * this->p + added_edge[1]] = cprime;
            this->cop_mix[offset + added_edge[1] * this->p + added_edge[0]] = cprime;
            tree->adj_matrix[added_edge[0] * this->p + added_edge[1]] = 
                                tree->adj_matrix[added_edge[1] * this->p + added_edge[0]] = 1;
            tree->adj_matrix[deleted_edge[0] * this->p + deleted_edge[1]] = 
                                tree->adj_matrix[deleted_edge[1] * this->p + deleted_edge[0]] = 0;
            Tree *new_tree = tree->clone();
            delete this->tree_mix[k];
            this->tree_mix[k] = new_tree;
            this->accepted_tree++;
        } else {
            delete cprime;
        }
        delete[] params_buffer;

    } else { /* Tree did not change: sample new parameters for the old copulas (i.e., via implied rank correlation matrix) */        
        
        Rprintf("NOT SUPPOSED TO GET HERE!!!!!!!!\n");

        double lqfwd1, lqbak1;
        Copula *c1 = this->cop_mix[offset + move[0] * this->p + move[1]];
        Copula *cprime1 = c1->RW(1, &lqfwd1, &lqbak1);
        double lqfwd2, lqbak2;
        Copula *c2 = this->cop_mix[offset + move[1] * this->p + move[2]];
        Copula *cprime2 = c2->RW(1, &lqfwd2, &lqbak2);
        
        /* Evaluate proposal factors */
        llik_prime    = cprime1->LLik(this->all_M[move[0]], this->all_M[move[1]], idx) + 
                        cprime2->LLik(this->all_M[move[1]], this->all_M[move[2]], idx);
        lprior_prime  = cprime1->LPrior() + cprime2->LPrior();
        lq_prime      = lqbak1 + lqbak2;
        llik          = c1->LLik(this->all_M[move[0]], this->all_M[move[1]], idx) +
                        c2->LLik(this->all_M[move[1]], this->all_M[move[2]], idx);
        lprior        = c1->LPrior() + c2->LPrior();
        lq            = lqfwd1 + lqfwd2;

        /* Evaluate proposal */
        double ratio = exp((llik_prime + lprior_prime + lq_tree_prime + lq_prime) - 
                           (llik + lprior + lq_tree + lq));

        /* Decide on acceptance */
        if (runif(0, 1) < ratio) {
            delete c1;
            this->cop_mix[offset + move[0] * this->p + move[1]] = cprime1;
            this->cop_mix[offset + move[1] * this->p + move[0]] = cprime1;
            delete c2;
            this->cop_mix[offset + move[1] * this->p + move[2]] = cprime2;
            this->cop_mix[offset + move[2] * this->p + move[1]] = cprime2;
            this->accepted_tree++;
        } else {
            delete cprime1;
            delete cprime2;
        }

    }

}   

/**
 * Sample k-th tree given parameters. Unlike the other methods, this one is not a random walk.
 * However, it does cost much more than the random walk algorithm and it's not necessarily
 * better, since it does not sample the respective new parameters at the same time
 * (unlike previous papers that used this algorithm, we have to condition on the parameters
 * instead of marginalizing them out).
 *
 * Algorithm taken from ``Generating random spanning trees more quickly than cover time"
 * (D. Wilson)
 *
 * In practice, I found this method to be essentially useless. Weights have extreme
 * distributions: quite often, the walk done in this algorithm gets stuck and never converges.
 * The main reason should be that one shouldn't try to sample a whole tree given that
 * the current parameters only really favor the current tree...
 */

void MCMCStructure::updateStructureRandomized(unsigned int k)
{
  /* Skip this update if tree is not associated with any data point */
  if (this->M > 0 && this->z_count[k] == 0)
    return;

  /* Find corresponding copula parameters */
  int offset = k * this->p * this->p;

  int trials = 0;

  /* Compute weights */
  double* weights = new double[this->p * this->p];
  for (unsigned int i = 0; i < this->p; i++)
      for (unsigned int j = i + 1; j < this->p; j++) {
          Copula* copula = this->cop_mix[offset + i * this->p + j];
          weights[i * this->p + j] = copula->LLik(this->all_M[i], this->all_M[j], this->idx[k]) +
                                     copula->LPrior();
          weights[j * this->p + i] = weights[i * this->p + j];
      }

  /* Randomize sampling algorithm */
  int root = (int) floor(runif(0, 1) * this->p); if (root == this->p) root--;
  bool* in_tree = new bool[this->p];
  for (unsigned int i = 0; i < this->p; i++)
      in_tree[i] = false;
  int* next = new int[this->p];
  in_tree[root] = true;
  next[root] = -1;
  for (unsigned int i = 0; i < this->p; i++) {
     unsigned u = i;
     while (!in_tree[u]) {
         next[u] = randomSuccessor(u, weights);
         u = next[u];
         trials++;
         if (trials > 100 * this->p) {
             Rprintf("EXACT SAMPLING TAKING TOO LONG... Changing to simple sampler\n");
             delete[] weights;
             delete[] in_tree;
             delete[] next;
             updateStructureRandomWalk(k);
             return;
         }
     }
     u = i;
     while (!in_tree[u]) {
         in_tree[u] = true;
         u = next[u];
     }
  }

  /* Create tree from 'next' */
  int* adjacencies = new int[this->p * this->p];
  for (unsigned int i = 0; i < this->p; i++)
      for (unsigned int j = i; j < this->p; j++)
          adjacencies[i * this->p + j] = adjacencies[j * this->p + i] = 0;
  for (unsigned int i = 0; i < this->p; i++)
      if (next[i] >= 0)
          adjacencies[i * this->p + next[i]] = adjacencies[next[i] * this->p + i] = 1;
  this->tree_mix[k] = new Tree(adjacencies, this->p);
  this->accepted_tree++;

  /* Clean-up */
  delete[] weights;
  delete[] in_tree;
  delete[] next;
  delete[] adjacencies;
}   



/**
 * Calculates the log-marginal likelihood of the copula c(u, v) (i.e., 
 * the probability of the data when integrating the prior over the copula parameter).
 *
 * This uses only the points associated with cluster k.
 */

double MCMCStructure::calcPostEdgeWeight(unsigned int k, unsigned int u, unsigned int v)
{
   assert(this->one_param);
  
   double params[1], old_params[1];
   double *lliks = new double[this->quadrature_size];
   double *priors = new double[this->quadrature_size];
   double max_llik = -std::numeric_limits<double>::max();

   for (unsigned int q = 0; q < this->quadrature_size; q++) {
       this->all_cop[u * this->p + v]->Params(old_params);
       params[0] = this->copula_quadrature[u * this->p + v][q];       
       this->all_cop[u * this->p + v]->setParams(params);

       lliks[q] = this->all_cop[u * this->p + v]->LLik(this->all_M[u], this->all_M[v], this->idx[k]);
       priors[q] = this->all_cop[u * this->p + v]->LPrior();
       if (lliks[q] > max_llik)
           max_llik = lliks[q];

       this->all_cop[u * this->p + v]->setParams(old_params);
   }

   double sum = 0;
   for (unsigned int q = 0; q < this->quadrature_size; q++)
       sum += exp(lliks[q] - max_llik + priors[q]);

   delete[] priors; delete[] lliks;
   return max_llik + log(sum);
}

unsigned int MCMCStructure::randomSuccessor(unsigned int u, double *weights)
{
   double* prob = new double[this->p - 1];
   double  sum = 0.;
   double  max_weight = -std::numeric_limits<double>::max(); // ``Log(0) == Negative infinity'

   for (unsigned int i = 0; i < this->p; i++) {
       if (i < u) {
           prob[i] = weights[u * this->p + i];
           if (prob[i] > max_weight)
               max_weight = prob[i];
       }
       else if (i > u) {
           prob[i - 1] = weights[u * this->p + i];
           if (prob[i - 1] > max_weight)
               max_weight = prob[i - 1];
       }
   }
   for (unsigned int i = 0; i < this->p - 1; i++) {
       prob[i] -= max_weight;
       prob[i] = exp(prob[i]);
       sum += prob[i];
   }

   for (unsigned int i = 0; i < this->p - 1; i++)
       prob[i] /= sum;

   unsigned int output = sample_discrete(prob, this->p - 1);
   delete[] prob;

   return output;
}




/**
 * completeData:
 *
 * Complete a dataset with the average value of the training data.
 *
 * NOTE: THIS IS FOR TESTING/DEBUGGIN PURPOSES ONLY -- IT WILL BE REMOVED
 * IN THE FUTRE SO, AND DATA AUGMENTATION WILL BE USED
 */

/**
 * Samples a tree given current parameters using the exact method
 * kindly provided by Sergey Kirshner.
 */

void MCMCStructure::updateStructureExact(unsigned int k)
{
  /* Skip this update if tree is not associated with any data point */
  if (this->M > 0 && this->z_count[k] == 0)
    return;

  int *adjacencies = new int[this->p * this->p];
  for (unsigned int i = 0; i < this->p; i++)
      for (unsigned int j = 0; j < this->p; j++)
          adjacencies[i * this->p + j] = 0;

  int offset = k * this->p * this->p;

  /* Initialize weight matrix */
  double *M = new double[this->p];
  double *B = new double[this->p * this->p];
  for (unsigned int i = 0; i < this->p; i++) {
      B[i * this->p + i] = 0;
      M[i] = 0.;
      for (set<int>::iterator j = this->idx[k]->begin(); j != this->idx[k]->end(); j++)
          M[i] += this->all_M[i]->LLik(this->data[this->z[*j] * this->p + i]);
      for (unsigned int j = i + 1; j < this->p; j++) {
          B[i * this->p + j] = this->cop_mix[offset + i * this->p + j]->LLik(this->all_M[i], this->all_M[j], this->idx[k]);
          B[j * this->p + i] = B[i * this->p + j];          
      }
  }
  double max_b = -std::numeric_limits<double>::max();
  for (unsigned int i = 0; i < this->p; i++) {
      for (unsigned int j = 0; j < this->p; j++) 
          if (i != j) {
             B[i * this->p + j] += M[j]; //B[i, j] == P(j | i)
             if (B[i * this->p + j] > max_b)
                max_b = B[i * this->p + j];
          }
  }
  for (unsigned int i = 0; i < this->p; i++) {
      for (unsigned int j = 0; j < this->p; j++) 
          if (i != j) {
             B[i * this->p + j] -= max_b;
             B[i * this->p + j] = exp(B[i * this->p + j]); 
          }
  }

  /* Number of nodes that will have parents */
  unsigned int V = this->p - 1;

  /* Laplacian and its inverse */
  double *L = new double[V * V];
  for (unsigned int i = 0; i < V; i++) {
      for (unsigned int j = 0; j < V; j++)
          if (i != j)
             L[i * V + j] = -B[j * this->p + i];
  }
  for (unsigned int i = 0; i < V; i++) {
      L[i * V + i] = 0;
      for (unsigned int j = 0; j < V; j++)
          L[i * V + i] += B[j * this->p + i];
  }
  
  /* Inverse of the Laplacian */
  double *K = new double[V * V];
  memcpy(K, L, sizeof(double) * V * V);
  //Rprintf("Before inversion...\n");
  inverse(K, V);
  //Rprintf("Inversion OK!\n");

  double *cof = new double[this->p];
  double *delta = new double[V * V];
  double *aux_delta = new double[V];
  double *P = new double[this->p];
  double *norm_P = new double[this->p];

  for (unsigned int v = 0; v < V; v++) {      
      for (unsigned int i = 0; i < this->p; i++)
          cof[i] = K[v * V + v];
      for (unsigned int i = 0; i < V; i++)
          cof[i] -= K[i * V + v];      
      bool all_invalid = true;
      double sP = 0;
      for (unsigned int i = 0; i < this->p; i++) {
          P[i] = cof[i] * B[i * this->p + v];
          if (P[i] < 1.e-16)
              P[i] = 0.;          
          if (!isnan(P[i]) && !isinf(P[i]) && P[i] != 0) {
              all_invalid = false;
              sP += P[i];
          }
          else
              P[i] = 0;
      }

      for (unsigned int i = 0; i < this->p; i++) {
          norm_P[i] = P[i] / sP;
          //Rprintf("P[%d] = %f\n", i, norm_P[i]);
      }
      //Rprintf("\n");

      /*if (all_invalid || isinf(sP)) {
          delete[] cof; delete[] delta; delete[] aux_delta; delete[] P; delete[] norm_P;
          delete[] adjacencies; delete[] M; delete[] B; delete[] L; delete[] K;
          Rprintf("Switch...\n");
          updateStructureOneParam(k);
          return;
      }*/

      if (all_invalid || isinf(sP))
          break;

      unsigned int u = sample_discrete(norm_P, this->p);
      if (u >= this->p)
         Rprintf("!!!! u = %d, p = %d\n", u, this->p);

      adjacencies[u * this->p + v] = adjacencies[v * this->p + u] = 1;

      if (u >= V) {
         for (unsigned int i = 0; i < V; i++)
             aux_delta[i] = -B[u * this->p + v] * K[v * V + i];         
      } else {
         for (unsigned int i = 0; i < V; i++)
             aux_delta[i] = -B[u * this->p + v] * (K[v * V + i] - K[u * V + i]);
      }
      aux_delta[v]++;
      for (unsigned int i = 0; i < V; i++)
          for (unsigned int j = 0; j < V; j++)
              delta[i * V + j] = K[i * V + v] * aux_delta[j];
      for (unsigned int i = 0; i < V; i++)
          for (unsigned int j = 0; j < V; j++)
              K[i * V + j] += delta[i * V + j] / P[u];      
  }

  /* Check if the result is a connected tree and CHECK FOR CYCLES!!! */
  int num_edges = 0;
  for (unsigned int i = 0; i < this->p; i++)
      for (unsigned int j = i + 1; j < this->p; j++)
          if (adjacencies[i * this->p + j] == 1)
              num_edges++;
  if (!this->tree_mix[k]->validAdjacencies(adjacencies, this->p) || num_edges < this->p - 1) {
      //Rprintf("Not a valid connected tree... (num_edges = %d)\n", num_edges);
      delete[] cof; delete[] delta; delete[] aux_delta; delete[] P; delete[] norm_P; 
      delete[] adjacencies; delete[] M; delete[] B; delete[] L; delete[] K;
      updateStructureStandard(k);
      return;
  }

  /* Finalization */
  
  /*Rprintf("\n");
  for (int i = 0; i < this->p; i++) {
      Rprintf("%d: ", i);
      for (int j = 0; j < this->p; j++)
          if (adjacencies[i * this->p + j] != 0)
              Rprintf("%d ", j);
      Rprintf("\n");
  }
  Rprintf("\n");*/
  delete this->tree_mix[k];
  this->tree_mix[k] = new Tree(adjacencies, this->p);
  this->accepted_tree++;							 
  delete[] cof; delete[] delta; delete[] aux_delta; delete[] P; delete[] norm_P; 
  delete[] adjacencies; delete[] M; delete[] B; delete[] L; delete[] K;
}   



void MCMCStructure::completeData(double *data, unsigned int n_d, int *mars_matrix_data)
{
    double *avg = new double[this->p];
    int    *count = new int[this->p];
    for (unsigned int i = 0; i < this->p; i++) {
        avg[i] = 0.;
        count[i] = 0;
    }
    for (unsigned int i = 0; i < this->n; i++)
        for (unsigned int j = 0; j < this->p; j++)
            if (this->mars_matrix[i * this->p + j] == 1) {
                avg[j] += this->data[i * this->p + j];
                count[j]++;
            }   
    for (unsigned int i = 0; i < this->p; i++)
        avg[i] /= count[i];

    for (unsigned int i = 0; i < n_d; i++)
        for (unsigned int j = 0; j < this->p; j++)
            if (mars_matrix_data[i * this->p + j] == 0)
                data[i * this->p + j] = avg[p];

    delete[] avg; delete[] count;
}


/**
 * predLogLikelihood:
 *
 * Predictive log-likelihood. A test set 'test' is given, with the respective
 * number of test points 'n_t'.
 *
 * For practical reasons, the computed resulted is the average predictive
 * log-likelihood of the sample: we compute the predictive log-lik for each
 * test point individually, and return the average.
 */

double MCMCStructure::predLogLikelihood(unsigned int T, unsigned int burn_in, double *test, 
                                        unsigned int n_t, int *mars_matrix_test)
{
    if (this->M == 0) {
        warning("The log-likelihood prediction is currently implemented for Dirichlet process mixtures only.\n");
        return 0.;
    }

    /* Also, output files */
    ofstream predictive_file; predictive_file.open("predictive.mcmc");
    double *acc_lik = new double[n_t];

    /* THIS IS WHAT IS GOING TO BE REPLACED WITH DATA AUGMENTATION!!! */
    completeData(test, n_t, mars_matrix_test);

    /* allocate space needed */
    int *z = new int[this->n];
    double *pred_lik = new double[n_t];
    for (unsigned int i = 0; i < n_t; i++)
         pred_lik[i] = 0.;

    /* open the parameter trace files and skip the burn in samples */
    PredInit(burn_in);

    /* now loop over each MCMC sample */
    for (unsigned int t = burn_in; t < T; t++) {    

        //Get alpha
        k_in >> this->alpha; k_in >> this->alpha;

        //Get trees, allocates the tree_mix data structure
        GetTreesIn();

        /* allocate one more component than is necessary to store the trace
           of the copula parameters */
        this->cop_mix = new Copula*[(this->K + 1) * this->p * this->p];

        /* get the marginal and copula parameters, using space allocated above */
        GetParamsIn();

        /* adding another component of random copulas for the DP mixture implementation 
           for calculating the predictive log likelihood */
        for (unsigned int v1 = 0; v1 < this->p; v1++)
             for (unsigned int v2 = v1 + 1; v2 < this->p; v2++) {
                  unsigned int offset = K * this->p * this->p;
                  unsigned int v1v2 = v1 * this->p + v2;
                  this->cop_mix[offset + v1v2] = new_copula(&copd[v1v2 * this->copd_size]); 
                  this->cop_mix[offset + v2 * this->p + v1] = this->cop_mix[offset + v1v2];
             }
 
        //Get assignments from the z_in file
        for (unsigned int i = 0; i < this->n; i++)  z_in >> z[i]; 

        //Do the actual (likelihood) calculation for each point
	    predLikelihood(pred_lik, test, n_t, z);

        /* Output to file */
        double outputI = 0.;
        for (unsigned int i = 0; i < n_t; i++) {
            acc_lik[i] += pred_lik[i];
            double point = acc_lik[i] / (t - this->burn_in);
            outputI += log(point);
        }
        double pr = outputI / this->n_test;
        if (isinf(pr) || isnan(pr))
            pr = 0.;
        predictive_file << pr << "\n";


	    /* each round adds one one sample to each of pred_lik */

        /* print a progress meter */
        if (t % 10 == 0) Rprintf("Running through MCMC sample #%d\n", t);

        //Clean-up for the current MCMC iteration 
        for (int k = 0; k < (int) K; k++) delete tree_mix[k];
        delete[] tree_mix;    
        for (int k = 0; k < (int) K + 1; k++) {
             int offset = k * this->p * this->p;
             for (unsigned int v1 = 0; v1 < this->p; v1++)
                  for (unsigned int v2 = v1 + 1; v2 < this->p; v2++)
                      delete this->cop_mix[offset + v1 * this->p + v2];
        }
        delete[] this->cop_mix;
    }

    /* convert the sum in each pred_lik into a mean,
       then take the logarithm and ultimately calculate
       the average of the log(pred_lik) */

    double output = 0.;
    for (unsigned int i = 0; i < n_t; i++) {
         pred_lik[i] /= (T - burn_in);         
         pred_lik[i] = log(pred_lik[i]);
         if (isinf(pred_lik[i]))
             warning("Test point %d had zero likelihood\n", i);
         output += pred_lik[i];
    }
    output /= n_t;

    /* clean up for the function */
    z_in.close(); k_in.close(); tree_in.close(); param_in.close();
    delete[] z; delete[] pred_lik;
    delete[] acc_lik;
    predictive_file.close();

    return output;
}


/*
 * predLogLikelihood:
 *
 * increment pred_lik by the likelihood of the test point(s) in the
 * test vector of length n_t
 */

void MCMCStructure::predLikelihood(double *pred_lik, double *test, unsigned int n_t, int *z)
{
  for (unsigned int i = 0; i < n_t; i++) {
       double llik = predLoglikelihood(i, test, z);
       pred_lik[i] += exp(llik);            
  }
}

double MCMCStructure::predLoglikelihood(unsigned int i, double *test, int *z)
{
    int test_pos = i * this->p;
    double t_lik = 0.;
    if (runif(0, 1) > this->alpha / (this->n + this->alpha)) {
        int choice = sample_uniform(this->n);
        t_lik = copulaLogLikelihood(tree_mix[z[choice]], z[choice], &test[test_pos]);      
    } else {
        Tree *tree_zero = sample_unif_tree(this->p);
        for (unsigned int v1 = 0; v1 < this->p; v1++)
            for (unsigned int v2 = v1 + 1; v2 < this->p; v2++)
                 this->cop_mix[K * this->p * this->p + v1 * this->p + v2]->randomize();
        t_lik = copulaLogLikelihood(tree_zero, K, &test[test_pos]);
        delete tree_zero;
    }
    t_lik += marginalsLogLikelihood(&test[test_pos]);
    return t_lik;
}

/* 
 * GetParamsIn:
 *
 * get the marginal and copoula parameters from their trace files
 * and and point the class pointers to them -- assumes that both
 * the marginal and copula_mix data structures have been preallocated.
 *
 * In particular, if marginals are fixed a prior, then do not load them
 * from data. IT IS ASSUMED THAT THE MARGINALS CURRENTLY STORED AT THOSE
 * MARGINALS OF INTEREST.
 */   

void MCMCStructure::GetParamsIn(void)
{
  //Organize marginal parameters
  if (this->do_marg_update)
      for (unsigned int v = 0; v < this->p; v++) {
           int n_params = this->all_M[v]->NumParams();
           double *mp = new double[n_params];
           for (int m = 0; m < n_params; m++)
                param_in >> mp[m];
           this->all_M[v]->setParams(mp);
           delete[] mp;
      }
  //Organize copula parameters
  for (int k = 0; k < (int) this->K; k++) {
       int offset = k * this->p * this->p;
       for (unsigned int v1 = 0; v1 < this->p; v1++)
            for (unsigned int v2 = v1 + 1; v2 < this->p; v2++) {
                 unsigned int v1v2 = v1 * this->p + v2;
                 this->cop_mix[offset + v1v2] = new_copula(&copd[v1v2 * this->copd_size]); 
                 this->cop_mix[offset + v2 * this->p + v1] = this->cop_mix[offset + v1v2];
                 int n_params = this->all_cop[v1 * this->p + v2]->NumParams();
                 double *cp = new double[n_params];
                 for (int m = 0; m < n_params; m++) param_in >> cp[m];
                      this->cop_mix[offset + v1v2]->setParams(cp);
                 delete[] cp;
            }
  }
}


/*
 * GetTreesIn:
 *
 * allocates the tree data structure and,
 * read the next set of trees from the tree_in file and
 * point this->tree_mix to them
 */

void MCMCStructure::GetTreesIn(void)
{	
  int *adjacencies = new int[this->p * this->p];

  tree_in >> this->K; //get number of trees for the t-th simulation
  tree_mix = new Tree*[this->K];
  for (int k = 0; k < (int) this->K; k++) {
    for (unsigned int v1 = 0; v1 < this->p; v1++) {
      adjacencies[v1 * this->p + v1] = 0;
      for (unsigned int v2 = v1 + 1; v2 < this->p; v2++)
        adjacencies[v1 * this->p + v2] = adjacencies[v2 * this->p + v1] = 0;
    }           
    int n_edges;
    tree_in >> n_edges;
    for (int e_iter = 0; e_iter < n_edges; e_iter++) {
      int v1, v2;
      tree_in >> v1; tree_in >> v2;
      adjacencies[v1 * this->p + v2] = adjacencies[v2 * this->p + v1] = 1;
    }
    tree_mix[k] = new Tree(adjacencies, this->p);
  }

  delete[] adjacencies;
}


/*
 * PredInit:
 *
 * Open the files that have the traces of the model parameters,
 * and skip the ones that correspond to burn in 
 */

void MCMCStructure::PredInit(unsigned int burn_in)
{
  /* open the files that have the parameter traces in it */
  z_in.open("z.mcmc");
  k_in.open("k.mcmc");
  tree_in.open("trees.mcmc");
  param_in.open("param.mcmc");

  double dummy;

  /* skip burn_in lines in the trace files */
  for (unsigned int t = 0; t < burn_in; t++) {

       /* skipping in tree */
       int t_K;
       tree_in >> t_K;
       for (int k = 0; k < t_K; k++) {
            int n_edges;
            tree_in >> n_edges;
            for (int e_iter = 0; e_iter < n_edges; e_iter++) {
                 int v1, v2;
                 tree_in >> v1; tree_in >> v2;
             }
       }

       /* skipping in z */
       for (unsigned int i = 0; i < this->n; i++)  z_in >> dummy; 

       /* skipping in k */
       k_in >> dummy; k_in >> dummy;

       /* skipping in marginal params */
       if (this->do_marg_update)
           for (unsigned int v = 0; v < this->p; v++) {
                int n_params = this->all_M[v]->NumParams();
                for (int i = 0; i < n_params; i++)
                     param_in >> dummy;
           }

       /* skipping in copula params */
       for (int k = 0; k < (int) t_K; k++)
            for (unsigned int v1 = 0; v1 < this->p; v1++)
       for (unsigned int v2 = v1 + 1; v2 < this->p; v2++) {
            int n_params = this->all_cop[v1 * this->p + v2]->NumParams();
            for (int i = 0; i < n_params; i++)
                 param_in >> dummy;
       }
  }
}


/**
 * Given datapoint 'point', calculate the likelihood of the tree-copula model based 
 * on 'tree' and the parameters currently stored in this->all_cop
 */

double MCMCStructure::copulaLogLikelihood(Tree *tree, unsigned int k, double *point)
{
    double output = 0.;
    int offset = k * this->p * this->p;
    unsigned int num_edges = tree->num_edges;
    for (unsigned int e = 0; e < num_edges; e++) {
        int v1 = tree->edges[e * 2], v2 = tree->edges[e * 2 + 1];        
        output += this->cop_mix[offset + v1 * this->p + v2]->LLik(this->all_M[v1], this->all_M[v2],
                                                                  point[v1], point[v2]);
        if (isnan(output)) 
            Rprintf("NaN here = edge(%d %d), ranks = (%f %f)\n", v1, v2, this->all_M[v1]->CalcRanks(point[v1]), this->all_M[v2]->CalcRanks(point[v2]));
    }
    return output;
}

/**
 * Given datapoint 'point', calculate the product of the likelihoods of the 
 * marginals using the parameters currently stored in this->all_M.
 */

double MCMCStructure::marginalsLogLikelihood(double *point)
{
    double output = 0.;    
    for (unsigned int v = 0; v < this->p; v++)
         output += this->all_M[v]->LLik(point[v]);        
    return output;
}

/**
 * setTestSet:
 *
 * Set information about test set. This is used during sampling
 * to compute predictive log-likelihood over the test set
 */

void MCMCStructure::setTestSet(unsigned int n_test, double *test_data, int *mars_matrix_test) 
{
  this->n_test = n_test;
  this->test_data = test_data;
  this->mars_matrix_test = mars_matrix_test;
}

/**
 * setBurnIn:
 *
 * Set information about burn in period. This is used during sampling
 * to compute predictive log-likelihood over the test set, and also to store
 * statistics of the missing data prediction.
 */

void MCMCStructure::setBurnIn(unsigned int burn_in) 
{
  this->burn_in = burn_in;
}

/**
 * setMCMCStep:
 *
 * Thin the sampled chain by recording (i.e., outputting to a file) points
 * at every mcmc_step. Useful when chains are long and we need to save disk 
 * space.
 *
 * THE COMPUTATION OF PREDICTIVE LOG-LIKELIHOOD IGNORES 'MCMC_STEP' 
 * (but not 'burn_in').
 */

void MCMCStructure::setMCMCStep(unsigned int mcmc_step) 
{
  this->mcmc_step = mcmc_step;
}

/**
 * setTreeProposal:
 *
 * Set a value that changes the choice of tree proposal mechanism
 *
 * == 0: triangular proposals with ``smart'' choice of copula parameter (cheap)
 * == 1: fully uniform choice of next tree (cheap)
 * == 2: ``exact'' algorithm that samples from the posterior distribution (expensive)
 */

void MCMCStructure::setTreeProposal(int v)
{
  this->tree_prop = v;  
}

/**
 * setMarginalUpdater:
 *
 * Set a flag that indicates if marginals should be sampled (or set to constants).
 */

void MCMCStructure::setMarginalUpdater(bool m)
{
  this->do_marg_update = m;
}


/**
 * setCopulaSliceSampler:
 *
 * Set a flag that indicates if slice sampler should be used to sample copula parameters.
 */

void MCMCStructure::setCopulaSliceSampler(bool u)
{
  this->useCopulaSlice = u;
}

/**
 * setIndCopulaUpdate:
 *
 * Set a flag that indicates if slice sampler should be used to sample copula parameters.
 */

void MCMCStructure::setIndCopulaUpdate(bool u)
{
  this->do_indcopula_update = u;
}

/**
 * storePredictiveLogLikelihood:
 *
 * During sampling, store loglikelihoods for each test point and MCMC iteration.
 */

void MCMCStructure::storePredictiveLogLikelihood(unsigned int t)
{
  for (unsigned int i = 0; i < this->n_test; i++)
       this->predictions[t * this->n_test + i] = predLoglikelihood(i, this->test_data, this->z);
}

/**
 * calculatePredictiveLoglikelihood:
 *
 * Use the stored predictions to compute the predictive log-likelihood of the test set.
 * Notice that we subtract the maximuma point in order to achieve some
 * stability before taking the exponential.
 */

double MCMCStructure::calculatePredictiveLoglikelihood(unsigned int T)
{
  assert(this->n_test > 0);

  double  output = 0.;    
  double* max = new double[this->n_test];
  double* point_t = new double[this->n_test];

  for (unsigned int i = 0; i < this->n_test; i++) {
       /* First calculate the average MCMC log-prediction for that sample */
       double point = 0.;
       point_t[i] = 0.; max[i] = this->predictions[this->burn_in * this->n_test + i];
       for (unsigned int t = this->burn_in + 1; t < T; t++)
            if (this->predictions[t * this->n_test + i] > max[i])
                max[i] = this->predictions[t * this->n_test + i];

       /* When exponentiating, subtract the average first */
       for (unsigned int t = this->burn_in; t < T; t++)
            point += exp(this->predictions[t * this->n_test + i] - max[i]);
       point /= T - this->burn_in;

       /* Add the maximum back */
       output += log(point) + max[i];
  }
  output /= this->n_test;

  /* Output to a file */
  Rprintf("Outputing trace of predictive log-likelihood... ");
  ofstream predictive_file; predictive_file.open("predictive.mcmc");
  for (unsigned int inner_T = this->burn_in; inner_T < T; inner_T++) {
       double outputI = 0;
       for (unsigned int i = 0; i < this->n_test; i++) {
            point_t[i] += exp(this->predictions[inner_T * this->n_test + i] - max[i]);
            double point = point_t[i] / (inner_T - this->burn_in);
            outputI += log(point) + max[i];
        }
        double pr = outputI / this->n_test;
        if (isinf(pr) || isnan(pr))
            pr = 0.;
        predictive_file << pr << "\n";
  }
  predictive_file.close();
  Rprintf("done\n");

  delete[] max; delete[] point_t;
  return output;
}

/**
 * evaluateMissingValues:
 *
 * Compare the average augmented values against the true values in the data.
 * Report root mean square error.
 */

double MCMCStructure::evaluateMissingValues()
{
  double error = 0.;
  int    total_vars = 0;

  /* iterate over all rows that have missing variables */
  for (unsigned int i = 0; i < this->num_missing_rows; i++) {
       unsigned int row = this->missing_rows[i];

       /* iteratate over all variabes that are missing in that particular row */
       for (unsigned int j = 0; j < this->num_missing_cols[i]; j++) {
            unsigned int col = this->missing_columns[i][j];  

            /* Get values to be compared */
            double true_value = this->dataSnapshot[row * this->p + col];
            double imputed_value = this->imputedData[row * this->p + col];
               
            /* Do the comparison: MSE */
            double diff = true_value - imputed_value;
            error += diff * diff;
            total_vars++;
       }
  }

  /* Report average */
  return sqrt(error / total_vars);
}

extern "C" {

  /**
   * R interface function.
   *
   * Sampling from the posterior distribution of parameters and tree
   * structures using the Metropolis-Hastings procedure. The results
   * are stored in text files.
   *
   * Parameters:
   *
   * - adjacencies: an adjacency matrix stored as an p * p array.
   *               It should contain 1 for each adjacent pair, 0 o.w.
   *               Notice this is used to initialize the chain only;
   *
   * - n: the number of points in the training set;
   *
   * - p: the number of variables in the domain;
   *
   * - a, b: hyperparameters for the Dirichlet process with concentration
   *         parameter alpha following a Gamma(a, b);
   * 
   * - data: training set, stored rowise as a n * p array;
   * - mars_matrix: ``missing at random'' matrix. A n * p array, with 0
   *                indicating that the respective entry in 'data' is missing;
   * 
   * - n_test: number of samples in the test set. If positive, we compute the
   *           predictive log-likelihood of the test data;
   * - test_data: training set, stored rowise as a n_test * p array;
   * - mars_matrix_test: ``missing at random'' matrix. A n_test * p array, with 0
   *                      indicating that the respective entry in 'data' is missing;
   *
   * - md: an array containing all the information for each of the p marginals.
   *       It is assumed that the information for each marginal is encoded by
   *       the same number of double-valued variables;
   * - md_size: the number of double-valued variables encoding information for
   *            each marginal (so 'md' is of size md_size * p);
   *
   * - copd: an array containing all the information for each of the p * (p - 1) / 2
   *         copulas. It should be provided in a redundant way: all (i, j) pairs
   *         are stored (pairs (i, i) are just ignored). It is assumed that the information for 
   *         each copula is encoded by the same number of double-valued variables;
   * - copd_size: the number of double-valued variables encoding information for
   *              each copula (so 'copd' is of size copd_size * p * (p - 1) / 2);   
   *
   * - num_samples: number of MCMC samples to be generated;
   * - burn_in: number of initial MCMC samples to skip when dumping output to 
   *            text files;
   * - options: an array of options which can be set from R to define some 
   *            computation to be done (so far, of size 5).
   *              
   *            * options[0] == 1: the missing values in the training set are
   *              not really missing, just marked as missing. Compare the average 
   *              predicted MCMC prediction of missing values against the true ones;
   *            * options[1] == 1: don't allocate memory for the predictive 
   *              log-likelihood, but do it later by reading from the log files;
   *            * options[2] == 1: for the Gaussian copula case, dump the implied covariance
   *              matrix in a file (this can be slow, since log files have to be read);
   *            * options[3] == 1: override choice of tree sampler - use always the
   *              'simple' sampler that proposes parameters for the new copula from
   *              the standard proposal distribution instead of the Spearman proposal;
   *            * options[4] == 1: when set to 1, treats marginals as constants
   *              (the constants are given as part of the 'md' array);
   *            * options[5] == 1: when set to 1, sample copulas from slice sampler;
   *
   * - output: buffer where the result of the predictive log-likelihood with test_data
   *           is going to be stored (if n_test > 0) in output[0]. The results of other
   *           operations may be stored in this array also;
   */

    void mcmc_structure_treevariate(int *n, int *p, double *a, double *b, 
                                    double *data, int *mars_matrix,
                                    int *n_test, double *test_data, int *mars_matrix_test,
                                    double *md, int *md_size, double *copd, int *copd_size,
                                    int *num_samples, int *burn_in, int *mcmc_step,
                                    int *initial_K, int *initial_clustering, int *initial_adj, double *initial_params,
                                    int *options, double *output) 
    {        
        GetRNGstate();
        
        Rprintf("#MCMC samples = %d\n", *num_samples);
        Tree **initial_trees = new Tree*[*initial_K];
        for (int i = 0; i < *initial_K; i++)
            initial_trees[i] = new Tree(&initial_adj[*p * *p * i], (unsigned int) *p);
        MCMCStructure *mcmc_treevariate = new MCMCStructure((unsigned int) *n, (unsigned int) *p, 
                                              *a, *b, data, mars_matrix, md, (unsigned int) *md_size,
                                              copd, (unsigned int) *copd_size, 
                                              (unsigned int) *initial_K, initial_clustering, initial_trees, initial_params);

        /* Execute sampling and store predictive log-likelihood (if applicable) */
        Rprintf("Starting to sample\n");
        if (*n_test > 0 && options[1] != 1)
            mcmc_treevariate->setTestSet((unsigned int) *n_test, test_data, mars_matrix_test);
        if (options[0] == 1 || *n_test > 0)
            mcmc_treevariate->setBurnIn((unsigned int) *burn_in);
        mcmc_treevariate->setMCMCStep((unsigned int) *mcmc_step);
        mcmc_treevariate->setTreeProposal(options[3]);
        if (options[4] == 1) {
            Rprintf("Marginals are fixed...\n");
            mcmc_treevariate->setMarginalUpdater(false);
        }
        if (options[5] == 1) {
            Rprintf("Using slice sampler for copulas...\n");
            mcmc_treevariate->setCopulaSliceSampler(true);
        }
        if (options[6] == 1) {
            Rprintf("Explicitly generating samples for dettached copulas...\n");
            mcmc_treevariate->setIndCopulaUpdate(true);
        }
        mcmc_treevariate->Rounds((unsigned int) *num_samples, options[1], options[0], output);
        
        /* Execute the options */

        if (options[1] == 1 && *n_test > 0) {
           /* Compute predictive log-likelihood from log files (to be used only if test set is too large and 
              predictions don't fit in memory). */
           output[0] = mcmc_treevariate->predLogLikelihood((unsigned int) *num_samples, 
                      (unsigned int) *burn_in, test_data, (unsigned int) *n_test, mars_matrix_test);
           Rprintf("Predictive log-likelihood = %f\n", output[0]);
        }

        /* Clean-up */
        for (int i = 0; i < *initial_K; i++)
            delete initial_trees[i];
        delete[] initial_trees; delete mcmc_treevariate;
        PutRNGstate();
    }

}

